@@ -19,7 +19,6 @@
 //
 // Copyright (c) 2006 Novell, Inc.
 //
-#if NET_2_0
 
 using System;
 using System.Drawing;
@@ -35,4 +34,3 @@ namespace System.Windows.Forms {
 	}
 
 }
-#endif
