@@ -26,7 +26,6 @@
 //	Jonathan Pobst (monkey@jpobst.com)
 //
 
-#if NET_2_0	
 using System.Drawing;
 using System.Runtime.InteropServices;
 using System.ComponentModel;
@@ -225,4 +224,3 @@ namespace System.Windows.Forms
 		#endregion
 	}
 }
-#endif
