@@ -12,6 +12,7 @@
 //     Lluis Sanchez Gual (lluis@ximian.com)
 //
 // (c)copyright 2003 Ville Palo
+// Copyright 2011 Xamarin Inc.
 //
 
 //
@@ -54,6 +55,7 @@ namespace System.Data {
 		private DataTable table;
 		private Hashtable DiffGrRows = new Hashtable ();
 		private Hashtable ErrorRows = new Hashtable ();
+
 		#endregion // Fields
 
 		#region ctors
@@ -221,7 +223,7 @@ namespace System.Data {
 
 		private void LoadColumnAttributes (DataTable Table, DataRow Row,
 			XmlReader reader, DataRowVersion loadType)
-		{			
+		{
 			if (!reader.HasAttributes // this check will be faster
 				|| !reader.MoveToFirstAttribute ())
 				return;
@@ -272,10 +274,8 @@ namespace System.Data {
 				{
 					object data = XmlDataLoader.StringToObject (Table.Columns[colName].DataType, reader.ReadString ());
 					
-					if (loadType == DataRowVersion.Current)
-						Row [colName] = data;
-					else
-						Row.SetOriginalValue (colName, data);
+					if (loadType == DataRowVersion.Current) Row [colName] = data;
+					else Row.SetOriginalValue (colName, data);
 					reader.Read ();
 				}
 				else 
@@ -290,6 +290,7 @@ namespace System.Data {
 						reader.Skip ();
 				}
 			}
+			
 			reader.ReadEndElement ();
 		}
 
@@ -351,6 +352,7 @@ namespace System.Data {
 
 		DataTable GetTable (string name)
 		{
+			name = XmlConvert.DecodeName (name);
 			if (DSet != null) 
 				return DSet.Tables [name];
 			else if (name == table.TableName) 
