@@ -46,6 +46,8 @@ namespace System.ServiceModel.MonoInternal
 	{
 		ContractDescription Contract { get; }
 
+		OperationContext Context { set; }
+
 		object Process (MethodBase method, string operationName, object [] parameters);
 
 		IAsyncResult BeginProcess (MethodBase method, string operationName, object [] parameters, AsyncCallback callback, object asyncState);
@@ -147,6 +149,10 @@ namespace System.ServiceModel.MonoInternal
 			get { return channel as IDuplexChannel; }
 		}
 
+		public OperationContext Context {
+			set { context = value; }
+		}
+
 		#region IClientChannel
 
 		bool did_interactive_initialization;
@@ -198,7 +204,6 @@ namespace System.ServiceModel.MonoInternal
 				}
 			}
 
-#if !MOONLIGHT
 			public override bool WaitOne (int millisecondsTimeout, bool exitContext)
 			{
 				return WaitHandle.WaitAll (ResultWaitHandles, millisecondsTimeout, exitContext);
@@ -208,7 +213,6 @@ namespace System.ServiceModel.MonoInternal
 			{
 				return WaitHandle.WaitAll (ResultWaitHandles, timeout, exitContext);
 			}
-#endif
 		}
 
 		class DisplayUIAsyncResult : IAsyncResult
@@ -445,31 +449,43 @@ namespace System.ServiceModel.MonoInternal
 				throw new InvalidOperationException ("another operation is in progress");
 			context = OperationContext.Current;
 
-			return _processDelegate.BeginInvoke (method, operationName, parameters, callback, asyncState);
+			try {
+				return _processDelegate.BeginInvoke (method, operationName, parameters, callback, asyncState);
+			} catch {
+				context = null;
+				throw;
+			}
 		}
 
 		public object EndProcess (MethodBase method, string operationName, object [] parameters, IAsyncResult result)
 		{
-			context = null;
-			if (result == null)
-				throw new ArgumentNullException ("result");
-			if (parameters == null)
-				throw new ArgumentNullException ("parameters");
-			// FIXME: the method arguments should be verified to be 
-			// identical to the arguments in the corresponding begin method.
-			return _processDelegate.EndInvoke (result);
+			try {
+				if (result == null)
+					throw new ArgumentNullException ("result");
+				if (parameters == null)
+					throw new ArgumentNullException ("parameters");
+				// FIXME: the method arguments should be verified to be 
+				// identical to the arguments in the corresponding begin method.
+				return _processDelegate.EndInvoke (result);
+			} finally {
+				context = null;
+			}
 		}
 
 		public object Process (MethodBase method, string operationName, object [] parameters)
 		{
+			var previousContext = OperationContext.Current;
 			try {
+				// Inherit the context from the calling thread
+				if (this.context != null) 
+					OperationContext.Current = this.context;
+
 				return DoProcess (method, operationName, parameters);
 			} catch (Exception ex) {
-#if MOONLIGHT // just for debugging
-				Console.Write ("Exception in async operation: ");
-				Console.WriteLine (ex);
-#endif
 				throw;
+			} finally {
+				// Reset the context before the thread goes back into the pool
+				OperationContext.Current = previousContext;
 			}
 		}
 
@@ -529,14 +545,12 @@ namespace System.ServiceModel.MonoInternal
 						Type detailType = typeof (ExceptionDetail);
 						var freader = fault.GetReaderAtDetailContents ();
 						DataContractSerializer ds = null;
-#if !NET_2_1
 						foreach (var fci in op.FaultContractInfos)
 							if (res.Headers.Action == fci.Action || fci.Serializer.IsStartObject (freader)) {
 								detailType = fci.Detail;
 								ds = fci.Serializer;
 								break;
 							}
-#endif
 						if (ds == null)
 							ds = new DataContractSerializer (detailType);
 						var detail = ds.ReadObject (freader);
