@@ -36,33 +36,44 @@ using System.Collections;
 using System.Collections.Generic;
 using System.Collections.Specialized;
 using System.Reflection;
-#if MONOWEB_DEP
 using Mono.Web.Util;
-#endif
 using System.Xml;
 using System.Configuration;
 using System.Configuration.Internal;
 using _Configuration = System.Configuration.Configuration;
 using System.Web.Util;
 using System.Threading;
+using System.Web.Hosting;
 
 namespace System.Web.Configuration {
 
 	public static class WebConfigurationManager
 	{
+		sealed class ConfigPath 
+		{
+			public string Path;
+			public bool InAnotherApp;
+
+			public ConfigPath (string path, bool inAnotherApp)
+			{
+				this.Path = path;
+				this.InAnotherApp = inAnotherApp;
+			}
+		}
+		
 		const int SAVE_LOCATIONS_CHECK_INTERVAL = 6000; // milliseconds
+		const int SECTION_CACHE_LOCK_TIMEOUT = 200; // milliseconds
 
 		static readonly char[] pathTrimChars = { '/' };
 		static readonly object suppressAppReloadLock = new object ();
 		static readonly object saveLocationsCacheLock = new object ();
-#if SYSTEMCORE_DEP
-		static readonly ReaderWriterLockSlim sectionCacheLock;
-#endif
 		
+		// See comment for the cacheLock field at top of System.Web.Caching/Cache.cs
+		static readonly ReaderWriterLockSlim sectionCacheLock;
+
 #if !TARGET_J2EE
 		static IInternalConfigConfigurationFactory configFactory;
 		static Hashtable configurations = Hashtable.Synchronized (new Hashtable ());
-		static Dictionary <int, object> sectionCache = new Dictionary <int, object> ();
 		static Hashtable configPaths = Hashtable.Synchronized (new Hashtable ());
 		static bool suppressAppReload;
 #else
@@ -174,9 +185,28 @@ namespace System.Web.Configuration {
 				}
 			}
 		}
+
+		const int DEFAULT_SECTION_CACHE_SIZE = 100;
+		const string CACHE_SIZE_OVERRIDING_KEY = "MONO_ASPNET_WEBCONFIG_CACHESIZE";
+		static LruCache<int, object> sectionCache;
 		
 		static WebConfigurationManager ()
 		{
+			var section_cache_size = DEFAULT_SECTION_CACHE_SIZE;
+			int section_cache_size_override;
+			bool size_overriden = false;
+			if (int.TryParse (Environment.GetEnvironmentVariable (CACHE_SIZE_OVERRIDING_KEY), out section_cache_size_override)) {
+				section_cache_size = section_cache_size_override;
+				size_overriden = true;
+				Console.WriteLine ("WebConfigurationManager's LRUcache Size overriden to: {0} (via {1})", section_cache_size_override, CACHE_SIZE_OVERRIDING_KEY);
+			}
+			sectionCache = new LruCache<int, object> (section_cache_size);
+			string eviction_warning = "WebConfigurationManager's LRUcache evictions count reached its max size";
+			if (!size_overriden)
+				eviction_warning += String.Format ("{0}Cache Size: {1} (overridable via {2})",
+				                                   Environment.NewLine, section_cache_size, CACHE_SIZE_OVERRIDING_KEY);
+			sectionCache.EvictionWarning = eviction_warning;
+
 			configFactory = ConfigurationManager.ConfigurationFactory;
 			_Configuration.SaveStart += ConfigurationSaveHandler;
 			_Configuration.SaveEnd += ConfigurationSaveHandler;
@@ -189,9 +219,7 @@ namespace System.Web.Configuration {
 					fi.SetValue (null, typeof (ApplicationSettingsConfigurationFileMap));
 			}
 
-#if SYSTEMCORE_DEP
 			sectionCacheLock = new ReaderWriterLockSlim ();
-#endif
 		}
 
 		static void ReenableWatcherOnConfigLocation (object state)
@@ -217,18 +245,11 @@ namespace System.Web.Configuration {
 		
 		static void ConfigurationSaveHandler (_Configuration sender, ConfigurationSaveEventArgs args)
 		{
-			bool locked = false;
 			try {
-#if SYSTEMCORE_DEP
 				sectionCacheLock.EnterWriteLock ();
-#endif
-				locked = true;
 				sectionCache.Clear ();
 			} finally {
-#if SYSTEMCORE_DEP
-				if (locked)
-					sectionCacheLock.ExitWriteLock ();
-#endif
+				sectionCacheLock.ExitWriteLock ();
 			}
 			
 			lock (suppressAppReloadLock) {
@@ -334,15 +355,16 @@ namespace System.Web.Configuration {
 			if (String.IsNullOrEmpty (path))
 				path = "/";
 
+			bool inAnotherApp = false;
 			if (!fweb && !String.IsNullOrEmpty (path))
-				path = FindWebConfig (path);
+				path = FindWebConfig (path, out inAnotherApp);
 
 			string confKey = path + site + locationSubPath + server + userName + password;
 			_Configuration conf = null;
 			conf = (_Configuration) configurations [confKey];
 			if (conf == null) {
 				try {
-					conf = ConfigurationFactory.Create (typeof (WebConfigurationHost), null, path, site, locationSubPath, server, userName, password);
+					conf = ConfigurationFactory.Create (typeof (WebConfigurationHost), null, path, site, locationSubPath, server, userName, password, inAnotherApp);
 					configurations [confKey] = conf;
 				} catch (Exception ex) {
 					lock (hasConfigErrorsLock) {
@@ -425,7 +447,7 @@ namespace System.Web.Configuration {
 			defaultConfiguration = cnew;
 			return true;
 		}
-
+		
 		internal static object GetSection (string sectionName, string path, HttpContext context)
 		{
 			if (String.IsNullOrEmpty (sectionName))
@@ -437,7 +459,6 @@ namespace System.Web.Configuration {
 			int cacheKey;
 			bool pathPresent = !String.IsNullOrEmpty (path);
 			string locationPath = null;
-			bool locked = false;
 
 			if (pathPresent)
 				locationPath = "location_" + path;
@@ -447,10 +468,7 @@ namespace System.Web.Configuration {
 				baseCacheKey ^= configPath.GetHashCode ();
 			
 			try {
-#if SYSTEMCORE_DEP
-				sectionCacheLock.EnterReadLock ();
-#endif
-				locked = true;
+				sectionCacheLock.EnterWriteLock ();
 				
 				object o;
 				if (pathPresent) {
@@ -466,10 +484,7 @@ namespace System.Web.Configuration {
 				if (sectionCache.TryGetValue (baseCacheKey, out o))
 					return o;
 			} finally {
-#if SYSTEMCORE_DEP
-				if (locked)
-					sectionCacheLock.ExitReadLock ();
-#endif
+				sectionCacheLock.ExitWriteLock ();
 			}
 
 			string cachePath = null;
@@ -486,11 +501,9 @@ namespace System.Web.Configuration {
 				} else
 					relPath = path;
 
-				_Configuration cnew;
-
 				HttpRequest req = context != null ? context.Request : null;
 				if (req != null) {
-					string vdir = VirtualPathUtility.GetDirectory (req.Path);
+					string vdir = VirtualPathUtility.GetDirectory (req.PathNoValidation);
 					if (vdir != null) {
 						vdir = vdir.TrimEnd (pathTrimChars);
 						if (String.Compare (c.ConfigPath, vdir, StringComparison.Ordinal) != 0 && LookUpLocation (vdir.Trim (pathTrimChars), ref c))
@@ -516,11 +529,7 @@ namespace System.Web.Configuration {
 				value = collection;
 			}
 #else
-#if MONOWEB_DEP
 			object value = SettingsMappingManager.MapSection (get_runtime_object.Invoke (section, new object [0]));
-#else
-			object value = null;
-#endif
 #endif
 			if (cachePath != null)
 				cacheKey = baseCacheKey ^ cachePath.GetHashCode ();
@@ -564,59 +573,91 @@ namespace System.Web.Configuration {
 			
 			return curPath.Substring (0, idx);
 		}
-		
+
 		internal static string FindWebConfig (string path)
 		{
+			bool dummy;
+
+			return FindWebConfig (path, out dummy);
+		}
+		
+		internal static string FindWebConfig (string path, out bool inAnotherApp)
+		{
+			inAnotherApp = false;
+			
 			if (String.IsNullOrEmpty (path))
 				return path;
-
-			string dir;
-			if (path [path.Length - 1] == '/')
-				dir = path;
-			else {
-				dir = VirtualPathUtility.GetDirectory (path, false);
-				if (dir == null)
-					return path;
+				
+			if (HostingEnvironment.VirtualPathProvider != null) {
+				if (HostingEnvironment.VirtualPathProvider.DirectoryExists (path))
+					path = VirtualPathUtility.AppendTrailingSlash (path);
 			}
+				
 			
-			string curPath = configPaths [dir] as string;
-			if (curPath != null)
-				return curPath;
+			string rootPath = HttpRuntime.AppDomainAppVirtualPath;
+			ConfigPath curPath;
+			curPath = configPaths [path] as ConfigPath;
+			if (curPath != null) {
+				inAnotherApp = curPath.InAnotherApp;
+				return curPath.Path;
+			}
 			
 			HttpContext ctx = HttpContext.Current;
 			HttpRequest req = ctx != null ? ctx.Request : null;
+			string physPath = req != null ? VirtualPathUtility.AppendTrailingSlash (MapPath (req, path)) : null;
+			string appDomainPath = HttpRuntime.AppDomainAppPath;
+			
+			if (physPath != null && appDomainPath != null && !physPath.StartsWith (appDomainPath, StringComparison.Ordinal))
+				inAnotherApp = true;
+			
+			string dir;
+			if (inAnotherApp || path [path.Length - 1] == '/')
+				dir = path;
+			else {
+			 	dir = VirtualPathUtility.GetDirectory (path, false);
+			 	if (dir == null)
+			 		return path;
+			}
+			
+			curPath = configPaths [dir] as ConfigPath;
+			if (curPath != null) {
+				inAnotherApp = curPath.InAnotherApp;
+				return curPath.Path;
+			}
+			
 			if (req == null)
 				return path;
 
-			curPath = path;
-			string rootPath = HttpRuntime.AppDomainAppVirtualPath;
-			string physPath;
-
-			while (String.Compare (curPath, rootPath, StringComparison.Ordinal) != 0) {
-				physPath = MapPath (req, curPath);
+			curPath = new ConfigPath (path, inAnotherApp);
+			while (String.Compare (curPath.Path, rootPath, StringComparison.Ordinal) != 0) {
+				physPath = MapPath (req, curPath.Path);
 				if (physPath == null) {
-					curPath = rootPath;
+					curPath.Path = rootPath;
 					break;
 				}
 
 				if (WebConfigurationHost.GetWebConfigFileName (physPath) != null)
 					break;
 				
-				curPath = GetParentDir (rootPath, curPath);
-				if (curPath == null || curPath == "~") {
-					curPath = rootPath;
+				curPath.Path = GetParentDir (rootPath, curPath.Path);
+				if (curPath.Path == null || curPath.Path == "~") {
+					curPath.Path = rootPath;
 					break;
 				}
 			}
 
-			configPaths [dir] = curPath;
-			return curPath;
+			if (String.Compare (curPath.Path, path, StringComparison.Ordinal) != 0)
+				configPaths [path] = curPath;
+			else
+				configPaths [dir] = curPath;
+			
+			return curPath.Path;
 		}
 		
 		static string GetCurrentPath (HttpContext ctx)
 		{
 			HttpRequest req = ctx != null ? ctx.Request : null;
-			return req != null ? req.Path : HttpRuntime.AppDomainAppVirtualPath;
+			return req != null ? req.PathNoValidation : HttpRuntime.AppDomainAppVirtualPath;
 		}
 		
 		internal static bool SuppressAppReload (bool newValue)
@@ -636,9 +677,7 @@ namespace System.Web.Configuration {
 			configurations.Remove (GetCurrentPath (ctx));
 		}
 
-#if TARGET_J2EE || MONOWEB_DEP
 		readonly static MethodInfo get_runtime_object = typeof (ConfigurationSection).GetMethod ("GetRuntimeObject", BindingFlags.NonPublic | BindingFlags.Instance);
-#endif		
 
 		public static object GetWebApplicationSection (string sectionName)
 		{
@@ -663,35 +702,21 @@ namespace System.Web.Configuration {
 		static void AddSectionToCache (int key, object section)
 		{
 			object cachedSection;
-			bool locked = false;
 
+			bool locked = false;
 			try {
-#if SYSTEMCORE_DEP
-				sectionCacheLock.EnterUpgradeableReadLock ();
-#endif
+				if (!sectionCacheLock.TryEnterWriteLock (SECTION_CACHE_LOCK_TIMEOUT))
+					return;
 				locked = true;
-					
+
 				if (sectionCache.TryGetValue (key, out cachedSection) && cachedSection != null)
 					return;
 
-				bool innerLocked = false;
-				try {
-#if SYSTEMCORE_DEP
-					sectionCacheLock.EnterWriteLock ();
-#endif
-					innerLocked = true;
-					sectionCache.Add (key, section);
-				} finally {
-#if SYSTEMCORE_DEP
-					if (innerLocked)
-						sectionCacheLock.ExitWriteLock ();
-#endif
-				}
+				sectionCache.Add (key, section);
 			} finally {
-#if SYSTEMCORE_DEP
-				if (locked)
-					sectionCacheLock.ExitUpgradeableReadLock ();
-#endif
+				if (locked) {
+					sectionCacheLock.ExitWriteLock ();
+				}
 			}
 		}
 		
