@@ -211,11 +211,14 @@ namespace System.Web
 				if (modcoll != null)
 					return;
 
+				bool mustNullContext = context == null;
 				try {
 					HttpModulesSection modules;
 					modules = (HttpModulesSection) WebConfigurationManager.GetWebApplicationSection ("system.web/httpModules");
 					HttpContext saved = HttpContext.Current;
 					HttpContext.Current = new HttpContext (new System.Web.Hosting.SimpleWorkerRequest (String.Empty, String.Empty, new StringWriter()));
+					if (context == null)
+						context = HttpContext.Current;
 					HttpModuleCollection coll = modules.LoadModules (this);
 					Interlocked.CompareExchange (ref modcoll, coll, null);
 					HttpContext.Current = saved;
@@ -227,6 +230,9 @@ namespace System.Web
 					}
 				} catch (Exception e) {
 					initialization_exception = e;
+				} finally {
+					if (mustNullContext)
+						context = null;
 				}
 			}
 		}
@@ -956,7 +962,13 @@ namespace System.Web
 				stop_processing = true;
 				PipelineDone ();
 			} catch (Exception e) {
-				ProcessError (e);
+				ThreadAbortException inner = e.InnerException as ThreadAbortException;
+				if (inner != null && FlagEnd.Value == inner.ExceptionState && !HttpRuntime.DomainUnloading) {
+					context.ClearError ();
+					Thread.ResetAbort ();
+				} else {
+					ProcessError (e);
+				}
 				stop_processing = true;
 				PipelineDone ();
 			}
@@ -1265,10 +1277,6 @@ namespace System.Web
 				context.Handler = handler;
 				context.PushHandler (handler);
 			} catch (FileNotFoundException fnf){
-#if TARGET_JVM
-				Console.WriteLine ("$$$$$$$$$$:Sys.Web Pipeline");
-				Console.WriteLine (fnf.ToString ());
-#endif
 				if (context.Request.IsLocal)
 					ProcessError (HttpException.NewWithCode (404,
 										 String.Format ("File not found {0}", fnf.FileName),
@@ -1498,9 +1506,7 @@ namespace System.Web
 					th.CurrentUICulture = new_app_culture;
 			}
 
-#if !TARGET_JVM
 			prev_user = Thread.CurrentPrincipal;
-#endif
 		}
 
 		void PostDone ()
@@ -1511,10 +1517,8 @@ namespace System.Web
 			}
 
 			Thread th = Thread.CurrentThread;
-#if !TARGET_JVM
 			if (Thread.CurrentPrincipal != prev_user)
 				Thread.CurrentPrincipal = prev_user;
-#endif
 			if (prev_appui_culture != null && prev_appui_culture != th.CurrentUICulture)
 				th.CurrentUICulture = prev_appui_culture;
 			if (prev_app_culture != null && prev_app_culture != th.CurrentCulture)
@@ -1660,11 +1664,7 @@ namespace System.Web
 			cultures [0] = Thread.CurrentThread.CurrentCulture;
 			cultures [1] = Thread.CurrentThread.CurrentUICulture;
 			
-#if TARGET_JVM
-			if (true)
-#else
 			if (Thread.CurrentThread.IsThreadPoolThread)
-#endif
 				Start (null);
 			else
 				ThreadPool.QueueUserWorkItem (x => {
@@ -1717,29 +1717,34 @@ namespace System.Web
 		bool RedirectCustomError (ref HttpException httpEx)
 		{
 			try {
-			if (!context.IsCustomErrorEnabledUnsafe)
-				return false;
+				if (!context.IsCustomErrorEnabledUnsafe)
+					return false;
 			
-			CustomErrorsSection config = (CustomErrorsSection)WebConfigurationManager.GetSection ("system.web/customErrors");			
-			if (config == null) {
-				if (context.ErrorPage != null)
-					return RedirectErrorPage (context.ErrorPage);
+				CustomErrorsSection config = (CustomErrorsSection)WebConfigurationManager.GetSection ("system.web/customErrors");			
+				if (config == null) {
+					if (context.ErrorPage != null)
+						return RedirectErrorPage (context.ErrorPage);
 				
-				return false;
-			}
+					return false;
+				}
+			
+				CustomError err = config.Errors [context.Response.StatusCode.ToString()];
+				string redirect = err == null ? null : err.Redirect;
+				if (redirect == null) {
+					redirect = context.ErrorPage;
+					if (redirect == null)
+						redirect = config.DefaultRedirect;
+				}
 			
-			CustomError err = config.Errors [context.Response.StatusCode.ToString()];
-			string redirect = err == null ? null : err.Redirect;
-			if (redirect == null) {
-				redirect = context.ErrorPage;
 				if (redirect == null)
-					redirect = config.DefaultRedirect;
-			}
-			
-			if (redirect == null)
-				return false;
-			
-			return RedirectErrorPage (redirect);
+					return false;
+
+				if (config.RedirectMode == CustomErrorsRedirectMode.ResponseRewrite) {
+					context.Server.Execute (redirect);
+					return true;
+				}
+				
+				return RedirectErrorPage (redirect);
 			}
 			catch (Exception ex) {
 				httpEx = HttpException.NewWithCode (500, String.Empty, ex, WebEventCodes.WebErrorOtherError);
@@ -1799,7 +1804,6 @@ namespace System.Web
 			if (type != null)
 				return type;
 
-#if !TARGET_JVM
 			Assembly [] assemblies = AppDomain.CurrentDomain.GetAssemblies ();
 			foreach (Assembly ass in assemblies) {
 				type = ass.GetType (typeName, false);
@@ -1828,7 +1832,6 @@ namespace System.Web
 			
 			if (type != null)
 				return type;
-#endif
 			if (throwOnMissing)
 				throw new TypeLoadException (String.Format ("Type '{0}' cannot be found", typeName), loadException);
 			
