@@ -9,7 +9,6 @@
 // Copyright (C) Tim Coleman, 2002
 // (C) 2003 Erik LeBel
 //
-
 //
 // Permission is hereby granted, free of charge, to any person obtaining
 // a copy of this software and associated documentation files (the
@@ -32,6 +31,7 @@
 //
 
 using System.Collections;
+using System.Collections.Generic;
 using System.Globalization;
 using System.Reflection;
 using System.Xml.Schema;
@@ -117,9 +117,9 @@ namespace System.Xml.Serialization {
 			string ns, 
 			XmlReflectionMember[] members, 
 			bool hasWrapperElement, 
-			bool writeAccessors)
+			bool rpc)
 		{
-			return ImportMembersMapping (elementName, ns, members, hasWrapperElement, writeAccessors, true);
+			return ImportMembersMapping (elementName, ns, members, hasWrapperElement, rpc, true);
 		}
 
 #if NET_2_0
@@ -130,10 +130,10 @@ namespace System.Xml.Serialization {
 			string ns, 
 			XmlReflectionMember[] members, 
 			bool hasWrapperElement, 
-			bool writeAccessors, 
-			bool validate)
+			bool rpc, 
+			bool openModel)
 		{
-			return ImportMembersMapping (elementName, ns, members, hasWrapperElement, writeAccessors, validate, XmlMappingAccess.Read | XmlMappingAccess.Write);
+			return ImportMembersMapping (elementName, ns, members, hasWrapperElement, rpc, openModel, XmlMappingAccess.Read | XmlMappingAccess.Write);
 		}
 
 #if NET_2_0
@@ -144,25 +144,29 @@ namespace System.Xml.Serialization {
 			string ns, 
 			XmlReflectionMember[] members, 
 			bool hasWrapperElement, 
-			bool writeAccessors, 
-			bool validate,
+			bool rpc, 
+			bool openModel,
 			XmlMappingAccess access)
 		{
 //			Reset ();	Disabled. See ChangeLog
 
-			XmlMemberMapping[] mapping = new XmlMemberMapping[members.Length];
+			ArrayList mapping = new ArrayList ();
 			for (int n=0; n<members.Length; n++)
 			{
 				XmlTypeMapMember mapMem = CreateMapMember (null, members[n], ns);
-				mapping[n] = new XmlMemberMapping (members[n].MemberName, ns, mapMem, false);
+				mapMem.GlobalIndex = n;
+				mapMem.CheckOptionalValueType (members);
+				mapping.Add (new XmlMemberMapping (members[n].MemberName, ns, mapMem, false));
 			}
 			elementName = XmlConvert.EncodeLocalName (elementName);
-			XmlMembersMapping mps = new XmlMembersMapping (elementName, ns, hasWrapperElement, false, mapping);
+			XmlMembersMapping mps = new XmlMembersMapping (elementName, ns, hasWrapperElement, false, (XmlMemberMapping[])mapping.ToArray (typeof(XmlMemberMapping)));
 			mps.RelatedMaps = relatedMaps;
 			mps.Format = SerializationFormat.Literal;
 			Type[] extraTypes = includedTypes != null ? (Type[])includedTypes.ToArray(typeof(Type)) : null;
+#if !NET_2_1
 			mps.Source = new MembersSerializationSource (elementName, hasWrapperElement, members, false, true, ns, extraTypes);
 			if (allowPrivateTypes) mps.Source.CanBeGenerated = false;
+#endif
 			return mps;
 		}
 
@@ -176,9 +180,9 @@ namespace System.Xml.Serialization {
 			return ImportTypeMapping (type, null, defaultNamespace);
 		}
 
-		public XmlTypeMapping ImportTypeMapping (Type type, XmlRootAttribute group)
+		public XmlTypeMapping ImportTypeMapping (Type type, XmlRootAttribute root)
 		{
-			return ImportTypeMapping (type, group, null);
+			return ImportTypeMapping (type, root, null);
 		}
 
 		public XmlTypeMapping ImportTypeMapping (Type type, XmlRootAttribute root, string defaultNamespace)
@@ -223,11 +227,17 @@ namespace System.Xml.Serialization {
 					default: throw new NotSupportedException ("Type " + typeData.Type.FullName + " not supported for XML stialization");
 				}
 
+#if NET_2_0
+				// bug #372780
+				map.SetKey (typeData.Type.ToString ());
+#endif
 				map.RelatedMaps = relatedMaps;
 				map.Format = SerializationFormat.Literal;
 				Type[] extraTypes = includedTypes != null ? (Type[]) includedTypes.ToArray (typeof (Type)) : null;
+#if !NET_2_1
 				map.Source = new XmlTypeSerializationSource (typeData.Type, root, attributeOverrides, defaultNamespace, extraTypes);
 				if (allowPrivateTypes) map.Source.CanBeGenerated = false;
+#endif
 				return map;
 			} catch (InvalidOperationException ex) {
 				throw new InvalidOperationException (string.Format (CultureInfo.InvariantCulture,
@@ -237,7 +247,8 @@ namespace System.Xml.Serialization {
 
 		XmlTypeMapping CreateTypeMapping (TypeData typeData, XmlRootAttribute root, string defaultXmlType, string defaultNamespace)
 		{
-			string rootNamespace = defaultNamespace;
+			bool hasTypeNamespace = !string.IsNullOrEmpty (defaultNamespace);
+			string rootNamespace = null;
 			string typeNamespace = null;
 			string elementName;
 			bool includeInSchema = true;
@@ -263,8 +274,10 @@ namespace System.Xml.Serialization {
 
 			if (atts.XmlType != null)
 			{
-				if (atts.XmlType.Namespace != null)
+				if (atts.XmlType.Namespace != null) {
 					typeNamespace = atts.XmlType.Namespace;
+					hasTypeNamespace = true;
+				}
 
 				if (atts.XmlType.TypeName != null && atts.XmlType.TypeName != string.Empty)
 					defaultXmlType = XmlConvert.EncodeLocalName (atts.XmlType.TypeName);
@@ -278,18 +291,20 @@ namespace System.Xml.Serialization {
 			{
 				if (root.ElementName.Length != 0)
 					elementName = XmlConvert.EncodeLocalName(root.ElementName);
-				if (root.Namespace != null)
+				if (root.Namespace != null) {
 					rootNamespace = root.Namespace;
+					hasTypeNamespace = true;
+				}
 				nullable = root.IsNullable;
 			}
 
-			if (rootNamespace == null) rootNamespace = "";
-			if (typeNamespace == null) typeNamespace = rootNamespace;
+			rootNamespace = rootNamespace ?? defaultNamespace ?? string.Empty;
+			typeNamespace = typeNamespace ?? rootNamespace;
 			
 			XmlTypeMapping map;
 			switch (typeData.SchemaType) {
 				case SchemaTypes.XmlSerializable:
-					map = new XmlSerializableMapping (elementName, rootNamespace, typeData, defaultXmlType, typeNamespace);
+					map = new XmlSerializableMapping (root, elementName, rootNamespace, typeData, defaultXmlType, typeNamespace);
 					break;
 				case SchemaTypes.Primitive:
 					if (!typeData.IsXsdType)
@@ -300,7 +315,7 @@ namespace System.Xml.Serialization {
 							typeData, defaultXmlType, typeNamespace);
 					break;
 				default:
-					map = new XmlTypeMapping (elementName, rootNamespace, typeData, defaultXmlType, typeNamespace);
+					map = new XmlTypeMapping (elementName, rootNamespace, typeData, defaultXmlType, hasTypeNamespace ? typeNamespace : null);
 					break;
 			}
 
@@ -311,21 +326,21 @@ namespace System.Xml.Serialization {
 			return map;
 		}
 
-		XmlTypeMapping ImportClassMapping (Type type, XmlRootAttribute root, string defaultNamespace)
+		XmlTypeMapping ImportClassMapping (Type type, XmlRootAttribute root, string defaultNamespace, bool isBaseType = false)
 		{
 			TypeData typeData = TypeTranslator.GetTypeData (type);
-			return ImportClassMapping (typeData, root, defaultNamespace);
+			return ImportClassMapping (typeData, root, defaultNamespace, isBaseType);
 		}
 
-		XmlTypeMapping ImportClassMapping (TypeData typeData, XmlRootAttribute root, string defaultNamespace)
+		XmlTypeMapping ImportClassMapping (TypeData typeData, XmlRootAttribute root, string defaultNamespace, bool isBaseType = false)
 		{
 			Type type = typeData.Type;
 
+			if (!allowPrivateTypes && !isBaseType)
+				ReflectionHelper.CheckSerializableType (type, false);
+
 			XmlTypeMapping map = helper.GetRegisteredClrType (type, GetTypeNamespace (typeData, root, defaultNamespace));
 			if (map != null) return map;
-
-			if (!allowPrivateTypes)
-				ReflectionHelper.CheckSerializableType (type, false);
 			
 			map = CreateTypeMapping (typeData, root, null, defaultNamespace);
 			helper.RegisterClrType (map, type, map.XmlTypeNamespace);
@@ -336,14 +351,30 @@ namespace System.Xml.Serialization {
 			ClassMap classMap = new ClassMap ();
 			map.ObjectMap = classMap;
 
-			ICollection members = GetReflectionMembers (type);
+			var members = GetReflectionMembers (type);
+			bool? isOrderExplicit = null;
+			foreach (XmlReflectionMember rmember in members)
+			{
+				int? order = rmember.XmlAttributes.Order;
+				if (isOrderExplicit == null)
+				{
+					if (order != null)
+						isOrderExplicit = (int) order >= 0;
+				}
+				else if (order != null && isOrderExplicit != ((int) order >= 0))
+					throw new InvalidOperationException ("Inconsistent XML sequence was detected. If there are XmlElement/XmlArray/XmlAnyElement attributes with explicit Order, then every other member must have an explicit order too.");
+			}
+			if (isOrderExplicit == true)
+				members.Sort ((m1, m2) => (int) m1.XmlAttributes.SortableOrder - (int) m2.XmlAttributes.SortableOrder);
+
 			foreach (XmlReflectionMember rmember in members)
 			{
 				string ns = map.XmlTypeNamespace;
 				if (rmember.XmlAttributes.XmlIgnore) continue;
 				if (rmember.DeclaringType != null && rmember.DeclaringType != type) {
-					XmlTypeMapping bmap = ImportClassMapping (rmember.DeclaringType, root, defaultNamespace);
-					ns = bmap.XmlTypeNamespace;
+					XmlTypeMapping bmap = ImportClassMapping (rmember.DeclaringType, root, defaultNamespace, true);
+					if (bmap.HasXmlTypeNamespace)
+						ns = bmap.XmlTypeNamespace;
 				}
 
 				try {
@@ -369,7 +400,7 @@ namespace System.Xml.Serialization {
 
 			if (type.BaseType != null)
 			{
-				XmlTypeMapping bmap = ImportClassMapping (type.BaseType, root, defaultNamespace);
+				XmlTypeMapping bmap = ImportClassMapping (type.BaseType, root, defaultNamespace, true);
 				ClassMap cbmap = bmap.ObjectMap as ClassMap;
 				
 				if (type.BaseType != typeof (object)) {
@@ -393,8 +424,8 @@ namespace System.Xml.Serialization {
 				XmlTypeMapMember mem = classMap.XmlTextCollector;
 				if (mem.TypeData.Type != typeof(string) && 
 				   mem.TypeData.Type != typeof(string[]) && 
-				   mem.TypeData.Type != typeof(object[]) && 
-				   mem.TypeData.Type != typeof(XmlNode[]))
+				   mem.TypeData.Type != typeof(XmlNode[]) && 
+				   mem.TypeData.Type != typeof(object[]))
 				   
 					throw new InvalidOperationException (String.Format (errSimple2, map.TypeData.TypeName, mem.Name, mem.TypeData.TypeName));
 			}
@@ -619,10 +650,9 @@ namespace System.Xml.Serialization {
 			map.IsNullable = false;
 			helper.RegisterClrType (map, type, map.XmlTypeNamespace);
 
-			string [] names = Enum.GetNames (type);
 			ArrayList members = new ArrayList();
-			foreach (string name in names)
-			{
+			string [] names = Enum.GetNames (type);
+			foreach (string name in names) {
 				FieldInfo field = type.GetField (name);
 				string xmlName = null;
 				if (field.IsDefined(typeof(XmlIgnoreAttribute), false))
@@ -664,7 +694,7 @@ namespace System.Xml.Serialization {
 			}
 		}
 
-		ICollection GetReflectionMembers (Type type)
+		List<XmlReflectionMember> GetReflectionMembers (Type type)
 		{
 			// First we want to find the inheritance hierarchy in reverse order.
 			Type currentType = type;
@@ -679,16 +709,6 @@ namespace System.Xml.Serialization {
 			// Read all Fields via reflection.
 			ArrayList fieldList = new ArrayList();
 			FieldInfo[] tfields = type.GetFields (BindingFlags.Instance | BindingFlags.Public);
-#if TARGET_JVM
-			// This statement ensures fields are ordered starting from the base type.
-			for (int ti=0; ti<typeList.Count; ti++) {
-				for (int i=0; i<tfields.Length; i++) {
-					FieldInfo field = tfields[i];
-					if (field.DeclaringType == typeList[ti])
-						fieldList.Add (field);
-				}
-			}
-#else
 			currentType = null;
 			int currentIndex = 0;
 			foreach (FieldInfo field in tfields)
@@ -701,22 +721,9 @@ namespace System.Xml.Serialization {
 				}
 				fieldList.Insert(currentIndex++, field);
 			}
-#endif
 			// Read all Properties via reflection.
 			ArrayList propList = new ArrayList();
 			PropertyInfo[] tprops = type.GetProperties (BindingFlags.Instance | BindingFlags.Public);
-#if TARGET_JVM
-			// This statement ensures properties are ordered starting from the base type.
-			for (int ti=0; ti<typeList.Count; ti++) {
-				for (int i=0; i<tprops.Length; i++) {
-					PropertyInfo prop = tprops[i];
-					if (!prop.CanRead) continue;
-					if (prop.GetIndexParameters().Length > 0) continue;
-					if (prop.DeclaringType == typeList[ti])
-						propList.Add (prop);
-				}
-			}
-#else
 			currentType = null;
 			currentIndex = 0;
 			foreach (PropertyInfo prop in tprops)
@@ -731,8 +738,7 @@ namespace System.Xml.Serialization {
 				if (prop.GetIndexParameters().Length > 0) continue;
 				propList.Insert(currentIndex++, prop);
 			}
-#endif
-			ArrayList members = new ArrayList();
+			var members = new List<XmlReflectionMember>();
 			int fieldIndex=0;
 			int propIndex=0;
 			// We now step through the type hierarchy from the base (object) through
@@ -767,7 +773,11 @@ namespace System.Xml.Serialization {
 						XmlAttributes atts = attributeOverrides[type, prop.Name];
 						if (atts == null) atts = new XmlAttributes (prop);
 						if (atts.XmlIgnore) continue;
-						if (!prop.CanWrite && (TypeTranslator.GetTypeData (prop.PropertyType).SchemaType != SchemaTypes.Array || prop.PropertyType.IsArray)) continue;
+						if (!prop.CanWrite) {
+							if (prop.PropertyType.IsInterface && typeof (IEnumerable).IsAssignableFrom (prop.PropertyType)) continue;
+							if (prop.PropertyType.IsGenericType && TypeData.GetGenericListItemType (prop.PropertyType) == null) continue; // check this before calling GetTypeData() which raises error for missing Add(). See bug #704813.
+							if (TypeTranslator.GetTypeData (prop.PropertyType).SchemaType != SchemaTypes.Array || prop.PropertyType.IsArray) continue;
+						}
 						XmlReflectionMember member = new XmlReflectionMember(prop.Name, prop.PropertyType, atts);
 						member.DeclaringType = prop.DeclaringType;
 						members.Add(member);
@@ -775,7 +785,8 @@ namespace System.Xml.Serialization {
 					else break;
 				}
 			}
-			return members;		
+			
+			return members;
 		}
 		
 		private XmlTypeMapMember CreateMapMember (Type declaringType, XmlReflectionMember rmember, string defaultNamespace)
@@ -787,7 +798,8 @@ namespace System.Xml.Serialization {
 			if (atts.XmlArray != null) {
 				if (atts.XmlArray.Namespace != null && atts.XmlArray.Form == XmlSchemaForm.Unqualified)
 					throw new InvalidOperationException ("XmlArrayAttribute.Form must not be Unqualified when it has an explicit Namespace value.");
-				if (typeData.SchemaType != SchemaTypes.Array)
+				if (typeData.SchemaType != SchemaTypes.Array &&
+				    !(typeData.SchemaType == SchemaTypes.Primitive && typeData.Type == typeof (byte [])))
 					throw new InvalidOperationException ("XmlArrayAttribute can be applied to members of array or collection type.");
 			}
 
@@ -801,18 +813,13 @@ namespace System.Xml.Serialization {
 				else
 					throw new InvalidOperationException ("XmlAnyAttributeAttribute can only be applied to members of type XmlAttribute[] or XmlNode[]");
 			}
-			else if (atts.XmlAnyElements != null && atts.XmlAnyElements.Count > 0)
+			else
+			if (atts.XmlAnyElements != null && atts.XmlAnyElements.Count > 0)
 			{
-				if ( (rmember.MemberType.FullName == "System.Xml.XmlElement[]") ||
-					 (rmember.MemberType.FullName == "System.Xml.XmlNode[]") ||
-					 (rmember.MemberType.FullName == "System.Xml.XmlElement"))
-				{
-					XmlTypeMapMemberAnyElement member = new XmlTypeMapMemberAnyElement();
-					member.ElementInfo = ImportAnyElementInfo (defaultNamespace, rmember, member, atts);
-					mapMember = member;
-				}
-				else
-					throw new InvalidOperationException ("XmlAnyElementAttribute can only be applied to members of type XmlElement, XmlElement[] or XmlNode[]");
+				// no XmlNode type check is done here (seealso: bug #553032).
+				XmlTypeMapMemberAnyElement member = new XmlTypeMapMemberAnyElement();
+				member.ElementInfo = ImportAnyElementInfo (defaultNamespace, rmember, member, atts);
+				mapMember = member;
 			}
 			else if (atts.Xmlns)
 			{
@@ -832,7 +839,7 @@ namespace System.Xml.Serialization {
 				else 
 					mapAttribute.AttributeName = atts.XmlAttribute.AttributeName;
 
-				mapAttribute.AttributeName = XmlConvert.EncodeLocalName (mapAttribute.AttributeName);
+				mapAttribute.AttributeName = XmlConvert.EncodeName (mapAttribute.AttributeName);
 
 				if (typeData.IsComplexType)
 					mapAttribute.MappedType = ImportTypeMapping (typeData.Type, null, defaultNamespace);
@@ -893,6 +900,7 @@ namespace System.Xml.Serialization {
 					elem.MappedType = ImportListMapping (rmember.MemberType, null, elem.Namespace, atts, 0);
 					elem.IsNullable = (atts.XmlArray != null) ? atts.XmlArray.IsNullable : false;
 					elem.Form = (atts.XmlArray != null) ? atts.XmlArray.Form : XmlSchemaForm.Qualified;
+					elem.ExplicitOrder = (atts.XmlArray != null) ? atts.XmlArray.Order : -1;
 					// This is a bit tricky, but is done
 					// after filling descendant members, so
 					// that array items could be serialized
@@ -926,7 +934,7 @@ namespace System.Xml.Serialization {
 			Type choiceEnumType = null;
 			
 			XmlTypeMapElementInfoList list = new XmlTypeMapElementInfoList();
-			ImportTextElementInfo (list, defaultType, member, atts);
+			ImportTextElementInfo (list, defaultType, member, atts, defaultNamespace);
 			
 			if (atts.XmlChoiceIdentifier != null) {
 				if (cls == null)
@@ -974,7 +982,12 @@ namespace System.Xml.Serialization {
 				elem.Form = att.Form;
 				if (elem.Form != XmlSchemaForm.Unqualified)
 					elem.Namespace = (att.Namespace != null) ? att.Namespace : defaultNamespace;
-				elem.IsNullable = att.IsNullable;
+
+				// elem may already be nullable, and IsNullable property in XmlElement is false by default
+				if (att.IsNullable && !elem.IsNullable)
+					elem.IsNullable = att.IsNullable;
+
+				elem.ExplicitOrder = att.Order;
 
 				if (elem.IsNullable && !elem.TypeData.IsNullable)
 					throw new InvalidOperationException ("IsNullable may not be 'true' for value type " + elem.TypeData.FullTypeName + " in member '" + defaultName + "'");
@@ -1004,13 +1017,16 @@ namespace System.Xml.Serialization {
 
 				if (choiceEnumMap != null) {
 					string cname = choiceEnumMap.GetEnumName (choiceEnumType.FullName, elem.ElementName);
+					if (cname == null && elem.Namespace != null)
+						cname = choiceEnumMap.GetEnumName (choiceEnumType.FullName,
+							elem.Namespace.ToString () + ":" + elem.ElementName);
 					if (cname == null)
 						throw new InvalidOperationException (string.Format (
 							CultureInfo.InvariantCulture, "Type {0} is missing"
 							+ " enumeration value '{1}' for element '{1} from"
 							+ " namespace '{2}'.", choiceEnumType, elem.ElementName,
 							elem.Namespace));
-					elem.ChoiceValue = Enum.Parse (choiceEnumType, cname);
+					elem.ChoiceValue = Enum.Parse (choiceEnumType, cname, false);
 				}
 					
 				list.Add (elem);
@@ -1022,7 +1038,7 @@ namespace System.Xml.Serialization {
 		{
 			XmlTypeMapElementInfoList list = new XmlTypeMapElementInfoList();
 
-			ImportTextElementInfo (list, rmember.MemberType, member, atts);
+			ImportTextElementInfo (list, rmember.MemberType, member, atts, defaultNamespace);
 
 			foreach (XmlAnyElementAttribute att in atts.XmlAnyElements)
 			{
@@ -1039,12 +1055,13 @@ namespace System.Xml.Serialization {
 					if (att.Namespace != null) 
 						throw new InvalidOperationException ("The element " + rmember.MemberName + " has been attributed with an XmlAnyElementAttribute and a namespace '" + att.Namespace + "', but no name. When a namespace is supplied, a name is also required. Supply a name or remove the namespace.");
 				}
+				elem.ExplicitOrder = att.Order;
 				list.Add (elem);
 			}
 			return list;
 		}
 
-		void ImportTextElementInfo (XmlTypeMapElementInfoList list, Type defaultType, XmlTypeMapMemberElement member, XmlAttributes atts)
+		void ImportTextElementInfo (XmlTypeMapElementInfoList list, Type defaultType, XmlTypeMapMemberElement member, XmlAttributes atts, string defaultNamespace)
 		{
 			if (atts.XmlText != null)
 			{
@@ -1067,6 +1084,8 @@ namespace System.Xml.Serialization {
 				 )
 					throw new InvalidOperationException ("XmlText cannot be used to encode complex types");
 
+				if (elem.TypeData.IsComplexType)
+					elem.MappedType = ImportTypeMapping (defaultType, null, defaultNamespace);
 				elem.IsTextElement = true;
 				elem.WrappedElement = false;
 				list.Add (elem);
@@ -1116,7 +1135,6 @@ namespace System.Xml.Serialization {
 			string namedValue = Enum.Format (typeData.Type, defaultValue, "g");
 			// get decimal representation of enum value
 			string decimalValue = Enum.Format (typeData.Type, defaultValue, "d");
-
 			// if decimal representation matches string representation, then
 			// the value is not defined in the enum type (as the "g" format
 			// will return the decimal equivalent of the value if the value
