@@ -12,6 +12,7 @@
 
 using System;
 using System.Collections;
+using System.IO;
 using System.Xml;
 using System.Xml.Schema;
 using System.Xml.Serialization;
@@ -1627,6 +1628,35 @@ namespace MonoTests.System.XmlSerialization
 		{
 			new XmlSerializer (typeof (MyCollection));
 		}
+
+		[Test]
+		public void Bug704813Type ()
+		{
+			var xs = new XmlSerializer (typeof (Bug704813Type));
+			xs.Serialize (TextWriter.Null, new Bug704813Type ());
+		}
+
+		[Test]
+		public void Bug708178Type()
+		{
+			string file = Path.Combine (Path.GetTempPath (), "Bug708178Type.xml");
+			XmlSerializer xmlSerializer = new XmlSerializer (typeof(Bug708178Type));
+			Bug708178Type bugType = new Bug708178Type ();
+			bugType.Foo.Add ("test");
+			Assert.AreEqual (1, bugType.Foo.Count);
+		 
+			//xml Serialize
+			TextWriter WriteFileStream = new StreamWriter (file, false);
+			xmlSerializer.Serialize (WriteFileStream, bugType);
+			WriteFileStream.Close ();
+		 
+			//xml Deserialize
+			FileStream ReadFileStream = new FileStream (file, FileMode.Open, FileAccess.Read, FileShare.Read);
+			Bug708178Type bugTypeReload = (Bug708178Type)xmlSerializer.Deserialize (ReadFileStream);
+		 
+			//should have deserialized the relationship
+			Assert.AreEqual(1, bugTypeReload.Foo.Count);
+	       }
 #endif
 
 		public class Employee : IXmlSerializable
@@ -2247,6 +2277,79 @@ namespace MonoTests.System.XmlSerialization
 			}
 		}
 #endif
+
+		public class Bug594490Class
+		{
+			[XmlAttribute ("xml:lang")]
+			public string GroupName;
+		}
+
+		[Test]
+		public void Bug594490_SerializationOfXmlLangAttribute ()
+		{
+			var serializer = new XmlSerializer (typeof(Bug594490Class));
+
+			using (var writer = new StringWriter ()) {
+				var obj = new Bug594490Class ();
+
+				obj.GroupName = "hello world";
+
+				serializer.Serialize (writer, obj);
+				writer.Close ();
+
+				Assert.AreEqual (@"<?xml version=""1.0"" encoding=""utf-16""?>
+<Bug594490Class xmlns:xsi=""http://www.w3.org/2001/XMLSchema-instance"" xmlns:xsd=""http://www.w3.org/2001/XMLSchema"" xml:lang=""hello world"" />",
+					writer.ToString (),
+					"Novell bug #594490 (https://bugzilla.novell.com/show_bug.cgi?id=594490) not fixed.");
+			}
+		}
+
+		/*
+		 * The following code was generated from Microsoft's xsd.exe with the /classes switch.
+		 * It only includes the relevent details but was based on the following namespaces:
+		 *   urn:oasis:names:tc:SAML:2.0:protocol
+		 *   urn:oasis:names:tc:SAML:2.0:assertion
+		 *   http://www.w3.org/2000/09/xmldsig#
+		 *   http://www.w3.org/2001/04/xmlenc
+		 */
+
+		[XmlTypeAttribute (Namespace = "urn:oasis:names:tc:SAML:2.0:protocol")]
+		[XmlRootAttribute ("RequestedAuthnContext", Namespace = "urn:oasis:names:tc:SAML:2.0:protocol", IsNullable = false)]
+		public class RequestedAuthnContext
+		{
+			string[] items;
+			ItemsChoice7[] itemsElementName;
+
+			[XmlElementAttribute ("AuthnContextClassRef", typeof (string), Namespace = "urn:oasis:names:tc:SAML:2.0:assertion", DataType = "anyURI")]
+			[XmlElementAttribute ("AuthnContextDeclRef", typeof (string), Namespace = "urn:oasis:names:tc:SAML:2.0:assertion", DataType = "anyURI")]
+			[XmlChoiceIdentifierAttribute ("ItemsElementName")]
+			public string[] Items {
+				get { return this.items; }
+				set { this.items = value; }
+			}
+
+			[XmlElementAttribute ("ItemsElementName")]
+			[XmlIgnoreAttribute ()]
+			public ItemsChoice7[] ItemsElementName {
+				get { return this.itemsElementName; }
+				set { this.itemsElementName = value; }
+			}
+		}
+
+		[XmlTypeAttribute (Namespace = "urn:oasis:names:tc:SAML:2.0:protocol", IncludeInSchema = false)]
+		public enum ItemsChoice7 {
+			[XmlEnumAttribute ("urn:oasis:names:tc:SAML:2.0:assertion:AuthnContextClassRef")]
+			AuthnContextClassRef,
+			[XmlEnumAttribute ("urn:oasis:names:tc:SAML:2.0:assertion:AuthnContextDeclRef")]
+			AuthnContextDeclRef,
+		}
+		// End snippet from xsd.exe
+	
+		[Test]
+		public void FullyQualifiedName_XmlEnumAttribute ()
+		{
+			var serializer = new XmlSerializer (typeof (RequestedAuthnContext)); 
+		}
 	}
 }
 
