@@ -27,11 +27,9 @@
 // WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 //
 
-#if NET_2_0 && CONFIGURATION_DEP
+#if CONFIGURATION_DEP
 
-#if !TARGET_JVM
 extern alias PrebuiltSystem;
-#endif
 
 using System;
 using System.Collections;
@@ -44,11 +42,7 @@ using System.Security.Cryptography;
 using System.Text;
 using System.Xml;
 
-#if TARGET_JVM
-using NameValueCollection = System.Collections.Specialized.NameValueCollection;
-#else
 using NameValueCollection = PrebuiltSystem.System.Collections.Specialized.NameValueCollection;
-#endif
 
 namespace System.Configuration
 {
@@ -80,7 +74,9 @@ namespace System.Configuration
 		// MachineConfigFilename will be used to set the ExeConfigFilename.
 		//
 		// This is necessary to fix bug #491531
+#pragma warning disable 649
 		private static Type webConfigurationFileMapType;
+#pragma warning restore 649
 		
 		private static string userRoamingPath = "";
 		private static string userLocalPath = "";
@@ -262,7 +258,6 @@ namespace System.Configuration
 				return attrs [0].Company;
 			}
 
-#if !TARGET_JVM
 			MethodInfo entryPoint = assembly.EntryPoint;
 			Type entryType = entryPoint != null ? entryPoint.DeclaringType : null;
 			if (entryType != null && !String.IsNullOrEmpty (entryType.Namespace)) {
@@ -270,9 +265,6 @@ namespace System.Configuration
 				return end < 0 ? entryType.Namespace : entryType.Namespace.Substring (0, end);
 			}
 			return "Program";
-#else
-			return assembly.GetName ().Name;
-#endif
 		}
 
 		private static string GetProductName ()
@@ -281,26 +273,17 @@ namespace System.Configuration
 			if (assembly == null)
 				assembly = Assembly.GetCallingAssembly ();
 
-#if !TARGET_JVM
-
 			byte [] pkt = assembly.GetName ().GetPublicKeyToken ();
 			return String.Format ("{0}_{1}_{2}",
 				AppDomain.CurrentDomain.FriendlyName,
-				pkt != null ? "StrongName" : "Url",
+				pkt != null && pkt.Length > 0 ? "StrongName" : "Url",
 				GetEvidenceHash());
-
-#else // AssemblyProductAttribute-based code
-			AssemblyProductAttribute [] attrs = (AssemblyProductAttribute[]) assembly.GetCustomAttributes (typeof (AssemblyProductAttribute), true);
-		
-			if ((attrs != null) && attrs.Length > 0) {
-				return attrs [0].Product;
-			}
-			return assembly.GetName ().Name;
-#endif
 		}
 
-		// FIXME: it seems that something else is used
-		// here, to convert hash bytes to string.
+		// Note: Changed from base64() to hex output to avoid unexpected chars like '\' or '/' with filesystem meaning.
+		//       Otherwise eventually filenames, which are invalid on linux or windows, might be created.
+		// Signed-off-by:  Carsten Schlote <schlote@vahanus.net>
+		// TODO: Compare with .NET. It might be also, that their way isn't suitable for Unix OS derivates (slahes in output)
 		private static string GetEvidenceHash ()
 		{
 			Assembly assembly = Assembly.GetEntryAssembly ();
@@ -308,9 +291,11 @@ namespace System.Configuration
 				assembly = Assembly.GetCallingAssembly ();
 
 			byte [] pkt = assembly.GetName ().GetPublicKeyToken ();
-			byte [] hash = SHA1.Create ().ComputeHash (pkt != null ? pkt : Encoding.UTF8.GetBytes (assembly.EscapedCodeBase));
-
-			return Convert.ToBase64String (hash);
+			byte [] hash = SHA1.Create ().ComputeHash (pkt != null && pkt.Length >0 ? pkt : Encoding.UTF8.GetBytes (assembly.EscapedCodeBase));
+			System.Text.StringBuilder evidence_string = new System.Text.StringBuilder();
+			foreach (byte b in hash)
+				evidence_string.AppendFormat("{0:x2}",b);
+			return evidence_string.ToString ();
 		}
 
 		private static string GetProductVersion ()
@@ -337,19 +322,15 @@ namespace System.Configuration
 				ProductVersion = GetProductVersion ().Split('.');
 
 			// C:\Documents and Settings\(user)\Application Data
-#if !TARGET_JVM
 			if (userRoamingBasePath == "")
 				userRoamingPath = Environment.GetFolderPath (Environment.SpecialFolder.ApplicationData);
 			else
-#endif
 				userRoamingPath = userRoamingBasePath;
 
 			// C:\Documents and Settings\(user)\Local Settings\Application Data (on Windows)
-#if !TARGET_JVM
 			if (userLocalBasePath == "")
 				userLocalPath = Environment.GetFolderPath (Environment.SpecialFolder.LocalApplicationData);
 			else
-#endif
 				userLocalPath = userLocalBasePath;
 
 			if (isCompany) {
@@ -362,13 +343,11 @@ namespace System.Configuration
 					Assembly assembly = Assembly.GetEntryAssembly ();
 					if (assembly == null)
 						assembly = Assembly.GetCallingAssembly ();
-#if !TARGET_JVM
 					byte [] pkt = assembly.GetName ().GetPublicKeyToken ();
 					ProductName = String.Format ("{0}_{1}_{2}",
 						ProductName,
 						pkt != null ? "StrongName" : "Url",
 						GetEvidenceHash());
-#endif
 				}
 				userRoamingPath = Path.Combine (userRoamingPath, ProductName);
 				userLocalPath = Path.Combine (userLocalPath, ProductName);
@@ -611,42 +590,49 @@ namespace System.Configuration
 			if (userGroup == null) {
 				userGroup = new UserSettingsGroup ();
 				config.SectionGroups.Add ("userSettings", userGroup);
-				ApplicationSettingsBase asb = context.CurrentSettings;
-				ClientSettingsSection cs = new ClientSettingsSection ();
-				userGroup.Sections.Add ((asb != null ? asb.GetType () : typeof (ApplicationSettingsBase)).FullName, cs);
+			}
+			ApplicationSettingsBase asb = context.CurrentSettings;
+			string class_name = NormalizeInvalidXmlChars ((asb != null ? asb.GetType () : typeof (ApplicationSettingsBase)).FullName);
+			ClientSettingsSection userSection = null;
+			ConfigurationSection cnf = userGroup.Sections.Get (class_name);
+			userSection = cnf as ClientSettingsSection;
+			if (userSection == null) {
+				userSection = new ClientSettingsSection ();
+				userGroup.Sections.Add (class_name, userSection);
 			}
 
 			bool hasChanges = false;
 
-			foreach (ConfigurationSection section in userGroup.Sections) {
-				ClientSettingsSection userSection = section as ClientSettingsSection;
-				if (userSection == null)
+			if (userSection == null)
+				return;
+
+			foreach (SettingsPropertyValue value in collection) {
+				if (checkUserLevel && value.Property.Attributes.Contains (typeof (SettingsManageabilityAttribute)) != isRoaming)
+					continue;
+				// The default impl does not save the ApplicationScopedSetting properties
+				if (value.Property.Attributes.Contains (typeof (ApplicationScopedSettingAttribute)))
 					continue;
 
-				foreach (SettingsPropertyValue value in collection) {
-					if (checkUserLevel && value.Property.Attributes.Contains (typeof (SettingsManageabilityAttribute)) != isRoaming)
-						continue;
-					hasChanges = true;
-					SettingElement element = userSection.Settings.Get (value.Name);
-					if (element == null) {
-						element = new SettingElement (value.Name, value.Property.SerializeAs);
-						userSection.Settings.Add (element);
-					}
-					if (element.Value.ValueXml == null)
-						element.Value.ValueXml = new XmlDocument ().CreateElement ("value");
-					switch (value.Property.SerializeAs) {
-					case SettingsSerializeAs.Xml:
-						element.Value.ValueXml.InnerXml = (value.SerializedValue as string) ?? string.Empty;
-						break;
-					case SettingsSerializeAs.String:
-						element.Value.ValueXml.InnerText = value.SerializedValue as string;
-						break;
-					case SettingsSerializeAs.Binary:
-						element.Value.ValueXml.InnerText = value.SerializedValue != null ? Convert.ToBase64String (value.SerializedValue as byte []) : string.Empty;
-						break;
-					default:
-						throw new NotImplementedException ();
-					}
+				hasChanges = true;
+				SettingElement element = userSection.Settings.Get (value.Name);
+				if (element == null) {
+					element = new SettingElement (value.Name, value.Property.SerializeAs);
+					userSection.Settings.Add (element);
+				}
+				if (element.Value.ValueXml == null)
+					element.Value.ValueXml = new XmlDocument ().CreateElement ("value");
+				switch (value.Property.SerializeAs) {
+				case SettingsSerializeAs.Xml:
+					element.Value.ValueXml.InnerXml = (value.SerializedValue as string) ?? string.Empty;
+					break;
+				case SettingsSerializeAs.String:
+					element.Value.ValueXml.InnerText = value.SerializedValue as string;
+					break;
+				case SettingsSerializeAs.Binary:
+					element.Value.ValueXml.InnerText = value.SerializedValue != null ? Convert.ToBase64String (value.SerializedValue as byte []) : string.Empty;
+					break;
+				default:
+					throw new NotImplementedException ();
 				}
 			}
 			if (hasChanges)
@@ -690,6 +676,20 @@ namespace System.Configuration
 #endif
 		}
 
+		// NOTE: We should add here all the chars that are valid in a name of a class (Ecma-wise),
+		// but invalid in an xml element name, and provide a better impl if we get too many of them.
+		string NormalizeInvalidXmlChars (string str)
+		{
+			char [] invalid_chars = new char [] { '+' };
+
+			if (str == null || str.IndexOfAny (invalid_chars) == -1)
+				return str;
+
+			// Replace with its hexadecimal values.
+			str = str.Replace ("+", "_x002B_");
+			return str;
+		}
+
 		private void LoadPropertyValue (SettingsPropertyCollection collection, SettingElement element, bool allowOverwrite)
 		{
 			SettingsProperty prop = collection [element.Name];
@@ -769,20 +769,19 @@ namespace System.Configuration
 		{
 			CreateExeMap ();
 
-			if (values == null) {
-				values = new SettingsPropertyValueCollection ();
-				string groupName = context ["GroupName"] as string;
-				LoadProperties (exeMapCurrent, collection, ConfigurationUserLevel.None, "applicationSettings", false, groupName);
-				LoadProperties (exeMapCurrent, collection, ConfigurationUserLevel.None, "userSettings", false, groupName);
+			values = new SettingsPropertyValueCollection ();
+			string groupName = context ["GroupName"] as string;
+			groupName = NormalizeInvalidXmlChars (groupName); // we likely saved the element removing the non valid xml chars.
+			LoadProperties (exeMapCurrent, collection, ConfigurationUserLevel.None, "applicationSettings", false, groupName);
+			LoadProperties (exeMapCurrent, collection, ConfigurationUserLevel.None, "userSettings", false, groupName);
 
-				LoadProperties (exeMapCurrent, collection, ConfigurationUserLevel.PerUserRoaming, "userSettings", true, groupName);
-				LoadProperties (exeMapCurrent, collection, ConfigurationUserLevel.PerUserRoamingAndLocal, "userSettings", true, groupName);
+			LoadProperties (exeMapCurrent, collection, ConfigurationUserLevel.PerUserRoaming, "userSettings", true, groupName);
+			LoadProperties (exeMapCurrent, collection, ConfigurationUserLevel.PerUserRoamingAndLocal, "userSettings", true, groupName);
 
-				// create default values if not exist
-				foreach (SettingsProperty p in collection)
-					if (values [p.Name] == null)
-						values.Add (new SettingsPropertyValue (p));
-			}
+			// create default values if not exist
+			foreach (SettingsProperty p in collection)
+				if (values [p.Name] == null)
+					values.Add (new SettingsPropertyValue (p));
 			return values;
 		}
 
@@ -834,13 +833,13 @@ namespace System.Configuration
 
 		public void Reset (SettingsContext context)
 		{
-			SettingsPropertyCollection coll = new SettingsPropertyCollection ();
-			GetPropertyValues (context, coll);
-			foreach (SettingsPropertyValue propertyValue in values) {
-				propertyValue.PropertyValue = propertyValue.Property.DefaultValue;
-				propertyValue.IsDirty = true;
+			if (values != null) {
+				foreach (SettingsPropertyValue propertyValue in values) {
+					// Can't use propertyValue.Property.DefaultValue
+					// as it may cause InvalidCastException (see bug# 532180)
+					values[propertyValue.Name].PropertyValue = propertyValue.Reset ();
+				}
 			}
-			SetPropertyValues (context, values);
 		}
 
 		// FIXME: implement
