@@ -35,6 +35,7 @@ using System;
 using System.Collections;
 using System.Collections.Specialized;
 using System.Configuration;
+using System.Reflection;
 using System.Threading;
 #if (XML_DEP)
 using System.Xml;
@@ -81,11 +82,11 @@ namespace System.Diagnostics
 		}
 	}
 #if (XML_DEP)
-#if NET_2_0
 	[Obsolete ("This class is obsoleted")]
-#endif
 	public class DiagnosticsConfigurationHandler : IConfigurationSectionHandler
 	{
+		TraceImplSettings configValues;
+
 		delegate void ElementHandler (IDictionary d, XmlNode node);
 
 		IDictionary elementHandlers = new Hashtable ();
@@ -95,9 +96,7 @@ namespace System.Diagnostics
 			elementHandlers ["assert"] = new ElementHandler (AddAssertNode);
 			elementHandlers ["switches"] = new ElementHandler (AddSwitchesNode);
 			elementHandlers ["trace"] = new ElementHandler (AddTraceNode);
-#if NET_2_0
 			elementHandlers ["sources"] = new ElementHandler (AddSourcesNode);
-#endif
 		}
 
 		public virtual object Create (object parent, object configContext, XmlNode section)
@@ -108,7 +107,11 @@ namespace System.Diagnostics
 			else
 				d = (IDictionary) ((ICloneable)parent).Clone();
 
-#if NET_2_0
+			if (d.Contains (TraceImplSettings.Key))
+				configValues = (TraceImplSettings) d [TraceImplSettings.Key];
+			else
+				d.Add (TraceImplSettings.Key, configValues = new TraceImplSettings ());
+
 			// process <sharedListeners> first
 			foreach (XmlNode child in section.ChildNodes) {
 				switch (child.NodeType) {
@@ -119,7 +122,6 @@ namespace System.Diagnostics
 					break;
 				}
 			}
-#endif
 
 			foreach (XmlNode child in section.ChildNodes) {
 				XmlNodeType type = child.NodeType;
@@ -130,10 +132,8 @@ namespace System.Diagnostics
 				case XmlNodeType.Comment:
 					continue;
 				case XmlNodeType.Element:
-#if NET_2_0
 					if (child.LocalName == "sharedListeners")
 						continue;
-#endif
 					ElementHandler eh = (ElementHandler) elementHandlers [child.Name];
 					if (eh != null)
 						eh (d, child);
@@ -169,7 +169,7 @@ namespace System.Diagnostics
 			if (logfilename != null)
 				d ["logfilename"] = logfilename;
 
-			DefaultTraceListener dtl = (DefaultTraceListener) TraceImpl.Listeners["Default"];
+			DefaultTraceListener dtl = (DefaultTraceListener) configValues.Listeners["Default"];
 			if (dtl != null) {
 				if (assertuienabled != null)
 					dtl.AssertUiEnabled = (bool) d ["assertuienabled"];
@@ -186,7 +186,6 @@ namespace System.Diagnostics
 		// them
 		private void AddSwitchesNode (IDictionary d, XmlNode node)
 		{
-#if !TARGET_JVM
 			// There are no attributes on <switch/>
 			ValidateInvalidAttributes (node.Attributes, node);
 
@@ -224,23 +223,11 @@ namespace System.Diagnostics
 			}
 
 			d [node.Name] = newNodes;
-#endif
 		}
 
 		private static object GetSwitchValue (string name, string value)
 		{
-#if NET_2_0
 			return value;
-#else
-			try {
-				return int.Parse (value);
-			} catch {
-				throw new ConfigurationException (string.Format (
-					"Error in trace switch '{0}': " + 
-					"The value of a switch must be integral",
-					name));
-			}
-#endif
 		}
 
 		private void AddTraceNode (IDictionary d, XmlNode node)
@@ -253,7 +240,7 @@ namespace System.Diagnostics
 					continue;
 				if (t == XmlNodeType.Element) {
 					if (child.Name == "listeners")
-						AddTraceListeners (d, child, TraceImpl.Listeners);
+						AddTraceListeners (d, child, configValues.Listeners);
 					else
 						ThrowUnrecognizedElement (child);
 					ValidateInvalidAttributes (child.Attributes, child);
@@ -279,7 +266,7 @@ namespace System.Diagnostics
 					throw new ConfigurationException ("The `autoflush' attribute must be `true' or `false'",
 							e, node);
 				}
-				TraceImpl.AutoFlush = autoflush;
+				configValues.AutoFlush = autoflush;
 			}
 			if (indentsizeConf != null) {
 				int indentsize = 0;
@@ -290,16 +277,15 @@ namespace System.Diagnostics
 					throw new ConfigurationException ("The `indentsize' attribute must be an integral value.",
 							e, node);
 				}
-				TraceImpl.IndentSize = indentsize;
+				configValues.IndentSize = indentsize;
 			}
 		}
 
-#if NET_2_0
 		private TraceListenerCollection GetSharedListeners (IDictionary d)
 		{
 			TraceListenerCollection shared_listeners = d ["sharedListeners"] as TraceListenerCollection;
 			if (shared_listeners == null) {
-				shared_listeners = new TraceListenerCollection ();
+				shared_listeners = new TraceListenerCollection (false);
 				d ["sharedListeners"] = shared_listeners;
 			}
 			return shared_listeners;
@@ -356,7 +342,7 @@ namespace System.Diagnostics
 			if (sources.ContainsKey (name))
 				return;
 
-			TraceSourceInfo sinfo = new TraceSourceInfo (name, levels);
+			TraceSourceInfo sinfo = new TraceSourceInfo (name, levels, configValues);
 			sources.Add (sinfo.Name, sinfo);
 			
 			foreach (XmlNode child in node.ChildNodes) {
@@ -374,13 +360,11 @@ namespace System.Diagnostics
 					ThrowUnrecognizedNode (child);
 			}
 		}
-#endif
 
 		// only defines "add" and "remove", but "clear" also works
 		// for add, "name" is required; initializeData is optional; "type" is required in 1.x, optional in 2.0.
 		private void AddTraceListeners (IDictionary d, XmlNode listenersNode, TraceListenerCollection listeners)
 		{
-#if !TARGET_JVM
 			// There are no attributes on <listeners/>
 			ValidateInvalidAttributes (listenersNode.Attributes, listenersNode);
 
@@ -400,7 +384,7 @@ namespace System.Diagnostics
 							RemoveTraceListener (name);
 							break;
 						case "clear":
-							TraceImpl.Listeners.Clear ();
+							configValues.Listeners.Clear ();
 							break;
 						default:
 							ThrowUnrecognizedElement (child);
@@ -411,7 +395,6 @@ namespace System.Diagnostics
 				else
 					ThrowUnrecognizedNode (child);
 			}
-#endif
 		}
 
 		private void AddTraceListener (IDictionary d, XmlNode child, XmlAttributeCollection attributes, TraceListenerCollection listeners)
@@ -419,7 +402,7 @@ namespace System.Diagnostics
 			string name = GetAttribute (attributes, "name", true, child);
 			string type = null;
 
-#if NET_2_0 && CONFIGURATION_DEP
+#if CONFIGURATION_DEP
 			type = GetAttribute (attributes, "type", false, child);
 			if (type == null) {
 				// indicated by name.
@@ -431,7 +414,7 @@ namespace System.Diagnostics
 						"Listener '{0}' references a shared " +
 						"listener and can only have a 'Name' " +
 						"attribute.", name));
-				listeners.Add (shared);
+				listeners.Add (shared, configValues);
 				return;
 			}
 #else
@@ -454,14 +437,18 @@ namespace System.Diagnostics
 				types = Type.EmptyTypes;
 			}
 
-			System.Reflection.ConstructorInfo ctor = t.GetConstructor (types);
+			BindingFlags flags = BindingFlags.Public | BindingFlags.Instance;
+			if (t.Assembly == GetType ().Assembly)
+				flags |= BindingFlags.NonPublic;
+
+			ConstructorInfo ctor = t.GetConstructor (flags, null, types, null);
 			if (ctor == null) 
 				throw new ConfigurationException ("Couldn't find constructor for class " + type);
 			
 			TraceListener l = (TraceListener) ctor.Invoke (args);
 			l.Name = name;
 
-#if NET_2_0 && CONFIGURATION_DEP
+#if CONFIGURATION_DEP
 			string trace = GetAttribute (attributes, "traceOutputOptions", false, child);
 			if (trace != null) {
 				if (trace != trace.Trim ())
@@ -494,13 +481,13 @@ namespace System.Diagnostics
 			}
 #endif
 
-			listeners.Add (l);
+			listeners.Add (l, configValues);
 		}
 
 		private void RemoveTraceListener (string name)
 		{
 			try {
-				TraceImpl.Listeners.Remove (name);
+				configValues.Listeners.Remove (name);
 			}
 			catch (ArgumentException) {
 				// The specified listener wasn't in the collection
