@@ -11,7 +11,6 @@
 //    http://www.myelin.co.nz
 // (c) 2004-2011 Novell, Inc. (http://www.novell.com)
 //
-
 //
 // Permission is hereby granted, free of charge, to any person obtaining
 // a copy of this software and associated documentation files (the
@@ -45,9 +44,7 @@ using System.IO;
 using System.Net.Configuration;
 using System.Text;
 using System.Timers;
-#if !MOONLIGHT
 using System.Net.NetworkInformation;
-#endif
 
 namespace System.Net.Sockets 
 {
@@ -55,6 +52,9 @@ namespace System.Net.Sockets
 	{
 		private bool islistening;
 		private bool useoverlappedIO;
+		private const int SOCKET_CLOSED = 10004;
+
+		private static readonly string timeout_exc_msg = "A connection attempt failed because the connected party did not properly respond after a period of time, or established connection failed because connected host has failed to respond";
 
 		static void AddSockets (List<Socket> sockets, IList list, string name)
 		{
@@ -68,12 +68,10 @@ namespace System.Net.Sockets
 
 			sockets.Add (null);
 		}
-#if !TARGET_JVM
 		[MethodImplAttribute(MethodImplOptions.InternalCall)]
 		private extern static void Select_internal (ref Socket [] sockets,
 							int microSeconds,
 							out int error);
-#endif
 		public static void Select (IList checkRead, IList checkWrite, IList checkError, int microSeconds)
 		{
 			var list = new List<Socket> ();
@@ -200,11 +198,9 @@ namespace System.Net.Sockets
 		}
 #endif
 	
-#if !TARGET_JVM
 		// Returns the amount of data waiting to be read on socket
 		[MethodImplAttribute(MethodImplOptions.InternalCall)]
 		private extern static int Available_internal(IntPtr socket, out int error);
-#endif
 
 		public int Available {
 			get {
@@ -404,11 +400,9 @@ namespace System.Net.Sockets
 			}
 		}
 
-#if !TARGET_JVM
 		// Returns the local endpoint details in addr and port
 		[MethodImplAttribute(MethodImplOptions.InternalCall)]
 		private extern static SocketAddress LocalEndPoint_internal(IntPtr socket, int family, out int error);
-#endif
 
 		// Wish:  support non-IP endpoints.
 		public EndPoint LocalEndPoint {
@@ -498,7 +492,6 @@ namespace System.Net.Sockets
 			}
 		}
 
-#if !MOONLIGHT
 		public bool AcceptAsync (SocketAsyncEventArgs e)
 		{
 			// NO check is made whether e != null in MS.NET (NRE is thrown in such case)
@@ -523,10 +516,15 @@ namespace System.Net.Sockets
 			e.curSocket = this;
 			Worker w = e.Worker;
 			w.Init (this, e, SocketOperation.Accept);
-			socket_pool_queue (Worker.Dispatcher, w.result);
+			int count;
+			lock (readQ) {
+				readQ.Enqueue (e.Worker);
+				count = readQ.Count;
+			}
+			if (count == 1)
+				socket_pool_queue (Worker.Dispatcher, w.result);
 			return true;
 		}
-#endif
 		// Creates a new system socket, returning the handle
 		[MethodImplAttribute(MethodImplOptions.InternalCall)]
 		private extern static IntPtr Accept_internal(IntPtr sock, out int error, bool blocking);
@@ -537,21 +535,19 @@ namespace System.Net.Sockets
 
 			int error = 0;
 			IntPtr sock = (IntPtr) (-1);
-			blocking_thread = Thread.CurrentThread;
 			try {
+				RegisterForBlockingSyscall ();
 				sock = Accept_internal(socket, out error, blocking);
-			} catch (ThreadAbortException) {
-				if (disposed) {
-					Thread.ResetAbort ();
-					error = (int) SocketError.Interrupted;
-				}
 			} finally {
-				blocking_thread = null;
+				UnRegisterForBlockingSyscall ();
+			}
+
+			if (error != 0) {
+				if (closed)
+					error = SOCKET_CLOSED;
+				throw new SocketException(error);
 			}
 
-			if (error != 0)
-				throw new SocketException (error);
-			
 			Socket accepted = new Socket(this.AddressFamily, this.SocketType,
 				this.ProtocolType, sock);
 
@@ -567,21 +563,19 @@ namespace System.Net.Sockets
 			
 			int error = 0;
 			IntPtr sock = (IntPtr)(-1);
-			blocking_thread = Thread.CurrentThread;
 			
 			try {
+				RegisterForBlockingSyscall ();
 				sock = Accept_internal (socket, out error, blocking);
-			} catch (ThreadAbortException) {
-				if (disposed) {
-					Thread.ResetAbort ();
-					error = (int)SocketError.Interrupted;
-				}
 			} finally {
-				blocking_thread = null;
+				UnRegisterForBlockingSyscall ();
 			}
 			
-			if (error != 0)
+			if (error != 0) {
+				if (closed)
+					error = SOCKET_CLOSED;
 				throw new SocketException (error);
+			}
 			
 			acceptSocket.address_family = this.AddressFamily;
 			acceptSocket.socket_type = this.SocketType;
@@ -605,7 +599,13 @@ namespace System.Net.Sockets
 				throw new InvalidOperationException ();
 
 			SocketAsyncResult req = new SocketAsyncResult (this, state, callback, SocketOperation.Accept);
-			socket_pool_queue (Worker.Dispatcher, req);
+			int count;
+			lock (readQ) {
+				readQ.Enqueue (req.Worker);
+				count = readQ.Count;
+			}
+			if (count == 1)
+				socket_pool_queue (Worker.Dispatcher, req);
 			return req;
 		}
 
@@ -624,7 +624,13 @@ namespace System.Net.Sockets
 			req.Offset = 0;
 			req.Size = receiveSize;
 			req.SockFlags = SocketFlags.None;
-			socket_pool_queue (Worker.Dispatcher, req);
+			int count;
+			lock (readQ) {
+				readQ.Enqueue (req.Worker);
+				count = readQ.Count;
+			}
+			if (count == 1)
+				socket_pool_queue (Worker.Dispatcher, req);
 			return req;
 		}
 
@@ -661,7 +667,13 @@ namespace System.Net.Sockets
 			req.Size = receiveSize;
 			req.SockFlags = SocketFlags.None;
 			req.AcceptSocket = acceptSocket;
-			socket_pool_queue (Worker.Dispatcher, req);
+			int count;
+			lock (readQ) {
+				readQ.Enqueue (req.Worker);
+				count = readQ.Count;
+			}
+			if (count == 1)
+				socket_pool_queue (Worker.Dispatcher, req);
 			return(req);
 		}
 
@@ -723,6 +735,21 @@ namespace System.Net.Sockets
 			socket_pool_queue (Worker.Dispatcher, req);
 			return(req);
 		}
+
+		void CheckRange (byte[] buffer, int offset, int size)
+		{
+			if (offset < 0)
+				throw new ArgumentOutOfRangeException ("offset", "offset must be >= 0");
+				
+			if (offset > buffer.Length)
+				throw new ArgumentOutOfRangeException ("offset", "offset must be <= buffer.Length");
+
+			if (size < 0)                          
+				throw new ArgumentOutOfRangeException ("size", "size must be >= 0");
+				
+			if (size > buffer.Length - offset)
+				throw new ArgumentOutOfRangeException ("size", "size must be <= buffer.Length - offset");
+		}
 		
 		public IAsyncResult BeginReceive(byte[] buffer, int offset,
 						 int size,
@@ -736,11 +763,7 @@ namespace System.Net.Sockets
 			if (buffer == null)
 				throw new ArgumentNullException ("buffer");
 
-			if (offset < 0 || offset > buffer.Length)
-				throw new ArgumentOutOfRangeException ("offset");
-
-			if (size < 0 || offset + size > buffer.Length)
-				throw new ArgumentOutOfRangeException ("size");
+			CheckRange (buffer, offset, size);
 
 			SocketAsyncResult req = new SocketAsyncResult (this, state, callback, SocketOperation.Receive);
 			req.Buffer = buffer;
@@ -828,14 +851,7 @@ namespace System.Net.Sockets
 			if (remote_end == null)
 				throw new ArgumentNullException ("remote_end");
 
-			if (offset < 0)
-				throw new ArgumentOutOfRangeException ("offset", "offset must be >= 0");
-
-			if (size < 0)
-				throw new ArgumentOutOfRangeException ("size", "size must be >= 0");
-
-			if (offset + size > buffer.Length)
-				throw new ArgumentOutOfRangeException ("offset, size", "offset + size exceeds the buffer length");
+			CheckRange (buffer, offset, size);
 
 			SocketAsyncResult req = new SocketAsyncResult (this, state, callback, SocketOperation.ReceiveFrom);
 			req.Buffer = buffer;
@@ -868,11 +884,7 @@ namespace System.Net.Sockets
 			if (remoteEP == null)
 				throw new ArgumentNullException ("remoteEP");
 
-			if (offset < 0 || offset > buffer.Length)
-				throw new ArgumentOutOfRangeException ("offset");
-
-			if (size < 0 || offset + size > buffer.Length)
-				throw new ArgumentOutOfRangeException ("size");
+			CheckRange (buffer, offset, size);
 
 			throw new NotImplementedException ();
 		}
@@ -886,14 +898,7 @@ namespace System.Net.Sockets
 			if (buffer == null)
 				throw new ArgumentNullException ("buffer");
 
-			if (offset < 0)
-				throw new ArgumentOutOfRangeException ("offset", "offset must be >= 0");
-
-			if (size < 0)
-				throw new ArgumentOutOfRangeException ("size", "size must be >= 0");
-
-			if (offset + size > buffer.Length)
-				throw new ArgumentOutOfRangeException ("offset, size", "offset + size exceeds the buffer length");
+			CheckRange (buffer, offset, size);
 
 			if (!connected)
 				throw new SocketException ((int)SocketError.NotConnected);
@@ -1044,7 +1049,10 @@ namespace System.Net.Sockets
 				throw new FileNotFoundException ();
 
 			SendFileHandler d = new SendFileHandler (SendFile);
-			return new SendFileAsyncResult (d, d.BeginInvoke (fileName, preBuffer, postBuffer, flags, callback, state));
+			return new SendFileAsyncResult (d, d.BeginInvoke (fileName, preBuffer, postBuffer, flags, ar => {
+				SendFileAsyncResult sfar = new SendFileAsyncResult (d, ar);
+				callback (sfar);
+			}, state));
 		}
 
 		public IAsyncResult BeginSendTo(byte[] buffer, int offset,
@@ -1059,14 +1067,7 @@ namespace System.Net.Sockets
 			if (buffer == null)
 				throw new ArgumentNullException ("buffer");
 
-			if (offset < 0)
-				throw new ArgumentOutOfRangeException ("offset", "offset must be >= 0");
-
-			if (size < 0)
-				throw new ArgumentOutOfRangeException ("size", "size must be >= 0");
-
-			if (offset + size > buffer.Length)
-				throw new ArgumentOutOfRangeException ("offset, size", "offset + size exceeds the buffer length");
+			CheckRange (buffer, offset, size);
 
 			SocketAsyncResult req = new SocketAsyncResult (this, state, callback, SocketOperation.SendTo);
 			req.Buffer = buffer;
@@ -1166,7 +1167,6 @@ namespace System.Net.Sockets
 			Connect (addresses, port);
 		}
 
-#if !MOONLIGHT
 		public bool DisconnectAsync (SocketAsyncEventArgs e)
 		{
 			// NO check is made whether e != null in MS.NET (NRE is thrown in such case)
@@ -1178,7 +1178,6 @@ namespace System.Net.Sockets
 			socket_pool_queue (Worker.Dispatcher, e.Worker.result);
 			return true;
 		}
-#endif
 
 		[MethodImplAttribute(MethodImplOptions.InternalCall)]
 		extern static void Disconnect_internal(IntPtr sock, bool reuse, out int error);
@@ -1289,7 +1288,6 @@ namespace System.Net.Sockets
 			req.CheckIfThrowDelayedException();
 		}
 
-#if !MOONLIGHT
 		public void EndDisconnect (IAsyncResult asyncResult)
 		{
 			if (disposed && closed)
@@ -1309,7 +1307,6 @@ namespace System.Net.Sockets
 
 			req.CheckIfThrowDelayedException ();
 		}
-#endif
 
 		[MonoTODO]
 		public int EndReceiveMessageFrom (IAsyncResult asyncResult,
@@ -1350,7 +1347,6 @@ namespace System.Net.Sockets
 			ares.Delegate.EndInvoke (ares.Original);
 		}
 
-#if !MOONLIGHT
 		public int EndSendTo (IAsyncResult result)
 		{
 			if (disposed && closed)
@@ -1371,7 +1367,6 @@ namespace System.Net.Sockets
 			req.CheckIfThrowDelayedException();
 			return req.Total;
 		}
-#endif
 
 		[MethodImplAttribute(MethodImplOptions.InternalCall)]
 		private extern static void GetSocketOption_arr_internal(IntPtr socket,
@@ -1414,7 +1409,8 @@ namespace System.Net.Sockets
 		// See Socket.IOControl, WSAIoctl documentation in MSDN. The
 		// common options between UNIX and Winsock are FIONREAD,
 		// FIONBIO and SIOCATMARK. Anything else will depend on the
-		// system.
+		// system except SIO_KEEPALIVE_VALS which is properly handled
+		// on both windows and linux.
 		[MethodImplAttribute(MethodImplOptions.InternalCall)]
 		extern static int WSAIoctl (IntPtr sock, int ioctl_code, byte [] input,
 			byte [] output, out int error);
@@ -1437,13 +1433,9 @@ namespace System.Net.Sockets
 			return result;
 		}
 
-		[MonoTODO]
 		public int IOControl (IOControlCode ioControlCode, byte[] optionInValue, byte[] optionOutValue)
 		{
-			/* Probably just needs to mirror the int
-			 * overload, but more investigation needed.
-			 */
-			throw new NotImplementedException ();
+			return IOControl ((int) ioControlCode, optionInValue, optionOutValue);
 		}
 
 		[MethodImplAttribute(MethodImplOptions.InternalCall)]
@@ -1497,20 +1489,7 @@ namespace System.Net.Sockets
 
 		public int Receive (byte [] buffer)
 		{
-			if (disposed && closed)
-				throw new ObjectDisposedException (GetType ().ToString ());
-
-			if (buffer == null)
-				throw new ArgumentNullException ("buffer");
-
-			SocketError error;
-
-			int ret = Receive_nochecks (buffer, 0, buffer.Length, SocketFlags.None, out error);
-			
-			if (error != SocketError.Success)
-				throw new SocketException ((int) error);
-
-			return ret;
+			return Receive (buffer, SocketFlags.None);
 		}
 
 		public int Receive (byte [] buffer, SocketFlags flags)
@@ -1527,7 +1506,7 @@ namespace System.Net.Sockets
 			
 			if (error != SocketError.Success) {
 				if (error == SocketError.WouldBlock && blocking) // This might happen when ReceiveTimeout is set
-					throw new SocketException ((int) error, "Operation timed out.");
+					throw new SocketException ((int) error, timeout_exc_msg);
 				throw new SocketException ((int) error);
 			}
 
@@ -1542,8 +1521,7 @@ namespace System.Net.Sockets
 			if (buffer == null)
 				throw new ArgumentNullException ("buffer");
 
-			if (size < 0 || size > buffer.Length)
-				throw new ArgumentOutOfRangeException ("size");
+			CheckRange (buffer, 0, size);
 
 			SocketError error;
 
@@ -1551,7 +1529,7 @@ namespace System.Net.Sockets
 			
 			if (error != SocketError.Success) {
 				if (error == SocketError.WouldBlock && blocking) // This might happen when ReceiveTimeout is set
-					throw new SocketException ((int) error, "Operation timed out.");
+					throw new SocketException ((int) error, timeout_exc_msg);
 				throw new SocketException ((int) error);
 			}
 
@@ -1566,11 +1544,7 @@ namespace System.Net.Sockets
 			if (buffer == null)
 				throw new ArgumentNullException ("buffer");
 
-			if (offset < 0 || offset > buffer.Length)
-				throw new ArgumentOutOfRangeException ("offset");
-
-			if (size < 0 || offset + size > buffer.Length)
-				throw new ArgumentOutOfRangeException ("size");
+			CheckRange (buffer, offset, size);
 			
 			SocketError error;
 
@@ -1578,7 +1552,7 @@ namespace System.Net.Sockets
 			
 			if (error != SocketError.Success) {
 				if (error == SocketError.WouldBlock && blocking) // This might happen when ReceiveTimeout is set
-					throw new SocketException ((int) error, "Operation timed out.");
+					throw new SocketException ((int) error, timeout_exc_msg);
 				throw new SocketException ((int) error);
 			}
 
@@ -1593,16 +1567,11 @@ namespace System.Net.Sockets
 			if (buffer == null)
 				throw new ArgumentNullException ("buffer");
 
-			if (offset < 0 || offset > buffer.Length)
-				throw new ArgumentOutOfRangeException ("offset");
-
-			if (size < 0 || offset + size > buffer.Length)
-				throw new ArgumentOutOfRangeException ("size");
+			CheckRange (buffer, offset, size);
 			
 			return Receive_nochecks (buffer, offset, size, flags, out error);
 		}
 
-#if !MOONLIGHT
 		public bool ReceiveFromAsync (SocketAsyncEventArgs e)
 		{
 			if (disposed && closed)
@@ -1622,17 +1591,15 @@ namespace System.Net.Sockets
 			res.Size = e.Count;
 			res.EndPoint = e.RemoteEndPoint;
 			res.SockFlags = e.SocketFlags;
-			Worker worker = new Worker (e);
 			int count;
 			lock (readQ) {
-				readQ.Enqueue (worker);
+				readQ.Enqueue (e.Worker);
 				count = readQ.Count;
 			}
 			if (count == 1)
 				socket_pool_queue (Worker.Dispatcher, res);
 			return true;
 		}
-#endif
 
 		public int ReceiveFrom (byte [] buffer, ref EndPoint remoteEP)
 		{
@@ -1701,11 +1668,7 @@ namespace System.Net.Sockets
 			if (remoteEP == null)
 				throw new ArgumentNullException ("remoteEP");
 
-			if (offset < 0 || offset > buffer.Length)
-				throw new ArgumentOutOfRangeException ("offset");
-
-			if (size < 0 || offset + size > buffer.Length)
-				throw new ArgumentOutOfRangeException ("size");
+			CheckRange (buffer, offset, size);
 
 			return ReceiveFrom_nochecks (buffer, offset, size, flags, ref remoteEP);
 		}
@@ -1728,7 +1691,7 @@ namespace System.Net.Sockets
 					connected = false;
 				else if (err == SocketError.WouldBlock && blocking) { // This might happen when ReceiveTimeout is set
 					if (throwOnError)	
-						throw new SocketException ((int) SocketError.TimedOut, "Operation timed out");
+						throw new SocketException ((int) SocketError.TimedOut, timeout_exc_msg);
 					error = (int) SocketError.TimedOut;
 					return 0;
 				}
@@ -1783,11 +1746,7 @@ namespace System.Net.Sockets
 			if (remoteEP == null)
 				throw new ArgumentNullException ("remoteEP");
 
-			if (offset < 0 || offset > buffer.Length)
-				throw new ArgumentOutOfRangeException ("offset");
-
-			if (size < 0 || offset + size > buffer.Length)
-				throw new ArgumentOutOfRangeException ("size");
+			CheckRange (buffer, offset, size);
 
 			/* FIXME: figure out how we get hold of the
 			 * IPPacketInformation
@@ -1850,8 +1809,7 @@ namespace System.Net.Sockets
 			if (buf == null)
 				throw new ArgumentNullException ("buf");
 
-			if (size < 0 || size > buf.Length)
-				throw new ArgumentOutOfRangeException ("size");
+			CheckRange (buf, 0, size);
 
 			SocketError error;
 
@@ -1871,11 +1829,7 @@ namespace System.Net.Sockets
 			if (buf == null)
 				throw new ArgumentNullException ("buffer");
 
-			if (offset < 0 || offset > buf.Length)
-				throw new ArgumentOutOfRangeException ("offset");
-
-			if (size < 0 || offset + size > buf.Length)
-				throw new ArgumentOutOfRangeException ("size");
+			CheckRange (buf, offset, size);
 
 			SocketError error;
 
@@ -1895,11 +1849,7 @@ namespace System.Net.Sockets
 			if (buf == null)
 				throw new ArgumentNullException ("buffer");
 
-			if (offset < 0 || offset > buf.Length)
-				throw new ArgumentOutOfRangeException ("offset");
-
-			if (size < 0 || offset + size > buf.Length)
-				throw new ArgumentOutOfRangeException ("size");
+			CheckRange (buf, offset, size);
 
 			return Send_nochecks (buf, offset, size, flags, out error);
 		}
@@ -1940,7 +1890,6 @@ namespace System.Net.Sockets
 			}
 		}
 
-#if !MOONLIGHT
 		public bool SendToAsync (SocketAsyncEventArgs e)
 		{
 			// NO check is made whether e != null in MS.NET (NRE is thrown in such case)
@@ -1960,17 +1909,15 @@ namespace System.Net.Sockets
 			res.Size = e.Count;
 			res.SockFlags = e.SocketFlags;
 			res.EndPoint = e.RemoteEndPoint;
-			Worker worker = new Worker (e);
 			int count;
 			lock (writeQ) {
-				writeQ.Enqueue (worker);
+				writeQ.Enqueue (e.Worker);
 				count = writeQ.Count;
 			}
 			if (count == 1)
 				socket_pool_queue (Worker.Dispatcher, res);
 			return true;
 		}
-#endif
 		
 		public int SendTo (byte [] buffer, EndPoint remote_end)
 		{
@@ -2011,8 +1958,7 @@ namespace System.Net.Sockets
 			if (remote_end == null)
 				throw new ArgumentNullException ("remote_end");
 
-			if (size < 0 || size > buffer.Length)
-				throw new ArgumentOutOfRangeException ("size");
+			CheckRange (buffer, 0, size);
 
 			return SendTo_nochecks (buffer, 0, size, flags, remote_end);
 		}
@@ -2038,11 +1984,7 @@ namespace System.Net.Sockets
 			if (remote_end == null)
 				throw new ArgumentNullException("remote_end");
 
-			if (offset < 0 || offset > buffer.Length)
-				throw new ArgumentOutOfRangeException ("offset");
-
-			if (size < 0 || offset + size > buffer.Length)
-				throw new ArgumentOutOfRangeException ("size");
+			CheckRange (buffer, offset, size);
 
 			return SendTo_nochecks (buffer, offset, size, flags, remote_end);
 		}
