@@ -48,9 +48,6 @@ using System.Text;
 using System.Net.Configuration;
 using System.Net.NetworkInformation;
 #endif
-#if MOONLIGHT && !INSIDE_SYSTEM
-using System.Net.Policy;
-#endif
 
 namespace System.Net.Sockets {
 
@@ -92,7 +89,7 @@ namespace System.Net.Sockets {
 			public Socket Sock;
 			public IntPtr handle;
 			object state;
-			AsyncCallback callback;
+			AsyncCallback callback; // used from the runtime
 			WaitHandle waithandle;
 
 			Exception delayedException;
@@ -174,11 +171,7 @@ namespace System.Net.Sockets {
 			{
 				if (callback == null)
 					return;
-#if MOONLIGHT
-				ThreadPool.QueueUserWorkItem (_ => { callback (this); }, null);
-#else
 				ThreadPool.UnsafeQueueUserWorkItem (_ => { callback (this); }, null);
-#endif
 			}
 
 			public void Dispose ()
@@ -245,7 +238,8 @@ namespace System.Net.Sockets {
 				Queue queue = null;
 				if (operation == SocketOperation.Receive ||
 				    operation == SocketOperation.ReceiveFrom ||
-				    operation == SocketOperation.ReceiveGeneric) {
+				    operation == SocketOperation.ReceiveGeneric ||
+				    operation == SocketOperation.Accept) {
 					queue = Sock.readQ;
 				} else if (operation == SocketOperation.Send ||
 					   operation == SocketOperation.SendTo ||
@@ -258,7 +252,7 @@ namespace System.Net.Sockets {
 					Worker worker = null;
 					SocketAsyncCall sac = null;
 					lock (queue) {
-						// queue.Count will only be 0 if the socket is closed while receive/send
+						// queue.Count will only be 0 if the socket is closed while receive/send/accept
 						// operation(s) are pending and at least one call to this method is
 						// waiting on the lock while another one calls CompleteAllOnDispose()
 						if (queue.Count > 0)
@@ -419,15 +413,12 @@ namespace System.Net.Sockets {
 				else if (op == Socket.SocketOperation.Send || op == Socket.SocketOperation.SendGeneric ||
 					op == Socket.SocketOperation.SendJustCallback)
 					sar.Worker.Send ();
-#if !MOONLIGHT
 				else if (op == Socket.SocketOperation.ReceiveFrom)
 					sar.Worker.ReceiveFrom ();
 				else if (op == Socket.SocketOperation.SendTo)
 					sar.Worker.SendTo ();
-#endif
 				else if (op == Socket.SocketOperation.Connect)
 					sar.Worker.Connect ();
-#if !MOONLIGHT
 				else if (op == Socket.SocketOperation.Accept)
 					sar.Worker.Accept ();
 				else if (op == Socket.SocketOperation.AcceptReceive)
@@ -442,7 +433,6 @@ namespace System.Net.Sockets {
 				else if (op == Socket.SocketOperation.SendPackets)
 					async_op = SocketAsyncOperation.SendPackets;
 				*/
-#endif
 				else
 					throw new NotImplementedException (String.Format ("Operation {0} is not implemented", op));
 			}
@@ -459,32 +449,26 @@ namespace System.Net.Sockets {
 				//	-SendPackets and ReceiveMessageFrom are not implemented yet
 				if (op == Socket.SocketOperation.Connect)
 					async_op = SocketAsyncOperation.Connect;
-#if !MOONLIGHT
 				else if (op == Socket.SocketOperation.Accept)
 					async_op = SocketAsyncOperation.Accept;
 				else if (op == Socket.SocketOperation.Disconnect)
 					async_op = SocketAsyncOperation.Disconnect;
-#endif
 				else if (op == Socket.SocketOperation.Receive || op == Socket.SocketOperation.ReceiveGeneric)
 					async_op = SocketAsyncOperation.Receive;
-#if !MOONLIGHT
 				else if (op == Socket.SocketOperation.ReceiveFrom)
 					async_op = SocketAsyncOperation.ReceiveFrom;
-#endif
 				/*
 				else if (op == Socket.SocketOperation.ReceiveMessageFrom)
 					async_op = SocketAsyncOperation.ReceiveMessageFrom;
 				*/
 				else if (op == Socket.SocketOperation.Send || op == Socket.SocketOperation.SendGeneric)
 					async_op = SocketAsyncOperation.Send;
-#if !MOONLIGHT
 				/*
 				else if (op == Socket.SocketOperation.SendPackets)
 					async_op = SocketAsyncOperation.SendPackets;
 				*/
 				else if (op == Socket.SocketOperation.SendTo)
 					async_op = SocketAsyncOperation.SendTo;
-#endif
 				else
 					throw new NotImplementedException (String.Format ("Operation {0} is not implemented", op));
 
@@ -495,7 +479,6 @@ namespace System.Net.Sockets {
 
 			public void Accept ()
 			{
-#if !MOONLIGHT
 				Socket acc_socket = null;
 				try {
 					if (args != null && args.AcceptSocket != null) {
@@ -512,7 +495,6 @@ namespace System.Net.Sockets {
 				}
 
 				result.Complete (acc_socket);
-#endif
 			}
 
 			/* only used in 2.0 profile and newer, but
@@ -521,7 +503,6 @@ namespace System.Net.Sockets {
 			 */
 			public void AcceptReceive ()
 			{
-#if !MOONLIGHT
 				Socket acc_socket = null;
 				try {
 					if (result.AcceptSocket == null) {
@@ -559,7 +540,6 @@ namespace System.Net.Sockets {
 				}
 
 				result.Complete (acc_socket, total);
-#endif
 			}
 
 			public void Connect ()
@@ -570,16 +550,7 @@ namespace System.Net.Sockets {
 				}
 
 				SocketAsyncResult mconnect = result.AsyncState as SocketAsyncResult;
-#if !MOONLIGHT
 				bool is_mconnect = (mconnect != null && mconnect.Addresses != null);
-#else
-				if (result.ErrorCode == SocketError.AccessDenied) {
-					result.Complete ();
-					result.DoMConnectCallback ();
-					return;
-				}
-				bool is_mconnect = false;
-#endif
 				try {
 					int error_code;
 					EndPoint ep = result.EndPoint;
@@ -625,7 +596,6 @@ namespace System.Net.Sockets {
 			/* Also only used in 2.0 profile and newer */
 			public void Disconnect ()
 			{
-#if !MOONLIGHT
 				try {
 					if (args != null)
 						result.ReuseSocket = args.DisconnectReuseSocket;
@@ -635,7 +605,6 @@ namespace System.Net.Sockets {
 					return;
 				}
 				result.Complete ();
-#endif
 			}
 
 			public void Receive ()
@@ -650,7 +619,6 @@ namespace System.Net.Sockets {
 
 			public void ReceiveFrom ()
 			{
-#if !MOONLIGHT
 				int total = 0;
 				try {
 					total = result.Sock.ReceiveFrom_nochecks (result.Buffer,
@@ -664,7 +632,6 @@ namespace System.Net.Sockets {
 				}
 
 				result.Complete (total);
-#endif
 			}
 
 			public void ReceiveGeneric ()
@@ -709,13 +676,13 @@ namespace System.Net.Sockets {
 						return; // Have to finish writing everything. See bug #74475.
 					}
 					result.Total = send_so_far;
+					send_so_far = 0;
 				}
 				result.Complete ();
 			}
 
 			public void SendTo ()
 			{
-#if !MOONLIGHT
 				int total = 0;
 				try {
 					total = result.Sock.SendTo_nochecks (result.Buffer,
@@ -730,13 +697,14 @@ namespace System.Net.Sockets {
 						return; // Have to finish writing everything. See bug #74475.
 					}
 					result.Total = send_so_far;
+					send_so_far = 0;
 				} catch (Exception e) {
+					send_so_far = 0;
 					result.Complete (e);
 					return;
 				}
 
 				result.Complete ();
-#endif
 			}
 
 			public void SendGeneric ()
@@ -784,16 +752,33 @@ namespace System.Net.Sockets {
 			}
 
 			if (ipv6Supported == -1) {
+				// We need to put a try/catch around ConfigurationManager methods as will always throw an exception 
+				// when run in a mono embedded application.  This occurs as embedded applications do not have a setup
+				// for application config.  The exception is not thrown when called from a normal .NET application. 
+				//
+				// We, then, need to guard calls to the ConfigurationManager.  If the config is not found or throws an
+				// exception, will fall through to the existing Socket / API directly below in the code.
+				//
+				// Also note that catching ConfigurationErrorsException specifically would require library dependency
+				// System.Configuration, and wanted to avoid that.
 #if !NET_2_1
 #if CONFIGURATION_DEP
-				SettingsSection config;
-				config = (SettingsSection) System.Configuration.ConfigurationManager.GetSection ("system.net/settings");
-				if (config != null)
-					ipv6Supported = config.Ipv6.Enabled ? -1 : 0;
+				try {
+					SettingsSection config;
+					config = (SettingsSection) System.Configuration.ConfigurationManager.GetSection ("system.net/settings");
+					if (config != null)
+						ipv6Supported = config.Ipv6.Enabled ? -1 : 0;
+				} catch {
+					ipv6Supported = -1;
+				}
 #else
-				NetConfig config = System.Configuration.ConfigurationSettings.GetConfig("system.net/settings") as NetConfig;
-				if (config != null)
-					ipv6Supported = config.ipv6Enabled ? -1 : 0;
+				try {
+					NetConfig config = System.Configuration.ConfigurationSettings.GetConfig("system.net/settings") as NetConfig;
+					if (config != null)
+						ipv6Supported = config.ipv6Enabled ? -1 : 0;
+				} catch {
+					ipv6Supported = -1;
+				}
 #endif
 #endif
 				if (ipv6Supported != 0) {
@@ -858,7 +843,7 @@ namespace System.Net.Sockets {
 		private SocketType socket_type;
 		private ProtocolType protocol_type;
 		internal bool blocking=true;
-		Thread blocking_thread;
+		List<Thread> blocking_threads;
 		private bool isbound;
 		/* When true, the socket was connected at the time of
 		 * the last IO operation
@@ -878,52 +863,88 @@ namespace System.Net.Sockets {
  		 */
 		internal EndPoint seed_endpoint = null;
 
-#if !TARGET_JVM
+		void RegisterForBlockingSyscall ()
+		{
+			while (blocking_threads == null) {
+				//In the rare event this CAS fail, there's a good chance other thread won, so we're kosher.
+				//In the VERY rare event of all CAS fail together, we pay the full price of of failure.
+				Interlocked.CompareExchange (ref blocking_threads, new List<Thread> (), null);
+			}
+
+			try {
+				
+			} finally {
+				/* We must use a finally block here to make this atomic. */
+				lock (blocking_threads) {
+					blocking_threads.Add (Thread.CurrentThread);
+				}
+			}
+		}
+
+		/* This must be called from a finally block! */
+		void UnRegisterForBlockingSyscall ()
+		{
+			//If this NRE, we're in deep problems because Register Must have
+			lock (blocking_threads) {
+				blocking_threads.Remove (Thread.CurrentThread);
+			}
+		}
+
+		void AbortRegisteredThreads () {
+			if (blocking_threads == null)
+				return;
+
+			lock (blocking_threads) {
+				foreach (var t in blocking_threads)
+					cancel_blocking_socket_operation (t);
+				blocking_threads.Clear ();
+			}
+		}
+
 		// Creates a new system socket, returning the handle
 		[MethodImplAttribute(MethodImplOptions.InternalCall)]
 		private extern IntPtr Socket_internal(AddressFamily family,
 						      SocketType type,
 						      ProtocolType proto,
 						      out int error);
-#endif		
 		
-		public Socket(AddressFamily family, SocketType type, ProtocolType proto)
+		public Socket(AddressFamily addressFamily, SocketType socketType, ProtocolType protocolType)
 		{
 #if NET_2_1 && !MOBILE
-			switch (family) {
+			switch (addressFamily) {
 			case AddressFamily.InterNetwork:	// ok
 			case AddressFamily.InterNetworkV6:	// ok
 			case AddressFamily.Unknown:		// SocketException will be thrown later (with right error #)
 				break;
 			// case AddressFamily.Unspecified:
 			default:
-				throw new ArgumentException ("family");
+				throw new ArgumentException ("addressFamily");
 			}
 
-			switch (type) {
+			switch (socketType) {
 			case SocketType.Stream:			// ok
 			case SocketType.Unknown:		// SocketException will be thrown later (with right error #)
 				break;
 			default:
-				throw new ArgumentException ("type");
+				throw new ArgumentException ("socketType");
 			}
 
-			switch (proto) {
+			switch (protocolType) {
 			case ProtocolType.Tcp:			// ok
 			case ProtocolType.Unspecified:		// ok
 			case ProtocolType.Unknown:		// SocketException will be thrown later (with right error #)
 				break;
 			default:
-				throw new ArgumentException ("proto");
+				throw new ArgumentException ("protocolType");
 			}
 #endif
-			address_family=family;
-			socket_type=type;
-			protocol_type=proto;
+			address_family = addressFamily;
+			socket_type = socketType;
+			protocol_type = protocolType;
 			
 			int error;
 			
-			socket = Socket_internal (family, type, proto, out error);
+			socket = Socket_internal (addressFamily, socketType, protocolType, out error);
 			if (error != 0)
 				throw new SocketException (error);
 #if !NET_2_1 || MOBILE
@@ -941,12 +962,10 @@ namespace System.Net.Sockets {
 			get { return address_family; }
 		}
 
-#if !TARGET_JVM
 		[MethodImplAttribute(MethodImplOptions.InternalCall)]
 		private extern static void Blocking_internal(IntPtr socket,
 							     bool block,
 							     out int error);
-#endif
 
 		public bool Blocking {
 			get {
@@ -1085,10 +1104,6 @@ namespace System.Net.Sockets {
 				if (disposed && closed)
 					throw new ObjectDisposedException (GetType ().ToString ());
 				
-#if MOONLIGHT
-				if (!connected)
-					return seed_endpoint;
-#else
 				/*
 				 * If the seed EndPoint is null, Connect, Bind,
 				 * etc has not yet been called. MS returns null
@@ -1096,7 +1111,6 @@ namespace System.Net.Sockets {
 				 */
 				if (!connected || seed_endpoint == null)
 					return null;
-#endif			
 				SocketAddress sa;
 				int error;
 				
@@ -1137,8 +1151,10 @@ namespace System.Net.Sockets {
 					return; */
 			}
 		}
+		[MethodImplAttribute(MethodImplOptions.InternalCall)]
+		static extern void cancel_blocking_socket_operation (Thread thread);
 
-		protected virtual void Dispose (bool explicitDisposing)
+		protected virtual void Dispose (bool disposing)
 		{
 			if (disposed)
 				return;
@@ -1151,11 +1167,8 @@ namespace System.Net.Sockets {
 				closed = true;
 				IntPtr x = socket;
 				socket = (IntPtr) (-1);
-				Thread th = blocking_thread;
-				if (th != null) {
-					th.Abort ();
-					blocking_thread = null;
-				}
+				
+				AbortRegisteredThreads ();
 
 				if (was_connected)
 					Linger (x);
@@ -1167,7 +1180,7 @@ namespace System.Net.Sockets {
 			}
 		}
 
-#if NET_2_1 || NET_4_0
+#if NET_4_0
 		public void Dispose ()
 #else
 		void IDisposable.Dispose ()
@@ -1210,51 +1223,36 @@ namespace System.Net.Sockets {
 				throw new ArgumentNullException ("remoteEP");
 
 			IPEndPoint ep = remoteEP as IPEndPoint;
-#if !MOONLIGHT
 			if (ep != null && socket_type != SocketType.Dgram) /* Dgram uses Any to 'disconnect' */
-#else
-			if (ep != null)
-#endif
 				if (ep.Address.Equals (IPAddress.Any) || ep.Address.Equals (IPAddress.IPv6Any))
 					throw new SocketException ((int) SocketError.AddressNotAvailable);
 
-#if MOONLIGHT
-			if (protocol_type != ProtocolType.Tcp)
-				throw new SocketException ((int) SocketError.AccessDenied);
-#else
 			if (islistening)
 				throw new InvalidOperationException ();
-#endif
 			serial = remoteEP.Serialize ();
 
 			int error = 0;
 
-			blocking_thread = Thread.CurrentThread;
 			try {
+				RegisterForBlockingSyscall ();
 				Connect_internal (socket, serial, out error);
-			} catch (ThreadAbortException) {
-				if (disposed) {
-					Thread.ResetAbort ();
-					error = (int) SocketError.Interrupted;
-				}
 			} finally {
-				blocking_thread = null;
+				UnRegisterForBlockingSyscall ();
 			}
 
 			if (error == 0 || error == 10035)
 				seed_endpoint = remoteEP; // Keep the ep around for non-blocking sockets
 
-			if (error != 0)
+			if (error != 0) {
+				if (closed)
+					error = SOCKET_CLOSED;
 				throw new SocketException (error);
+			}
 
-#if !MOONLIGHT
-			if (socket_type == SocketType.Dgram && (ep.Address.Equals (IPAddress.Any) || ep.Address.Equals (IPAddress.IPv6Any)))
+			if (socket_type == SocketType.Dgram && ep != null && (ep.Address.Equals (IPAddress.Any) || ep.Address.Equals (IPAddress.IPv6Any)))
 				connected = false;
 			else
 				connected = true;
-#else
-			connected = true;
-#endif
 			isbound = true;
 		}
 
@@ -1459,9 +1457,7 @@ namespace System.Net.Sockets {
 #endif
 		}
 
-#if !MOONLIGHT
 		public
-#endif
 		IAsyncResult BeginConnect(EndPoint end_point, AsyncCallback callback, object state)
 		{
 			if (disposed && closed)
@@ -1524,11 +1520,7 @@ namespace System.Net.Sockets {
 			return req;
 		}
 
-#if !MOONLIGHT
 		public
-#else
-		internal
-#endif
 		IAsyncResult BeginConnect (IPAddress[] addresses, int port, AsyncCallback callback, object state)
 
 		{
@@ -1547,10 +1539,8 @@ namespace System.Net.Sockets {
 
 			if (port <= 0 || port > 65535)
 				throw new ArgumentOutOfRangeException ("port", "Must be > 0 and < 65536");
-#if !MOONLIGHT
 			if (islistening)
 				throw new InvalidOperationException ();
-#endif
 
 			SocketAsyncResult req = new SocketAsyncResult (this, state, callback, SocketOperation.Connect);
 			req.Addresses = addresses;
@@ -1591,79 +1581,41 @@ namespace System.Net.Sockets {
 		bool GetCheckedIPs (SocketAsyncEventArgs e, out IPAddress [] addresses)
 		{
 			addresses = null;
-#if MOONLIGHT || NET_4_0
+#if NET_4_0
 			// Connect to the first address that match the host name, like:
 			// http://blogs.msdn.com/ncl/archive/2009/07/20/new-ncl-features-in-net-4-0-beta-2.aspx
 			// while skipping entries that do not match the address family
 			DnsEndPoint dep = (e.RemoteEndPoint as DnsEndPoint);
 			if (dep != null) {
 				addresses = Dns.GetHostAddresses (dep.Host);
-				IPEndPoint endpoint;
-#if MOONLIGHT && !INSIDE_SYSTEM
-				if (!e.PolicyRestricted && !SecurityManager.HasElevatedPermissions) {
-					List<IPAddress> valid = new List<IPAddress> ();
-					foreach (IPAddress a in addresses) {
-						// if we're not downloading a socket policy then check the policy
-						// and if we're not running with elevated permissions (SL4 OoB option)
-						endpoint = new IPEndPoint (a, dep.Port);
-						if (!CrossDomainPolicyManager.CheckEndPoint (endpoint, e.SocketClientAccessPolicyProtocol))
-							continue;
-						valid.Add (a);
-					}
-					if (valid.Count == 0)
-		 				e.SocketError = SocketError.AccessDenied;
-					addresses = valid.ToArray ();
-				}
-#endif
 				return true;
 			} else {
 				e.ConnectByNameError = null;
-#if MOONLIGHT && !INSIDE_SYSTEM
-				if (!e.PolicyRestricted && !SecurityManager.HasElevatedPermissions) {
-					if (CrossDomainPolicyManager.CheckEndPoint (e.RemoteEndPoint, e.SocketClientAccessPolicyProtocol))
-						return false;
-		 			else
-						e.SocketError = SocketError.AccessDenied;
-				} else
-#endif
 					return false;
 			}
-			return true; // do not use remote endpoint
 #else
 			return false; // < NET_4_0 -> use remote endpoint
 #endif
 		}
 
 		bool ConnectAsyncReal (SocketAsyncEventArgs e)
-		{
-			IPAddress [] addresses = null;
+		{			
 			bool use_remoteep = true;
-#if MOONLIGHT || NET_4_0
+#if NET_4_0
+			IPAddress [] addresses = null;
 			use_remoteep = !GetCheckedIPs (e, out addresses);
-			bool policy_failed = (e.SocketError == SocketError.AccessDenied);
 #endif
 			e.curSocket = this;
 			Worker w = e.Worker;
 			w.Init (this, e, SocketOperation.Connect);
 			SocketAsyncResult result = w.result;
-#if MOONLIGHT
-			if (policy_failed) {
-				// SocketAsyncEventArgs.Completed must be called
-				connected = false;
-				result.EndPoint = e.RemoteEndPoint;
-				result.error = (int) SocketError.AccessDenied;
-				result.Complete ();
-				socket_pool_queue (Worker.Dispatcher, result);
-				return true;
-			}
-#endif
 			IAsyncResult ares = null;
 			try {
 				if (use_remoteep) {
 					result.EndPoint = e.RemoteEndPoint;
 					ares = BeginConnect (e.RemoteEndPoint, SocketAsyncEventArgs.Dispatcher, e);
 				}
-#if MOONLIGHT || NET_4_0
+#if NET_4_0
 				else {
 
 					DnsEndPoint dep = (e.RemoteEndPoint as DnsEndPoint);
@@ -1684,7 +1636,6 @@ namespace System.Net.Sockets {
 			return true;
 		}
 
-#if !MOONLIGHT
 		public bool ConnectAsync (SocketAsyncEventArgs e)
 		{
 			// NO check is made whether e != null in MS.NET (NRE is thrown in such case)
@@ -1697,66 +1648,10 @@ namespace System.Net.Sockets {
 
 			return ConnectAsyncReal (e);
 		}
-#endif
-#if MOONLIGHT
-		static void CheckConnect (SocketAsyncEventArgs e)
-		{
-			// NO check is made whether e != null in MS.NET (NRE is thrown in such case)
 
-			if (e.RemoteEndPoint == null)
-				throw new ArgumentNullException ("remoteEP");
-			if (e.BufferList != null)
-				throw new ArgumentException ("Multiple buffers cannot be used with this method.");
-		}
-
-		public bool ConnectAsync (SocketAsyncEventArgs e)
-		{
-			if (disposed && closed)
-				throw new ObjectDisposedException (GetType ().ToString ());
-
-			CheckConnect (e);
-			// if an address family is specified then they must match
-			AddressFamily raf = e.RemoteEndPoint.AddressFamily;
-			if ((raf != AddressFamily.Unspecified) && (raf != AddressFamily))
-				throw new NotSupportedException ("AddressFamily mismatch between socket and endpoint");
-
-			// connected, not yet connected or even policy denied, the Socket.RemoteEndPoint is always 
-			// available after the ConnectAsync call
-			seed_endpoint = e.RemoteEndPoint;
-			return ConnectAsyncReal (e);
-		}
-
-		public static bool ConnectAsync (SocketType socketType, ProtocolType protocolType, SocketAsyncEventArgs e)
-		{
-			// exception ordering requires to check before creating the socket (good thing resource wise too)
-			CheckConnect (e);
-
-			// create socket based on the endpoint address family (if specified), otherwise default fo IPv4
-			AddressFamily raf = e.RemoteEndPoint.AddressFamily;
-			if (raf == AddressFamily.Unspecified)
-				raf = AddressFamily.InterNetwork;
-			Socket s = new Socket (raf, socketType, protocolType);
-			return s.ConnectAsyncReal (e);
-		}
-
-		public static void CancelConnectAsync (SocketAsyncEventArgs e)
-		{
-			if (e == null)
-				throw new ArgumentNullException ("e");
-
-			// FIXME: this is canceling a synchronous connect, not an async one
-			Socket s = e.ConnectSocket;
-			if ((s != null) && (s.blocking_thread != null))
-				s.blocking_thread.Abort ();
-		}
-#endif
 		[MethodImplAttribute (MethodImplOptions.InternalCall)]
 		private extern static int Receive_internal (IntPtr sock, WSABUF[] bufarray, SocketFlags flags, out int error);
-#if !MOONLIGHT
 		public
-#else
-		internal
-#endif
 		int Receive (IList<ArraySegment<byte>> buffers)
 		{
 			int ret;
@@ -1769,11 +1664,7 @@ namespace System.Net.Sockets {
 		}
 
 		[CLSCompliant (false)]
-#if !MOONLIGHT
 		public
-#else
-		internal
-#endif
 		int Receive (IList<ArraySegment<byte>> buffers, SocketFlags socketFlags)
 		{
 			int ret;
@@ -1786,11 +1677,7 @@ namespace System.Net.Sockets {
 		}
 
 		[CLSCompliant (false)]
-#if !MOONLIGHT
 		public
-#else
-		internal
-#endif
 		int Receive (IList<ArraySegment<byte>> buffers, SocketFlags socketFlags, out SocketError errorCode)
 		{
 			if (disposed && closed)
@@ -1844,11 +1731,7 @@ namespace System.Net.Sockets {
 
 		[MethodImplAttribute (MethodImplOptions.InternalCall)]
 		private extern static int Send_internal (IntPtr sock, WSABUF[] bufarray, SocketFlags flags, out int error);
-#if !MOONLIGHT
 		public
-#else
-		internal
-#endif
 		int Send (IList<ArraySegment<byte>> buffers)
 		{
 			int ret;
@@ -1860,11 +1743,7 @@ namespace System.Net.Sockets {
 			return(ret);
 		}
 
-#if !MOONLIGHT
 		public
-#else
-		internal
-#endif
 		int Send (IList<ArraySegment<byte>> buffers, SocketFlags socketFlags)
 		{
 			int ret;
@@ -1877,11 +1756,7 @@ namespace System.Net.Sockets {
 		}
 
 		[CLSCompliant (false)]
-#if !MOONLIGHT
 		public
-#else
-		internal
-#endif
 		int Send (IList<ArraySegment<byte>> buffers, SocketFlags socketFlags, out SocketError errorCode)
 		{
 			if (disposed && closed)
@@ -1926,11 +1801,7 @@ namespace System.Net.Sockets {
 			return new InvalidOperationException (method + " can only be called once per asynchronous operation");
 		}
 
-#if !MOONLIGHT
 		public
-#else
-		internal
-#endif
 		int EndReceive (IAsyncResult result)
 		{
 			SocketError error;
@@ -1943,11 +1814,7 @@ namespace System.Net.Sockets {
 			return bytesReceived;
 		}
 
-#if !MOONLIGHT
 		public
-#else
-		internal
-#endif
 		int EndReceive (IAsyncResult asyncResult, out SocketError errorCode)
 		{
 			if (disposed && closed)
@@ -1974,11 +1841,7 @@ namespace System.Net.Sockets {
 			return(req.Total);
 		}
 
-#if !MOONLIGHT
 		public
-#else
-		internal
-#endif
 		int EndSend (IAsyncResult result)
 		{
 			SocketError error;
@@ -1991,11 +1854,7 @@ namespace System.Net.Sockets {
 			return bytesSent;
 		}
 
-#if !MOONLIGHT
 		public
-#else
-		internal
-#endif
 		int EndSend (IAsyncResult asyncResult, out SocketError errorCode)
 		{
 			if (disposed && closed)
@@ -2022,11 +1881,7 @@ namespace System.Net.Sockets {
 		}
 
 		// Used by Udpclient
-#if !MOONLIGHT
 		public
-#else
-		internal
-#endif
 		int EndReceiveFrom(IAsyncResult result, ref EndPoint end_point)
 		{
 			if (disposed && closed)
