@@ -25,12 +25,7 @@
 // OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 // WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 //
-#if NET_2_0
-
-
 namespace System.Net
 {
 	public delegate void UploadStringCompletedEventHandler (object sender, UploadStringCompletedEventArgs e);
 }
-
-#endif
