@@ -124,9 +124,7 @@ public class UTF8Encoding : Encoding
 						break;
 					} else {
 						// We have a surrogate tail without 
-						// leading surrogate. In NET_2_0 it
-						// uses fallback. In NET_1_1 we output
-						// wrong surrogate.
+						// leading surrogate.
 						char [] fallback_chars = GetFallbackChars (chars, start, fallback, ref buffer);
 						fixed (char *fb_chars = fallback_chars) {
 							char dummy = '\0';
@@ -292,9 +290,7 @@ public class UTF8Encoding : Encoding
 						break;
 					} else {
 						// We have a surrogate tail without 
-						// leading surrogate. In NET_2_0 it
-						// uses fallback. In NET_1_1 we output
-						// wrong surrogate.
+						// leading surrogate.
 						char [] fallback_chars = GetFallbackChars (chars, start, fallback, ref buffer); 
 						char dummy = '\0';
 						if (bytes + InternalGetByteCount (fallback_chars, 0, fallback_chars.Length, fallback, ref dummy, true) > end_bytes)
@@ -667,7 +663,7 @@ fail_no_space:
 			throw new ArgumentOutOfRangeException ("charIndex", _("ArgRange_Array"));
 		}
 
-		if (charIndex == chars.Length)
+		if (charIndex == chars.Length && byteCount == 0)
 			return 0;
 
 		fixed (char* cptr = chars) {
@@ -692,10 +688,14 @@ fail_no_space:
 		if (leftOverCount == 0) {
 			int end = byteIndex + byteCount;
 			for (; byteIndex < end; posn++, byteIndex++, byteCount--) {
-				if (bytes [byteIndex] < 0x80)
+				if (bytes [byteIndex] < 0x80) {
+					if (posn >= length) {
+						throw new ArgumentException (_("Arg_InsufficientSpace"), "chars");
+					}
 					chars [posn] = (char) bytes [byteIndex];
-				else
+				} else {
 					break;
+				}
 			}
 		}
 
@@ -852,7 +852,14 @@ fail_no_space:
 		if (charCount < 0) {
 			throw new ArgumentOutOfRangeException ("charCount", _("ArgRange_NonNegative"));
 		}
-		return charCount * 4;
+
+		// Add 1 to charCount since there may be a lead surrogate left from the previous call to GetBytes/Encoder.Convert
+		charCount = charCount + 1;
+		if (EncoderFallback.MaxCharCount > 1) {
+			charCount = charCount * EncoderFallback.MaxCharCount;
+		}
+
+		return charCount * 3;
 	}
 
 	// Get the maximum number of characters needed to decode a
@@ -862,19 +869,26 @@ fail_no_space:
 		if (byteCount < 0) {
 			throw new ArgumentOutOfRangeException ("byteCount", _("ArgRange_NonNegative"));
 		}
-		return byteCount;
+
+		// Add 1 to byteCount since there may be the bytes from part of a surrogate pair left from the previous call to GetChars/Decoder.Convert
+		int maxCharCount = byteCount + 1;
+		if (DecoderFallback.MaxCharCount > 1) {
+			maxCharCount = maxCharCount * DecoderFallback.MaxCharCount;
+		}
+
+		return maxCharCount;
 	}
 
 	// Get a UTF8-specific decoder that is attached to this instance.
 	public override Decoder GetDecoder ()
 	{
-		return new UTF8Decoder (DecoderFallback);
+		return new UTF8Decoder (this);
 	}
 
 	// Get a UTF8-specific encoder that is attached to this instance.
 	public override Encoder GetEncoder ()
 	{
-		return new UTF8Encoder (EncoderFallback, emitIdentifier);
+		return new UTF8Encoder (this);
 	}
 
 	// Get the UTF8 preamble.
@@ -883,7 +897,7 @@ fail_no_space:
 		if (emitIdentifier)
 			return new byte [] { 0xEF, 0xBB, 0xBF };
 
-		return new byte [0];
+		return EmptyArray<byte>.Value;
 	}
 
 	// Determine if this object is equal to another.
@@ -921,15 +935,15 @@ fail_no_space:
 
 	// UTF-8 decoder implementation.
 	[Serializable]
-	private class UTF8Decoder : Decoder
+	private class UTF8Decoder : EncodingDecoder
 	{
 		private uint leftOverBits;
 		private uint leftOverCount;
 
 		// Constructor.
-		public UTF8Decoder (DecoderFallback fallback)
+		public UTF8Decoder (Encoding encoding)
+			: base (encoding)
 		{
-			Fallback = fallback;
 			leftOverBits = 0;
 			leftOverCount = 0;
 		}
@@ -955,17 +969,16 @@ fail_no_space:
 
 	// UTF-8 encoder implementation.
 	[Serializable]
-	private class UTF8Encoder : Encoder
+	private class UTF8Encoder : EncodingEncoder
 	{
 //		private bool emitIdentifier;
 		private char leftOverForCount;
 		private char leftOverForConv;
 
 		// Constructor.
-		public UTF8Encoder (EncoderFallback fallback, bool emitIdentifier)
+		public UTF8Encoder (UTF8Encoding encoding)
+			: base (encoding)
 		{
-			Fallback = fallback;
-//			this.emitIdentifier = emitIdentifier;
 			leftOverForCount = '\0';
 			leftOverForConv = '\0';
 		}
