@@ -48,9 +48,9 @@ namespace System.Threading.Tasks
 		static Task current;
 		
 		// parent is the outer task in which this task is created
-		readonly Task parent;
-		// contAncestor is the Task on which this continuation was setup
-		readonly Task contAncestor;
+		Task parent;
+		// A reference to a Task on which this continuation is attached to
+		Task contAncestor;
 		
 		static int          id = -1;
 		static readonly TaskFactory defaultFactory = new TaskFactory ();
@@ -76,6 +76,8 @@ namespace System.Threading.Tasks
 		CancellationToken token;
 		CancellationTokenRegistration? cancellationRegistration;
 
+		ExecutionContext ec;
+
 		internal const TaskCreationOptions WorkerTaskNotSupportedOptions = TaskCreationOptions.LongRunning | TaskCreationOptions.PreferFairness;
 
 		const TaskCreationOptions MaxTaskCreationOptions =
@@ -158,6 +160,8 @@ namespace System.Threading.Tasks
 
 			if (token.CanBeCanceled && !ignoreCancellation)
 				cancellationRegistration = token.Register (l => ((Task) l).CancelReal (), this);
+
+			ec = ExecutionContext.Capture (false, true);
 		}
 
 		static bool HasFlag (TaskCreationOptions opt, TaskCreationOptions member)
@@ -182,8 +186,11 @@ namespace System.Threading.Tasks
 			if (IsContinuation)
 				throw new InvalidOperationException ("Start may not be called on a continuation task");
 
+			if (IsPromise)
+				throw new InvalidOperationException ("Start may not be called on a promise-style task");
+
 			SetupScheduler (scheduler);
-			Schedule ();
+			Schedule (true);
 		}
 
 		internal void SetupScheduler (TaskScheduler scheduler)
@@ -208,10 +215,13 @@ namespace System.Threading.Tasks
 			if (IsContinuation)
 				throw new InvalidOperationException ("RunSynchronously may not be called on a continuation task");
 
-			RunSynchronouslyCore (scheduler);
+			if (IsPromise)
+				throw new InvalidOperationException ("RunSynchronously may not be called on a promise-style task");
+
+			RunSynchronouslyCore (scheduler, true);
 		}
 
-		internal void RunSynchronouslyCore (TaskScheduler scheduler)
+		internal void RunSynchronouslyCore (TaskScheduler scheduler, bool throwException)
 		{
 			SetupScheduler (scheduler);
 			Status = TaskStatus.WaitingToRun;
@@ -220,11 +230,16 @@ namespace System.Threading.Tasks
 				if (scheduler.RunInline (this, false))
 					return;
 			} catch (Exception inner) {
-				throw new TaskSchedulerException (inner);
+				var ex = new TaskSchedulerException (inner);
+				TrySetException (new AggregateException (ex), false, true);
+				if (throwException)
+					throw ex;
+
+				return;
 			}
 
-			Schedule ();
-			Wait ();
+			Schedule (throwException);
+			WaitCore (Timeout.Infinite, CancellationToken.None, false);
 		}
 		#endregion
 		
@@ -330,18 +345,27 @@ namespace System.Threading.Tasks
 			ContinueWith (new TaskContinuation (continuation, options));
 		}
 		
-		internal void ContinueWith (IContinuation continuation)
+		internal bool ContinueWith (IContinuation continuation, bool canExecuteInline = true)
 		{
 			if (IsCompleted) {
+				if (!canExecuteInline)
+					return false;
+
 				continuation.Execute ();
-				return;
+				return true;
 			}
 			
 			continuations.Add (continuation);
 			
 			// Retry in case completion was achieved but event adding was too late
-			if (IsCompleted && continuations.Remove (continuation))
+			if (IsCompleted && continuations.Remove (continuation)) {
+				if (!canExecuteInline)
+					return false;
+
 				continuation.Execute ();
+			}
+
+			return true;
 		}
 
 		internal void RemoveContinuation (IContinuation continuation)
@@ -364,10 +388,17 @@ namespace System.Threading.Tasks
 		#endregion
 		
 		#region Internal and protected thingies
-		internal void Schedule ()
+		internal void Schedule (bool throwException)
 		{
 			Status = TaskStatus.WaitingToRun;
-			scheduler.QueueTask (this);
+			try {
+				scheduler.QueueTask (this);
+			} catch (Exception inner) {
+				var ex = new TaskSchedulerException (inner);
+				TrySetException (new AggregateException (ex), false, true);
+				if (throwException)
+					throw ex;
+			}
 		}
 		
 		void ThreadStart ()
@@ -401,7 +432,10 @@ namespace System.Threading.Tasks
 				status = TaskStatus.Running;
 				
 				try {
-					InnerInvoke ();
+					if (ec != null)
+						ExecutionContext.Run (ec, l => ((Task) l).InnerInvoke (), this);
+					else
+						InnerInvoke ();
 				} catch (OperationCanceledException oce) {
 					if (token != CancellationToken.None && oce.CancellationToken == token)
 						CancelReal ();
@@ -438,7 +472,7 @@ namespace System.Threading.Tasks
 			return true;
 		}
 
-		internal bool TrySetException (AggregateException aggregate)
+		internal bool TrySetException (AggregateException aggregate, bool cancellation, bool observed)
 		{
 			if (IsCompleted)
 				return false;
@@ -450,8 +484,19 @@ namespace System.Threading.Tasks
 
 				return false;
 			}
-			
-			HandleGenericException (aggregate);
+
+			if (cancellation) {
+				ExceptionSlot.Exception = aggregate;
+				Thread.MemoryBarrier ();
+
+				CancelReal ();
+			} else {
+				HandleGenericException (aggregate);
+			}
+
+			if (observed)
+				exSlot.Observed = true;
+
 			return true;
 		}
 
@@ -488,19 +533,17 @@ namespace System.Threading.Tasks
 				ProcessChildExceptions ();
 				Status = exSlot == null ? TaskStatus.RanToCompletion : TaskStatus.Faulted;
 				ProcessCompleteDelegates ();
-				if (parent != null &&
-#if NET_4_5
-				    !HasFlag (parent.CreationOptions, TaskCreationOptions.DenyChildAttach) &&
-#endif
-					HasFlag (creationOptions, TaskCreationOptions.AttachedToParent))
-					parent.ChildCompleted (this.Exception);
+				if (parent != null && NotifyParentOnFinish ())
+					parent = null;
 			}
 		}
 
 		void InnerInvoke ()
 		{
 			if (IsContinuation) {
-				invoker.Invoke (contAncestor, state, this);
+				var ancestor = contAncestor;
+				contAncestor = null;
+				invoker.Invoke (ancestor, state, this);
 			} else {
 				invoker.Invoke (this, state, this);
 			}
@@ -526,12 +569,12 @@ namespace System.Threading.Tasks
 				wait_handle.Set ();
 
 			// Tell parent that we are finished
-			if (parent != null && HasFlag (creationOptions, TaskCreationOptions.AttachedToParent) &&
-#if NET_4_5
-			    !HasFlag (parent.CreationOptions, TaskCreationOptions.DenyChildAttach) &&
-#endif
-				status != TaskStatus.WaitingForChildrenToComplete) {
-				parent.ChildCompleted (this.Exception);
+			if (parent != null && NotifyParentOnFinish ()) {
+				//
+				// Break the reference back to the parent, otherwise any Tasks created from another Task's thread of 
+				// execution will create an undesired linked-list that the GC cannot free. See bug #18398.
+				//
+				parent = null;
 			}
 
 			// Completions are already processed when task is canceled or faulted
@@ -548,6 +591,20 @@ namespace System.Threading.Tasks
 				cancellationRegistration.Value.Dispose ();
 		}
 
+		bool NotifyParentOnFinish ()
+		{
+			if (!HasFlag (creationOptions, TaskCreationOptions.AttachedToParent))
+				return true;
+#if NET_4_5
+			if (HasFlag (parent.CreationOptions, TaskCreationOptions.DenyChildAttach))
+				return true;
+#endif
+			if (status != TaskStatus.WaitingForChildrenToComplete)
+				parent.ChildCompleted (Exception);
+
+			return false;
+		}
+
 		void ProcessCompleteDelegates ()
 		{
 			if (continuations.HasElements) {
@@ -639,7 +696,7 @@ namespace System.Threading.Tasks
 			if (millisecondsTimeout < -1)
 				throw new ArgumentOutOfRangeException ("millisecondsTimeout");
 
-			bool result = WaitCore (millisecondsTimeout, cancellationToken);
+			bool result = WaitCore (millisecondsTimeout, cancellationToken, true);
 
 			if (IsCanceled)
 				throw new AggregateException (new TaskCanceledException (this));
@@ -651,14 +708,19 @@ namespace System.Threading.Tasks
 			return result;
 		}
 
-		internal bool WaitCore (int millisecondsTimeout, CancellationToken cancellationToken)
+		internal bool WaitCore (int millisecondsTimeout, CancellationToken cancellationToken, bool runInline)
 		{
 			if (IsCompleted)
 				return true;
 
 			// If the task is ready to be run and we were supposed to wait on it indefinitely without cancellation, just run it
-			if (Status == TaskStatus.WaitingToRun && millisecondsTimeout == Timeout.Infinite && scheduler != null && !cancellationToken.CanBeCanceled)
-				scheduler.RunInline (this, true);
+			if (runInline && Status == TaskStatus.WaitingToRun && millisecondsTimeout == Timeout.Infinite && scheduler != null && !cancellationToken.CanBeCanceled) {
+				try {
+					scheduler.RunInline (this, true);
+				} catch (Exception e) {
+					throw new TaskSchedulerException (e);
+				}
+			}
 
 			bool result = true;
 
@@ -837,6 +899,7 @@ namespace System.Threading.Tasks
 			// Set action to null so that the GC can collect the delegate and thus
 			// any big object references that the user might have captured in a anonymous method
 			if (disposing) {
+				parent = null;
 				invoker = null;
 				state = null;
 				if (cancellationRegistration != null)
@@ -957,11 +1020,23 @@ namespace System.Threading.Tasks
 			if (millisecondsDelay < -1)
 				throw new ArgumentOutOfRangeException ("millisecondsDelay");
 
-			var task = new Task (TaskActionInvoker.Delay, millisecondsDelay, cancellationToken, TaskCreationOptions.None, null, TaskConstants.Finished);
+			if (cancellationToken.IsCancellationRequested)
+				return TaskConstants.Canceled;
+
+			var task = new Task (TaskActionInvoker.Empty, null, cancellationToken, TaskCreationOptions.None, null, null);
 			task.SetupScheduler (TaskScheduler.Default);
-			
-			if (millisecondsDelay != Timeout.Infinite)
-				task.scheduler.QueueTask (task);
+
+			if (millisecondsDelay != Timeout.Infinite) {
+				var timer = new Timer (delegate (object state) {
+					var t = (Task) state;
+					if (t.Status == TaskStatus.WaitingForActivation) {
+						t.Status = TaskStatus.Running;
+						t.Finish ();
+					}
+				}, task, millisecondsDelay, -1);
+
+				task.ContinueWith (new DisposeContinuation (timer));
+			}
 
 			return task;
 		}
@@ -1068,6 +1143,9 @@ namespace System.Threading.Tasks
 
 		internal static Task<TResult[]> WhenAllCore<TResult> (IList<Task<TResult>> tasks)
 		{
+			if (tasks.Count == 0)
+				return FromResult(new TResult[0]);
+
 			foreach (var t in tasks) {
 				if (t == null)
 					throw new ArgumentException ("tasks", "the tasks argument contains a null element");
@@ -1228,7 +1306,7 @@ namespace System.Threading.Tasks
 		
 		public AggregateException Exception {
 			get {
-				if (exSlot == null)
+				if (exSlot == null || !IsFaulted)
 					return null;
 				exSlot.Observed = true;
 				return exSlot.Exception;
@@ -1269,7 +1347,7 @@ namespace System.Threading.Tasks
 			}
 		}
 
-		TaskExceptionSlot ExceptionSlot {
+		internal TaskExceptionSlot ExceptionSlot {
 			get {
 				if (exSlot != null)
 					return exSlot;
@@ -1286,7 +1364,7 @@ namespace System.Threading.Tasks
 		
 		bool IAsyncResult.CompletedSynchronously {
 			get {
-				return true;
+				return false;
 			}
 		}
 
@@ -1314,6 +1392,12 @@ namespace System.Threading.Tasks
 			}
 		}
 
+		bool IsPromise {
+			get {
+				return invoker == TaskActionInvoker.Promise;
+			}
+		}
+
 		internal Task ContinuationAncestor {
 			get {
 				return contAncestor;
