@@ -138,6 +138,7 @@ namespace MonoTests.System.IO
 		}
 
 		[Test]
+		[Category ("MobileNotWorking")]
 		public void Name ()
 		{
 			string path = TempFolder + DSC + "DIT.Name.Test";
@@ -558,6 +559,18 @@ namespace MonoTests.System.IO
 			}
 		}
 
+		[Test] //GetDirectories (String, SearchOptions)
+		public void GetDirectiories_SearchOptionAllDirectories ()
+		{
+			string directoryToBeLookedFor = "lookforme";
+			DirectoryInfo baseDir = Directory.CreateDirectory(Path.Combine(TempFolder, "GetDirectiories_SearchOptionAllDirectories"));
+			DirectoryInfo subdir = baseDir.CreateSubdirectory("subdir");
+			DirectoryInfo subsubdir = subdir.CreateSubdirectory(directoryToBeLookedFor);
+			DirectoryInfo[] directoriesFound = baseDir.GetDirectories(directoryToBeLookedFor, SearchOption.AllDirectories);
+			Assert.AreEqual(1, directoriesFound.Length, "There should be exactly one directory with the specified name.");
+			Assert.AreEqual(directoryToBeLookedFor, directoriesFound[0].Name, "The name of the directory found should match the expected one.");
+		}
+
 #if NET_2_0
 		[Test] // GetDirectories (String, SearchOption)
 		public void GetDirectories3_SearchPattern_Null ()
@@ -866,6 +879,32 @@ namespace MonoTests.System.IO
 				DeleteDir (path);
 			}
 		}
+		
+		[Test]
+		public void MoveTo_UpdateProperties ()
+		{
+			string path = TempFolder + DSC + "DIT.MoveToUpdateProperties.Test";
+			string path2 = TempFolder + DSC + "DIT.MoveToUpdateProperties2.Test";
+			string path3 = path2 + DSC + "DIT.MoveToUpdateProperties3.Test";
+			DeleteDir (path);
+			Directory.CreateDirectory (path);
+			Directory.CreateDirectory (path2);
+
+			DirectoryInfo info = new DirectoryInfo(path);
+			
+			Assert.IsTrue (Directory.Exists(info.FullName));
+			Assert.AreEqual (path, info.FullName);
+			Assert.AreEqual ("DIT.MoveToUpdateProperties.Test", info.Name);
+			Assert.AreEqual (TempFolder, info.Parent.FullName);
+			Assert.AreEqual (path, info.ToString ());
+
+			info.MoveTo (path3);
+			Assert.IsTrue (Directory.Exists(info.FullName));
+			Assert.AreEqual (path3, info.FullName);
+			Assert.AreEqual ("DIT.MoveToUpdateProperties3.Test", info.Name);
+			Assert.AreEqual (path2, info.Parent.FullName);
+			Assert.AreEqual (path3, info.ToString ());
+		}
 
 		[Test]
 		public void DirectoryNameWithSpace ()
@@ -928,7 +967,6 @@ namespace MonoTests.System.IO
 		}
 
 		[Test]
-		[Category("TargetJvmNotSupported")] // LastAccessTime not supported for TARGET_JVM
 		public void LastAccessTime ()
 		{
 			DirectoryInfo info = new DirectoryInfo (TempFolder);
@@ -936,7 +974,6 @@ namespace MonoTests.System.IO
 		}
 
 		[Test]
-		[Category("TargetJvmNotSupported")] // LastAccessTime not supported for TARGET_JVM
 		public void LastAccessTimeUtc ()
 		{
 			DirectoryInfo info = new DirectoryInfo (TempFolder);
@@ -944,7 +981,6 @@ namespace MonoTests.System.IO
 		}
 
 		[Test]
-		[Category("TargetJvmNotSupported")] // CreationTime not supported for TARGET_JVM
 		public void CreationTime ()
 		{
 			DirectoryInfo info = new DirectoryInfo (TempFolder);
@@ -952,7 +988,6 @@ namespace MonoTests.System.IO
 		}
 
 		[Test]
-		[Category("TargetJvmNotSupported")] // CreationTime not supported for TARGET_JVM
 		public void CreationTimeUtc ()
 		{
 			DirectoryInfo info = new DirectoryInfo (TempFolder);
@@ -991,7 +1026,7 @@ namespace MonoTests.System.IO
 		public void WindowsSystem32_76191 ()
 		{
 			if (RunningOnUnix)
-				return;
+				Assert.Ignore ("Running on Unix.");
 
 			Directory.SetCurrentDirectory (@"C:\WINDOWS\system32");
 			WindowsParentFullName ("C:", "C:\\WINDOWS");
@@ -1024,6 +1059,7 @@ namespace MonoTests.System.IO
 			Assert.AreEqual (TempFolder + DSC + "ToString.Test", info.ToString ());
 		}
 
+#if !MOBILE
 		[Test]
 		public void Serialization ()
 		{
@@ -1061,7 +1097,7 @@ namespace MonoTests.System.IO
 			Assert.AreEqual (info.Name, clone.Name, "#1");
 			Assert.AreEqual (info.FullName, clone.FullName, "#2");
 		}
-		
+
 		// Needed so that UnixSymbolicLinkInfo doesn't have to
 		// be JITted on windows
 		private void Symlink_helper ()
@@ -1094,20 +1130,16 @@ namespace MonoTests.System.IO
 			// Linux-like platforms but mono-on-windows
 			// doesn't set the NotDotNet category
 			if (!RunningOnUnix) {
-				return;
+				Assert.Ignore ("Not running on Unix.");
 			}
 
 			Symlink_helper ();
 		}
-
+#endif
 		static bool RunningOnUnix {
 			get {
-#if NET_2_0
-				return Environment.OSVersion.Platform == PlatformID.Unix;
-#else
-				int platform = (int) Environment.OSVersion.Platform;
-				return platform == 128;
-#endif
+				int p = (int) Environment.OSVersion.Platform;
+				return ((p == 4) || (p == 128) || (p == 6));
 			}
 		}
 
