@@ -52,8 +52,20 @@ namespace MonoTests.System.Runtime.CompilerServices {
 		}
 	}
 
-	class Key { public int Foo; }
-	class Val { public int Foo; }
+	class Key {
+		public int Foo;
+		public override string ToString () {
+				return "key-" + Foo;
+			}
+	}
+
+	class Val {
+		public int Foo;
+		public override string ToString () {
+			return "value-" + Foo;
+		}
+	}
+	
 
 	[Test]
 	public void GetValue () {
@@ -182,11 +194,15 @@ namespace MonoTests.System.Runtime.CompilerServices {
 	[Test]
 	public void Reachability () {
 		if (GC.MaxGeneration == 0) /*Boehm doesn't handle ephemerons */
-			return;
+			Assert.Ignore ("Not working on Boehm.");
 		var cwt = new ConditionalWeakTable <object,object> ();
-		List<object> keepAlive;
-		List<WeakReference> keys;
-		FillStuff (cwt, out keepAlive, out keys);
+		List<object> keepAlive = null;
+		List<WeakReference> keys = null;
+		Thread t = new Thread (delegate () {
+				FillStuff (cwt, out keepAlive, out keys);
+			});
+		t.Start ();
+		t.Join ();
 
 		GC.Collect ();
 
@@ -229,7 +245,7 @@ namespace MonoTests.System.Runtime.CompilerServices {
 	[Test]
 	public void InsertStress () {
 		if (GC.MaxGeneration == 0) /*Boehm doesn't handle ephemerons */
-			return;
+			Assert.Ignore ("Not working on Boehm.");
 		var cwt = new ConditionalWeakTable <object,object> ();
 
 		var a = new object ();
@@ -280,7 +296,7 @@ namespace MonoTests.System.Runtime.CompilerServices {
 	[Test]
 	public void OldGenStress () {
 		if (GC.MaxGeneration == 0) /*Boehm doesn't handle ephemerons */
-			return;
+			Assert.Ignore ("Not working on Boehm.");
 		var cwt = new ConditionalWeakTable <object,object>[1];
 		List<object> k = null;
 		List<WeakReference> res, res2;
@@ -424,7 +440,7 @@ namespace MonoTests.System.Runtime.CompilerServices {
 	public void FinalizableObjectsThatRetainDeadKeys ()
 	{
 		if (GC.MaxGeneration == 0) /*Boehm doesn't handle ephemerons */
-			return;
+			Assert.Ignore ("Not working on Boehm.");
 		lock (_lock1) { 
 			var cwt = new ConditionalWeakTable <object,object> ();
 			ThreadStart dele = () => { FillWithFinalizable (cwt); };
@@ -448,7 +464,7 @@ namespace MonoTests.System.Runtime.CompilerServices {
 	public void OldGenKeysMakeNewGenObjectsReachable ()
 	{
 		if (GC.MaxGeneration == 0) /*Boehm doesn't handle ephemerons */
-		return;
+			Assert.Ignore ("Not working on Boehm.");
 		ConditionalWeakTable<object, Val> table = new ConditionalWeakTable<object, Val>();
 		List<Key> keys = new List<Key>();
 
@@ -457,21 +473,21 @@ namespace MonoTests.System.Runtime.CompilerServices {
 		// should be collected ever.
 		//
 		for (int x = 0; x < 1000; x++) 
-		keys.Add(new Key() { Foo = x });
+			keys.Add (new Key () { Foo = x });
 
-		for (int i = 0; i < 10000; ++i) {
+		for (int i = 0; i < 1000; ++i) {
 			// Insert all keys into the ConditionalWeakTable
 			foreach (var key in keys)
-			table.Add(key, new Val() { Foo = key.Foo });
+				table.Add (key, new Val () { Foo = key.Foo });
 
 			// Look up all keys to verify that they are still there
 			Val val;
 			foreach (var key in keys)
-				Assert.IsTrue (table.TryGetValue(key, out val), "#1-" + i);
+				Assert.IsTrue (table.TryGetValue (key, out val), "#1-" + i + "-k-" + key);
 
 			// Remove all keys from the ConditionalWeakTable
 			foreach (var key in keys)
-				Assert.IsTrue (!table.Remove(key), "#2-" + i);
+				Assert.IsTrue (table.Remove (key), "#2-" + i + "-k-" + key);
 		}
 	}
 	}
