@@ -7,7 +7,6 @@
 //
 // Copyright (C) 2004-2007 Novell, Inc (http://www.novell.com)
 //
-#if !TARGET_JVM
 using NUnit.Framework;
 using System;
 using System.IO;
@@ -114,6 +113,13 @@ namespace MonoTests.System.Runtime.InteropServices
 			}
 		}
 
+		[Test]
+		public unsafe void Sizeof_Pointer ()
+		{
+			int size = Marshal.SizeOf (typeof (char*));
+			Assert.IsTrue (size == 4 || size == 8);
+		}
+
 		[Test]
 		public void PtrToStringWithNull ()
 		{
@@ -871,4 +877,3 @@ namespace MonoTests.System.Runtime.InteropServices
 	}
 #endif
 }
-#endif
