@@ -25,9 +25,7 @@ namespace MonoTests.System.Threading
 	{
 		public static void NoPrincipal () 
 		{
-#if !TARGET_JVM // AppDomain.SetPrincipalPolicy not supported for TARGET_JVM
 			AppDomain.CurrentDomain.SetPrincipalPolicy (PrincipalPolicy.NoPrincipal);
-#endif
 			IPrincipal p = Thread.CurrentPrincipal;
 			Assert.IsNull (p, "#1");
 
@@ -39,7 +37,6 @@ namespace MonoTests.System.Threading
 			// in this case we can return to null
 		}
 
-#if !TARGET_JVM // AppDomain.SetPrincipalPolicy not supported for TARGET_JVM
 		public static void UnauthenticatedPrincipal () 
 		{
 			AppDomain.CurrentDomain.SetPrincipalPolicy (PrincipalPolicy.UnauthenticatedPrincipal);
@@ -76,7 +73,6 @@ namespace MonoTests.System.Threading
 			Assert.IsNotNull (Thread.CurrentPrincipal, "#7");
 			// in this case we can't return to null
 		}
-#endif // TARGET_JVM
 
 		public static void CopyOnNewThread ()
 		{
@@ -86,8 +82,15 @@ namespace MonoTests.System.Threading
 	}
 
 	[TestFixture]
+	[Category("MobileNotWorking")] // Abort #10240
 	public class ThreadTest
 	{
+		TimeSpan Infinite = new TimeSpan (-10000);	// -10000 ticks == -1 ms
+		TimeSpan SmallNegative = new TimeSpan (-2);	// between 0 and -1.0 (infinite) ms
+		TimeSpan Negative = new TimeSpan (-20000);	// really negative
+		TimeSpan MaxValue = TimeSpan.FromMilliseconds ((long) Int32.MaxValue);
+		TimeSpan TooLarge = TimeSpan.FromMilliseconds ((long) Int32.MaxValue + 1);
+
 		static bool is_win32;
 		static bool is_mono;
 
@@ -228,7 +231,27 @@ namespace MonoTests.System.Threading
 		public void TestCtor1()
 		{
 			C1Test test1 = new C1Test();
-			Thread TestThread = new Thread(new ThreadStart(test1.TestMethod));
+			Thread t = new Thread (new ThreadStart (test1.TestMethod));
+
+			Assert.IsTrue (t.CurrentCulture.IsReadOnly, "CurrentCulture.IsReadOnly");
+			Assert.IsFalse (t.IsAlive, "IsAlive");
+			Assert.IsFalse (t.IsBackground, "IsBackground");
+			Assert.IsNull (t.Name, "Name");
+			Assert.AreEqual (ThreadState.Unstarted, t.ThreadState, "ThreadState");
+		}
+
+		[Test]
+		[Category ("NotWorking")] // we're not sharing (read-only) CultureInfo
+		public void CultureInfo_Shared_Across_Threads ()
+		{
+			Thread t = new Thread (TestCtor1);
+			Assert.AreSame (t.CurrentCulture, t.CurrentUICulture, "Culture");
+
+			Assert.AreSame (t.CurrentCulture, CultureInfo.CurrentCulture, "CultureInfo.CurrentCulture");
+			Assert.AreSame (t.CurrentUICulture, CultureInfo.CurrentUICulture, "CultureInfo.CurrentUICulture");
+
+			Assert.AreSame (t.CurrentCulture, Thread.CurrentThread.CurrentCulture, "Thread.CurrentThread.CurrentCulture");
+			Assert.AreSame (t.CurrentUICulture, Thread.CurrentThread.CurrentUICulture, "Thread.CurrentThread.CurrentUICulture");
 		}
 
 		[Test] // bug #325566
@@ -422,6 +445,16 @@ namespace MonoTests.System.Threading
 			}
 		}
 
+		[Test]
+		public void TestUndivisibleByPageSizeMaxStackSize ()
+		{
+			const int undivisible_stacksize = 1048573;
+
+			var thread = new Thread (new ThreadStart (delegate {}), undivisible_stacksize);
+			thread.Start ();
+			thread.Join ();
+		}
+
 		[Test]
 		public void TestIsBackground1 ()
 		{
@@ -441,7 +474,6 @@ namespace MonoTests.System.Threading
 		}
 
 		[Test]
-		[Category ("NotDotNet")] // on MS, ThreadState is immediately Stopped after Abort
 		public void TestIsBackground2 ()
 		{
 			C2Test test1 = new C2Test();
@@ -452,7 +484,14 @@ namespace MonoTests.System.Threading
 			} finally {
 				TestThread.Abort();
 			}
-			Assert.IsTrue (TestThread.IsBackground, "#52 Is Background Changed to Start ");
+			
+			if (TestThread.IsAlive) {
+				try {
+					Assert.IsTrue (TestThread.IsBackground, "#52 Is Background Changed to Start ");
+				} catch (ThreadStateException) {
+					// Ignore if thread died meantime
+				}
+			}
 		}
 
 		[Test]
@@ -476,6 +515,24 @@ namespace MonoTests.System.Threading
 			}
 		}
 
+		[Test]
+		public void Name ()
+		{
+			Thread t = new Thread (new ThreadStart (Name));
+			Assert.IsNull (t.Name, "Name-1");
+			t.Name = null;
+			Assert.IsNull (t.Name, "Name-2");
+		}
+
+		[Test]
+		[ExpectedException (typeof (InvalidOperationException))]
+		public void ReName ()
+		{
+			Thread t = new Thread (new ThreadStart (ReName));
+			t.Name = "a";
+			t.Name = "b";
+		}
+
 		[Test]
 		public void TestNestedThreads1()
 		{
@@ -518,6 +575,70 @@ namespace MonoTests.System.Threading
 			}
 		}
 
+		[Test]
+		[ExpectedException (typeof (ArgumentOutOfRangeException))]
+		public void Join_Int32_Negative ()
+		{
+			// -1 is Timeout.Infinite
+			Thread.CurrentThread.Join (-2);
+		}
+
+		[Test]
+		[ExpectedException (typeof (ArgumentOutOfRangeException))]
+		public void Join_TimeSpan_Negative ()
+		{
+			Thread.CurrentThread.Join (Negative);
+		}
+
+		[Test]
+		[ExpectedException (typeof (ArgumentOutOfRangeException))]
+		public void Join_TimeSpan_TooLarge ()
+		{
+			Thread.CurrentThread.Join (TooLarge);
+		}
+
+		[Test]
+		public void Join_TimeSpan_SmallNegative ()
+		{
+			Thread.CurrentThread.Join (SmallNegative);
+		}
+
+		[Test]
+		[ExpectedException (typeof (ArgumentOutOfRangeException))]
+		public void Sleep_Int32_Negative ()
+		{
+			// -1 is Timeout.Infinite
+			Thread.Sleep (-2);
+		}
+
+		[Test]
+		public void Sleep_TimeSpan_SmallNegative ()
+		{
+			Thread.Sleep (SmallNegative);
+		}
+
+		[Test]
+		[ExpectedException (typeof (ArgumentOutOfRangeException))]
+		public void Sleep_TimeSpan_Negative ()
+		{
+			Thread.Sleep (Negative);
+		}
+
+		[Test]
+		[ExpectedException (typeof (ArgumentOutOfRangeException))]
+		public void Sleep_TimeSpan_TooLarge ()
+		{
+			Thread.Sleep (TooLarge);
+		}
+
+		[Test]
+		public void SpinWait ()
+		{
+			// no exception for negative numbers
+			Thread.SpinWait (Int32.MinValue);
+			Thread.SpinWait (0);
+		}
+
 		[Test]
 		public void TestThreadState ()
 		{
@@ -559,7 +680,6 @@ namespace MonoTests.System.Threading
 			}
 		}
 
-#if !TARGET_JVM // AppDomain.SetPrincipalPolicy not supported for TARGET_JVM
 		[Test]
 		[Ignore ("see comment below.")]
 		public void CurrentPrincipal_PrincipalPolicy_UnauthenticatedPrincipal () 
@@ -588,7 +708,6 @@ namespace MonoTests.System.Threading
 				t.Abort ();
 			}
 		}
-#endif // TARGET_JVM
 		
 		[Test]
 		public void IPrincipal_CopyOnNewThread () 
@@ -684,6 +803,13 @@ namespace MonoTests.System.Threading
 			Assert.IsTrue (interruptedExceptionThrown, "ThreadInterruptedException expected.");
 		}
 
+		[Test]
+		[ExpectedException (typeof (ArgumentNullException))]
+		public void TestQueueUserWorkItemNullCallback ()
+		{
+			ThreadPool.QueueUserWorkItem (null, null);
+		}
+
 		private void Test_Interrupt_Worker (object o)
 		{
 			Thread t = o as Thread;
@@ -692,25 +818,24 @@ namespace MonoTests.System.Threading
 		}
 		
 		[Test]
+		[Category ("NotDotNet")] // it crashes nunit.
 		public void Test_InterruptCurrentThread ()
 		{
 			bool interruptedExceptionThrown = false;
 
+			Thread.CurrentThread.Interrupt ();
 			try {
-				try {
-					Thread.CurrentThread.Interrupt ();
-				} finally {
-					try {
-						Thread.Sleep (0);
-					} catch (ThreadInterruptedException) {
-						Assert.Fail ("ThreadInterruptedException should not be thrown.");
-					}
-				}
+				Thread.Sleep (0);
+				Assert.Fail ();
 			} catch (ThreadInterruptedException) {
-				interruptedExceptionThrown = true;
 			}
+		}
 
-			Assert.IsFalse (interruptedExceptionThrown, "ThreadInterruptedException should not be thrown.");
+		[Test]
+		public void GetNamedDataSlotTest ()
+		{
+			Assert.IsNotNull (Thread.GetNamedDataSlot ("te#st"), "#1");
+			Assert.AreSame (Thread.GetNamedDataSlot ("te#st"), Thread.GetNamedDataSlot ("te#st"), "#2");
 		}
 
 		void CheckIsRunning (string s, Thread t)
@@ -805,6 +930,107 @@ namespace MonoTests.System.Threading
 		}
 	}
 
+	[TestFixture]
+	[Serializable]
+	public class ThreadTest_ManagedThreadId
+	{
+		AppDomain ad1;
+		AppDomain ad2;
+		MBRO mbro = new MBRO ();
+
+		class MBRO : MarshalByRefObject {
+			public int id_a1;
+			public int id_b1;
+			public int id_b2;
+			public string ad_a1;
+			public string ad_b1;
+			public string ad_b2;
+			public string message;
+		}
+#if !MOBILE
+		[Test]
+		public void ManagedThreadId_AppDomains ()
+		{
+			AppDomain currentDomain = AppDomain.CurrentDomain;
+			ad1 = AppDomain.CreateDomain ("AppDomain 1", currentDomain.Evidence, currentDomain.SetupInformation);
+			ad2 = AppDomain.CreateDomain ("AppDomain 2", currentDomain.Evidence, currentDomain.SetupInformation);
+
+			Thread a = new Thread (ThreadA);
+			Thread b = new Thread (ThreadB);
+			// execute on AppDomain 1 thread A
+			// execute on AppDomain 2 thread B
+			// execute on AppDomain 1 thread B - must have same ManagedThreadId as Ad 2 on thread B
+			a.Start ();
+			a.Join ();
+			b.Start ();
+			b.Join ();
+
+			AppDomain.Unload (ad1);
+			AppDomain.Unload (ad2);
+
+			if (mbro.message != null)
+				Assert.Fail (mbro.message);
+
+			// Console.WriteLine ("Done id_a1: {0} id_b1: {1} id_b2: {2} ad_a1: {3} ad_b1: {4} ad_b2: {5}", mbro.id_a1, mbro.id_b1, mbro.id_b2, mbro.ad_a1, mbro.ad_b1, mbro.ad_b2);
+
+			Assert.AreEqual ("AppDomain 1", mbro.ad_a1, "Name #1");
+			Assert.AreEqual ("AppDomain 1", mbro.ad_b1, "Name #2");
+			Assert.AreEqual ("AppDomain 2", mbro.ad_b2, "Name #3");
+
+			Assert.AreNotEqual (mbro.id_a1, mbro.id_b1, "Id #1");
+			Assert.AreNotEqual (mbro.id_a1, mbro.id_b2, "Id #2");
+			Assert.AreEqual (mbro.id_b1, mbro.id_b2, "Id #3");
+
+			Assert.AreNotEqual (mbro.id_a1, Thread.CurrentThread.ManagedThreadId, "Id #4");
+			Assert.AreNotEqual (mbro.id_b1, Thread.CurrentThread.ManagedThreadId, "Id #5");
+			Assert.AreNotEqual (mbro.id_b2, Thread.CurrentThread.ManagedThreadId, "Id #6");
+			Assert.AreNotEqual (mbro.ad_a1, AppDomain.CurrentDomain.FriendlyName, "Name #4");
+			Assert.AreNotEqual (mbro.ad_b1, AppDomain.CurrentDomain.FriendlyName, "Name #5");
+			Assert.AreNotEqual (mbro.ad_b2, AppDomain.CurrentDomain.FriendlyName, "Name #6");
+		}
+#endif
+		void A1 ()
+		{
+			mbro.id_a1 = Thread.CurrentThread.ManagedThreadId;
+			mbro.ad_a1 = AppDomain.CurrentDomain.FriendlyName;
+		}
+		
+		void B2 ()
+		{
+			mbro.id_b2 = Thread.CurrentThread.ManagedThreadId;
+			mbro.ad_b2 = AppDomain.CurrentDomain.FriendlyName;
+		}
+
+		void B1 ()
+		{
+			mbro.id_b1 = Thread.CurrentThread.ManagedThreadId;
+			mbro.ad_b1 = AppDomain.CurrentDomain.FriendlyName;
+		}
+
+		void ThreadA (object obj)
+		{
+			// Console.WriteLine ("ThreadA");
+			try {
+				ad1.DoCallBack (A1);
+			} catch (Exception ex) {
+				mbro.message = string.Format ("ThreadA exception: {0}", ex);
+			}
+			// Console.WriteLine ("ThreadA Done");
+		}
+
+		void ThreadB (object obj)
+		{
+			// Console.WriteLine ("ThreadB");
+			try {
+				ad2.DoCallBack (B2);
+				ad1.DoCallBack (B1);
+			} catch (Exception ex) {
+				mbro.message = string.Format ("ThreadB exception: {0}", ex);
+			}
+			// Console.WriteLine ("ThreadB Done");
+		}
+	}
+
 	[TestFixture]
 	public class ThreadApartmentTest
 	{
@@ -896,6 +1122,117 @@ namespace MonoTests.System.Threading
 			}
 			Assert.IsTrue (exception_occured, "Thread1 Started Invalid Exception Occured");
 		}
+
+		[Test]
+		public void TestSetApartmentStateSameState ()
+		{
+			Thread t1 = new Thread (new ThreadStart (Start));
+			t1.SetApartmentState (ApartmentState.STA);
+			Assert.AreEqual (ApartmentState.STA, t1.ApartmentState, "Thread1 Set Once");
+
+			t1.SetApartmentState (ApartmentState.STA);
+			Assert.AreEqual (ApartmentState.STA, t1.ApartmentState, "Thread1 Set twice");
+		}
+
+		[Test]
+		[ExpectedException(typeof(InvalidOperationException))]
+		public void TestSetApartmentStateDiffState ()
+		{
+			Thread t1 = new Thread (new ThreadStart (Start));
+			t1.SetApartmentState (ApartmentState.STA);
+			Assert.AreEqual (ApartmentState.STA, t1.ApartmentState, "Thread1 Set Once");
+
+			t1.SetApartmentState (ApartmentState.MTA);
+		}
+
+		[Test]
+		public void TestTrySetApartmentState ()
+		{
+			Thread t1 = new Thread (new ThreadStart (Start));
+			t1.SetApartmentState (ApartmentState.STA);
+			Assert.AreEqual (ApartmentState.STA, t1.ApartmentState, "#1");
+
+			bool result = t1.TrySetApartmentState (ApartmentState.MTA);
+			Assert.IsFalse (result, "#2");
+
+			result = t1.TrySetApartmentState (ApartmentState.STA);
+			Assert.IsTrue (result, "#3");
+		}
+
+		[Test]
+		public void TestTrySetApartmentStateRunning ()
+		{
+			Thread t1 = new Thread (new ThreadStart (Start));
+			t1.SetApartmentState (ApartmentState.STA);
+			Assert.AreEqual (ApartmentState.STA, t1.ApartmentState, "#1");
+
+			t1.Start ();
+
+			try {
+				t1.TrySetApartmentState (ApartmentState.STA);
+				Assert.Fail ("#2");
+			} catch (ThreadStateException) {
+			}
+
+			t1.Join ();
+		}
+
+		[Test]
+		public void Volatile () {
+			double v3 = 55667;
+			Thread.VolatileWrite (ref v3, double.MaxValue);
+			Assert.AreEqual (v3, double.MaxValue);
+
+			float v4 = 1;
+			Thread.VolatileWrite (ref v4, float.MaxValue);
+			Assert.AreEqual (v4, float.MaxValue);
+		}
+
+		[Test]
+		public void Culture ()
+		{
+			Assert.IsNotNull (Thread.CurrentThread.CurrentCulture, "CurrentCulture");
+			Assert.IsNotNull (Thread.CurrentThread.CurrentUICulture, "CurrentUICulture");
+		}
+
+		[Test]
+		public void ThreadStartSimple ()
+		{
+			int i = 0;
+			Thread t = new Thread (delegate () {
+				// ensure the NSAutoreleasePool works
+				i++;
+			});
+			t.Start ();
+			t.Join ();
+			Assert.AreEqual (1, i, "ThreadStart");
+		}
+
+		[Test]
+		public void ParametrizedThreadStart ()
+		{
+			int i = 0;
+			object arg = null;
+			Thread t = new Thread (delegate (object obj) {
+				// ensure the NSAutoreleasePool works
+				i++;
+				arg = obj;
+			});
+			t.Start (this);
+			t.Join ();
+
+			Assert.AreEqual (1, i, "ParametrizedThreadStart");
+			Assert.AreEqual (this, arg, "obj");	
+		}		
+
+		[Test]
+		public void SetNameTpThread () {
+			ThreadPool.QueueUserWorkItem(new WaitCallback(ThreadProc));
+		}
+
+		static void ThreadProc(Object stateInfo) {
+			Thread.CurrentThread.Name = "My Worker";
+		}
 	}
 
 	public class TestUtil
