@@ -565,6 +565,7 @@ public class TimeSpanTest {
 		Assert.AreEqual (true, TimeSpan.Equals (null, null), "A10");
 	}
 
+	[Test]
 	public void TestFromXXXX ()
 	{
 		Assert.AreEqual ("12.08:16:48", TimeSpan.FromDays (12.345).ToString (), "A1");
@@ -573,6 +574,11 @@ public class TimeSpanTest {
 		Assert.AreEqual ("00:00:12.3450000", TimeSpan.FromSeconds (12.345).ToString (), "A4");
 		Assert.AreEqual ("00:00:00.0120000", TimeSpan.FromMilliseconds (12.345).ToString (), "A5");
 		Assert.AreEqual ("00:00:00.0012345", TimeSpan.FromTicks (12345).ToString (), "A6");
+		Assert.AreEqual ("-00:00:00.0010000", TimeSpan.FromMilliseconds (-0.5).ToString (), "A7");
+		Assert.AreEqual ("00:00:00.0010000", TimeSpan.FromMilliseconds (0.5).ToString (), "A8");
+		Assert.AreEqual ("-00:00:00.0030000", TimeSpan.FromMilliseconds (-2.5).ToString (), "A9");
+		Assert.AreEqual ("00:00:00.0030000", TimeSpan.FromMilliseconds (2.5).ToString (), "A10");
+		Assert.AreEqual ("00:00:00.0010000", TimeSpan.FromSeconds (0.0005).ToString (), "A11");
 	}
 
 	[Test]
@@ -818,7 +824,7 @@ public class TimeSpanTest {
 		ParseHelper (" 1:2:3:12345678 ", true, false, "dontcare"); 
 #endif
 
-#if NET_4_0		
+#if NET_4_0
 		ParseHelper ("10:11:12:13", false, false, "10.11:12:13"); // Days using : instead of . as separator
 		ParseHelper ("10.11", true, false, "dontcare"); // days+hours is invalid
 
@@ -1316,6 +1322,8 @@ public class TimeSpanTest {
 		TryParseExactHelper ("10:12", new string [0], true, "dontcare");
 		TryParseExactHelper ("10:12", new string [] { String.Empty }, true, "dontcare");
 		TryParseExactHelper ("10:12", new string [] { null }, true, "dontcare");
+
+		TryParseExactHelper (null, new string [] { null }, true, "dontcare");
 	}
 
 	void TryParseExactHelper (string input, string [] formats, bool error, string expected, IFormatProvider formatProvider = null,
@@ -1418,6 +1426,10 @@ public class TimeSpanTest {
 		Assert.AreEqual ("1", ts.ToString ("%h"), "#E0");
 		Assert.AreEqual ("3", ts.ToString ("%m"), "#E1");
 		Assert.AreEqual ("5", ts.ToString ("%s"), "#E2");
+
+		ts = new TimeSpan (123456789);
+		Assert.AreEqual ("12.3", ts.ToString ("s\\.f"), "#F0");
+		Assert.AreEqual ("12.3", ts.ToString ("s\\.F"), "#F1");
 	}
 
 	[Test]
