@@ -6,13 +6,14 @@
 // Dual licensed under the terms of the MIT X11 or GNU GPL
 //
 // Copyright 2009 Novell, Inc
+// Copyright 2011 Xamarin Inc.
 //
 
 using System;
 using System.Linq;
 using SLE = System.Linq.Expressions;
 
-#if NET_4_0
+#if NET_4_0 || MOBILE_DYNAMIC
 using System.Dynamic;
 #endif
 
@@ -45,19 +46,13 @@ namespace Mono.CSharp
 		public DynamicTypeExpr (Location loc)
 		{
 			this.loc = loc;
+		}
 
-			type = InternalType.Dynamic;
+		public override TypeSpec ResolveAsType (IMemberContext ec, bool allowUnboundTypeArguments)
+		{
 			eclass = ExprClass.Type;
-		}
-
-		public override bool CheckAccessLevel (IMemberContext ds)
-		{
-			return true;
-		}
-
-		protected override TypeExpr DoResolveAsTypeStep (IMemberContext ec)
-		{
-			return this;
+			type = ec.Module.Compiler.BuiltinTypes.Dynamic;
+			return type;
 		}
 	}
 
@@ -68,7 +63,7 @@ namespace Mono.CSharp
 	//
 	public class RuntimeValueExpression : Expression, IDynamicAssign, IMemoryLocation
 	{
-#if !NET_4_0
+#if !NET_4_0 && !MOBILE_DYNAMIC
 		public class DynamicMetaObject
 		{
 			public TypeSpec RuntimeType;
@@ -101,9 +96,14 @@ namespace Mono.CSharp
 			throw new NotImplementedException ();
 		}
 
+		public override bool ContainsEmitWithAwait ()
+		{
+			throw new NotSupportedException ();
+		}
+
 		public override Expression CreateExpressionTree (ResolveContext ec)
 		{
-			throw new NotImplementedException ();
+			throw new NotSupportedException ();
 		}
 
 		protected override Expression DoResolve (ResolveContext ec)
@@ -128,7 +128,7 @@ namespace Mono.CSharp
 			throw new NotImplementedException ();
 		}
 
-		public void EmitAssign (EmitContext ec, Expression source, bool leave_copy, bool prepare_for_load)
+		public void EmitAssign (EmitContext ec, Expression source, bool leave_copy, bool isCompound)
 		{
 			throw new NotImplementedException ();
 		}
@@ -142,17 +142,22 @@ namespace Mono.CSharp
 
 		public override SLE.Expression MakeExpression (BuilderContext ctx)
 		{
-#if NET_4_0		
-			if (type.IsStruct && !obj.Expression.Type.IsValueType)
-				return SLE.Expression.Unbox (obj.Expression, type.GetMetaInfo ());
+#if STATIC
+			return base.MakeExpression (ctx);
+#else
 
-			if (obj.Expression.NodeType == SLE.ExpressionType.Parameter) {
-				if (((SLE.ParameterExpression) obj.Expression).IsByRef)
-					return obj.Expression;
-			}
+#if NET_4_0 || MOBILE_DYNAMIC
+				if (type.IsStruct && !obj.Expression.Type.IsValueType)
+					return SLE.Expression.Unbox (obj.Expression, type.GetMetaInfo ());
+
+				if (obj.Expression.NodeType == SLE.ExpressionType.Parameter) {
+					if (((SLE.ParameterExpression) obj.Expression).IsByRef)
+						return obj.Expression;
+				}
+	#endif
+
+				return SLE.Expression.Convert (obj.Expression, type.GetMetaInfo ());
 #endif
-
-			return SLE.Expression.Convert (obj.Expression, type.GetMetaInfo ());
 		}
 	}
 
@@ -176,10 +181,14 @@ namespace Mono.CSharp
 			return this;
 		}
 
-#if NET_4_0
+#if NET_4_0 || MOBILE_DYNAMIC
 		public override SLE.Expression MakeExpression (BuilderContext ctx)
 		{
+#if STATIC
+			return base.MakeExpression (ctx);
+#else
 			return SLE.Expression.Block (expr.MakeExpression (ctx), SLE.Expression.Default (type.GetMetaInfo ()));
+#endif
 		}
 #endif
 	}
@@ -215,21 +224,22 @@ namespace Mono.CSharp
 		//
 		protected class BinderFlags : EnumConstant
 		{
-			DynamicExpressionStatement statement;
-			CSharpBinderFlags flags;
+			readonly DynamicExpressionStatement statement;
+			readonly CSharpBinderFlags flags;
 
 			public BinderFlags (CSharpBinderFlags flags, DynamicExpressionStatement statement)
 				: base (statement.loc)
 			{
 				this.flags = flags;
 				this.statement = statement;
+				eclass = 0;
 			}
 
 			protected override Expression DoResolve (ResolveContext ec)
 			{
-				Child = new IntConstant ((int) (flags | statement.flags), statement.loc).Resolve (ec);
+				Child = new IntConstant (ec.BuiltinTypes, (int) (flags | statement.flags), statement.loc);
 
-				type = TypeManager.binder_flags;
+				type = ec.Module.PredefinedTypes.BinderFlags.Resolve ();
 				eclass = Child.eclass;
 				return this;
 			}
@@ -242,6 +252,9 @@ namespace Mono.CSharp
 		// Used by BinderFlags
 		protected CSharpBinderFlags flags;
 
+		TypeSpec binder_type;
+		TypeParameters context_mvars;
+
 		public DynamicExpressionStatement (IDynamicBinder binder, Arguments args, Location loc)
 		{
 			this.binder = binder;
@@ -255,10 +268,9 @@ namespace Mono.CSharp
 			}
 		}
 
-		FieldSpec CreateSiteField (EmitContext ec, FullNamedExpression type)
+		public override bool ContainsEmitWithAwait ()
 		{
-			var site_container = ec.CreateDynamicSite ();
-			return site_container.CreateCallSiteField (type, loc);
+			return arguments.ContainsEmitWithAwait ();
 		}
 
 		public override Expression CreateExpressionTree (ResolveContext ec)
@@ -267,42 +279,30 @@ namespace Mono.CSharp
 			return null;
 		}
 
-		protected override Expression DoResolve (ResolveContext ec)
+		protected override Expression DoResolve (ResolveContext rc)
 		{
-			if (DoResolveCore (ec))
-				binder_expr = binder.CreateCallSiteBinder (ec, arguments);
+			if (DoResolveCore (rc))
+				binder_expr = binder.CreateCallSiteBinder (rc, arguments);
 
 			return this;
 		}
 
 		protected bool DoResolveCore (ResolveContext rc)
 		{
+			if (rc.CurrentTypeParameters != null && rc.CurrentTypeParameters[0].IsMethodTypeParameter)
+				context_mvars = rc.CurrentTypeParameters;
+
 			int errors = rc.Report.Errors;
+			var pt = rc.Module.PredefinedTypes;
 
-			if (TypeManager.binder_type == null) {
-				var t = TypeManager.CoreLookupType (rc.Compiler,
-					"Microsoft.CSharp.RuntimeBinder", "Binder", MemberKind.Class, true);
-				if (t != null)
-					TypeManager.binder_type = new TypeExpression (t, Location.Null);
-			}
-
-			if (TypeManager.call_site_type == null)
-				TypeManager.call_site_type = TypeManager.CoreLookupType (rc.Compiler,
-					"System.Runtime.CompilerServices", "CallSite", MemberKind.Class, true);
-
-			if (TypeManager.generic_call_site_type == null)
-				TypeManager.generic_call_site_type = TypeManager.CoreLookupType (rc.Compiler,
-					"System.Runtime.CompilerServices", "CallSite", 1, MemberKind.Class, true);
-
-			if (TypeManager.binder_flags == null) {
-				TypeManager.binder_flags = TypeManager.CoreLookupType (rc.Compiler,
-					"Microsoft.CSharp.RuntimeBinder", "CSharpBinderFlags", MemberKind.Enum, true);
-			}
+			binder_type = pt.Binder.Resolve ();
+			pt.CallSite.Resolve ();
+			pt.CallSiteGeneric.Resolve ();
 
 			eclass = ExprClass.Value;
 
 			if (type == null)
-				type = InternalType.Dynamic;
+				type = rc.BuiltinTypes.Dynamic;
 
 			if (rc.Report.Errors == errors)
 				return true;
@@ -324,42 +324,225 @@ namespace Mono.CSharp
 
 		protected void EmitCall (EmitContext ec, Expression binder, Arguments arguments, bool isStatement)
 		{
+			//
+			// This method generates all internal infrastructure for a dynamic call. The
+			// reason why it's quite complicated is the mixture of dynamic and anonymous
+			// methods. Dynamic itself requires a temporary class (ContainerX) and anonymous
+			// methods can generate temporary storey as well (AnonStorey). Handling MVAR
+			// type parameters rewrite is non-trivial in such case as there are various
+			// combinations possible therefore the mutator is not straightforward. Secondly
+			// we need to keep both MVAR(possibly VAR for anon storey) and type VAR to emit
+			// correct Site field type and its access from EmitContext.
+			//
+
 			int dyn_args_count = arguments == null ? 0 : arguments.Count;
-			TypeExpr site_type = CreateSiteType (ec, arguments, dyn_args_count, isStatement);
+			int default_args = isStatement ? 1 : 2;
+			var module = ec.Module;
 
-			FieldExpr site_field_expr = new FieldExpr (CreateSiteField (ec, site_type), loc);
+			bool has_ref_out_argument = false;
+			var targs = new TypeExpression[dyn_args_count + default_args];
+			targs[0] = new TypeExpression (module.PredefinedTypes.CallSite.TypeSpec, loc);
 
-			SymbolWriter.OpenCompilerGeneratedBlock (ec);
+			TypeExpression[] targs_for_instance = null;
+			TypeParameterMutator mutator;
 
-			Arguments args = new Arguments (1);
-			args.Add (new Argument (binder));
-			StatementExpression s = new StatementExpression (new SimpleAssign (site_field_expr, new Invocation (new MemberAccess (site_type, "Create"), args)));
-			
-			BlockContext bc = new BlockContext (ec.MemberContext, null, TypeManager.void_type);		
-			if (s.Resolve (bc)) {
-				Statement init = new If (new Binary (Binary.Operator.Equality, site_field_expr, new NullLiteral (loc), loc), s, loc);
-				init.Emit (ec);
+			var site_container = ec.CreateDynamicSite ();
+
+			if (context_mvars != null) {
+				TypeParameters tparam;
+				TypeContainer sc = site_container;
+				do {
+					tparam = sc.CurrentTypeParameters;
+					sc = sc.Parent;
+				} while (tparam == null);
+
+				mutator = new TypeParameterMutator (context_mvars, tparam);
+
+				if (!ec.IsAnonymousStoreyMutateRequired) {
+					targs_for_instance = new TypeExpression[targs.Length];
+					targs_for_instance[0] = targs[0];
+				}
+			} else {
+				mutator = null;
 			}
 
-			args = new Arguments (1 + dyn_args_count);
-			args.Add (new Argument (site_field_expr));
-			if (arguments != null) {
-				foreach (Argument a in arguments) {
-					if (a is NamedArgument) {
-						// Name is not valid in this context
-						args.Add (new Argument (a.Expr, a.ArgType));
-						continue;
+			for (int i = 0; i < dyn_args_count; ++i) {
+				Argument a = arguments[i];
+				if (a.ArgType == Argument.AType.Out || a.ArgType == Argument.AType.Ref)
+					has_ref_out_argument = true;
+
+				var t = a.Type;
+
+				// Convert any internal type like dynamic or null to object
+				if (t.Kind == MemberKind.InternalCompilerType)
+					t = ec.BuiltinTypes.Object;
+
+				if (targs_for_instance != null)
+					targs_for_instance[i + 1] = new TypeExpression (t, loc);
+
+				if (mutator != null)
+					t = t.Mutate (mutator);
+
+				targs[i + 1] = new TypeExpression (t, loc);
+			}
+
+			TypeExpr del_type = null;
+			TypeExpr del_type_instance_access = null;
+			if (!has_ref_out_argument) {
+				string d_name = isStatement ? "Action" : "Func";
+
+				TypeSpec te = null;
+				Namespace type_ns = module.GlobalRootNamespace.GetNamespace ("System", true);
+				if (type_ns != null) {
+					te = type_ns.LookupType (module, d_name, dyn_args_count + default_args, LookupMode.Normal, loc);
+				}
+
+				if (te != null) {
+					if (!isStatement) {
+						var t = type;
+						if (t.Kind == MemberKind.InternalCompilerType)
+							t = ec.BuiltinTypes.Object;
+
+						if (targs_for_instance != null)
+							targs_for_instance[targs_for_instance.Length - 1] = new TypeExpression (t, loc);
+
+						if (mutator != null)
+							t = t.Mutate (mutator);
+
+						targs[targs.Length - 1] = new TypeExpression (t, loc);
 					}
 
-					args.Add (a);
+					del_type = new GenericTypeExpr (te, new TypeArguments (targs), loc);
+					if (targs_for_instance != null)
+						del_type_instance_access = new GenericTypeExpr (te, new TypeArguments (targs_for_instance), loc);
+					else
+						del_type_instance_access = del_type;
 				}
 			}
 
-			Expression target = new DelegateInvocation (new MemberAccess (site_field_expr, "Target", loc).Resolve (bc), args, loc).Resolve (bc);
-			if (target != null)
-				target.Emit (ec);
+			//
+			// Create custom delegate when no appropriate predefined delegate has been found
+			//
+			Delegate d;
+			if (del_type == null) {
+				TypeSpec rt = isStatement ? ec.BuiltinTypes.Void : type;
+				Parameter[] p = new Parameter[dyn_args_count + 1];
+				p[0] = new Parameter (targs[0], "p0", Parameter.Modifier.NONE, null, loc);
 
-			SymbolWriter.CloseCompilerGeneratedBlock (ec);
+				var site = ec.CreateDynamicSite ();
+				int index = site.Containers == null ? 0 : site.Containers.Count;
+
+				if (mutator != null)
+					rt = mutator.Mutate (rt);
+
+				for (int i = 1; i < dyn_args_count + 1; ++i) {
+					p[i] = new Parameter (targs[i], "p" + i.ToString ("X"), arguments[i - 1].Modifier, null, loc);
+				}
+
+				d = new Delegate (site, new TypeExpression (rt, loc),
+					Modifiers.INTERNAL | Modifiers.COMPILER_GENERATED,
+					new MemberName ("Container" + index.ToString ("X")),
+					new ParametersCompiled (p), null);
+
+				d.CreateContainer ();
+				d.DefineContainer ();
+				d.Define ();
+				d.PrepareEmit ();
+
+				site.AddTypeContainer (d);
+
+				//
+				// Add new container to inflated site container when the
+				// member cache already exists
+				//
+				if (site.CurrentType is InflatedTypeSpec && index > 0)
+					site.CurrentType.MemberCache.AddMember (d.CurrentType);
+
+				del_type = new TypeExpression (d.CurrentType, loc);
+				if (targs_for_instance != null) {
+					del_type_instance_access = null;
+				} else {
+					del_type_instance_access = del_type;
+				}
+			} else {
+				d = null;
+			}
+
+			var site_type_decl = new GenericTypeExpr (module.PredefinedTypes.CallSiteGeneric.TypeSpec, new TypeArguments (del_type), loc);
+			var field = site_container.CreateCallSiteField (site_type_decl, loc);
+			if (field == null)
+				return;
+
+			if (del_type_instance_access == null) {
+				var dt = d.CurrentType.DeclaringType.MakeGenericType (module, context_mvars.Types);
+				del_type_instance_access = new TypeExpression (MemberCache.GetMember (dt, d.CurrentType), loc);
+			}
+
+			var instanceAccessExprType = new GenericTypeExpr (module.PredefinedTypes.CallSiteGeneric.TypeSpec,
+				new TypeArguments (del_type_instance_access), loc);
+
+			if (instanceAccessExprType.ResolveAsType (ec.MemberContext) == null)
+				return;
+
+			bool inflate_using_mvar = context_mvars != null && ec.IsAnonymousStoreyMutateRequired;
+
+			TypeSpec gt;
+			if (inflate_using_mvar || context_mvars == null) {
+				gt = site_container.CurrentType;
+			} else {
+				gt = site_container.CurrentType.MakeGenericType (module, context_mvars.Types);
+			}
+
+			// When site container already exists the inflated version has to be
+			// updated manually to contain newly created field
+			if (gt is InflatedTypeSpec && site_container.AnonymousMethodsCounter > 1) {
+				var tparams = gt.MemberDefinition.TypeParametersCount > 0 ? gt.MemberDefinition.TypeParameters : TypeParameterSpec.EmptyTypes;
+				var inflator = new TypeParameterInflator (module, gt, tparams, gt.TypeArguments);
+				gt.MemberCache.AddMember (field.InflateMember (inflator));
+			}
+
+			FieldExpr site_field_expr = new FieldExpr (MemberCache.GetMember (gt, field), loc);
+
+			BlockContext bc = new BlockContext (ec.MemberContext, null, ec.BuiltinTypes.Void);
+
+			Arguments args = new Arguments (1);
+			args.Add (new Argument (binder));
+			StatementExpression s = new StatementExpression (new SimpleAssign (site_field_expr, new Invocation (new MemberAccess (instanceAccessExprType, "Create"), args)));
+
+			using (ec.With (BuilderContext.Options.OmitDebugInfo, true)) {
+				if (s.Resolve (bc)) {
+					Statement init = new If (new Binary (Binary.Operator.Equality, site_field_expr, new NullLiteral (loc)), s, loc);
+					init.Emit (ec);
+				}
+
+				args = new Arguments (1 + dyn_args_count);
+				args.Add (new Argument (site_field_expr));
+				if (arguments != null) {
+					int arg_pos = 1;
+					foreach (Argument a in arguments) {
+						if (a is NamedArgument) {
+							// Name is not valid in this context
+							args.Add (new Argument (a.Expr, a.ArgType));
+						} else {
+							args.Add (a);
+						}
+
+						if (inflate_using_mvar && a.Type != targs[arg_pos].Type)
+							a.Expr.Type = targs[arg_pos].Type;
+
+						++arg_pos;
+					}
+				}
+
+				Expression target = new DelegateInvocation (new MemberAccess (site_field_expr, "Target", loc).Resolve (bc), args, false, loc).Resolve (bc);
+				if (target != null)
+					target.Emit (ec);
+			}
+		}
+
+		public override void FlowAnalysis (FlowAnalysisContext fc)
+		{
+			arguments.FlowAnalysis (fc);
 		}
 
 		public static MemberAccess GetBinderNamespace (Location loc)
@@ -368,109 +551,84 @@ namespace Mono.CSharp
 				new QualifiedAliasMember (QualifiedAliasMember.GlobalAlias, "Microsoft", loc), "CSharp", loc), "RuntimeBinder", loc);
 		}
 
-		public static MemberAccess GetBinder (string name, Location loc)
+		protected MemberAccess GetBinder (string name, Location loc)
 		{
-			return new MemberAccess (TypeManager.binder_type, name, loc);
-		}
-
-		TypeExpr CreateSiteType (EmitContext ec, Arguments arguments, int dyn_args_count, bool is_statement)
-		{
-			int default_args = is_statement ? 1 : 2;
-
-			bool has_ref_out_argument = false;
-			FullNamedExpression[] targs = new FullNamedExpression[dyn_args_count + default_args];
-			targs [0] = new TypeExpression (TypeManager.call_site_type, loc);
-			for (int i = 0; i < dyn_args_count; ++i) {
-				Argument a = arguments [i];
-				if (a.ArgType == Argument.AType.Out || a.ArgType == Argument.AType.Ref)
-					has_ref_out_argument = true;
-
-				targs [i + 1] = new TypeExpression (a.Type, loc);
-			}
-
-			TypeExpr del_type = null;
-			if (!has_ref_out_argument) {
-				string d_name = is_statement ? "Action" : "Func";
-
-				TypeSpec t = TypeManager.CoreLookupType (ec.MemberContext.Compiler, "System", d_name, dyn_args_count + default_args, MemberKind.Delegate, false);
-				if (t != null) {
-					if (!is_statement)
-						targs [targs.Length - 1] = new TypeExpression (type, loc);
-
-					del_type = new GenericTypeExpr (t, new TypeArguments (targs), loc);
-				}
-			}
-
-			//
-			// Create custom delegate when no appropriate predefined one is found
-			//
-			if (del_type == null) {
-				TypeSpec rt = is_statement ? TypeManager.void_type : type;
-				Parameter[] p = new Parameter [dyn_args_count + 1];
-				p[0] = new Parameter (targs [0], "p0", Parameter.Modifier.NONE, null, loc);
-
-				for (int i = 1; i < dyn_args_count + 1; ++i)
-					p[i] = new Parameter (targs[i], "p" + i.ToString ("X"), arguments[i - 1].Modifier, null, loc);
-
-				TypeContainer site = ec.CreateDynamicSite ();
-				int index = site.Types == null ? 0 : site.Types.Count;
-
-				Delegate d = new Delegate (site.NamespaceEntry, site, new TypeExpression (rt, loc),
-					Modifiers.INTERNAL | Modifiers.COMPILER_GENERATED,
-					new MemberName ("Container" + index.ToString ("X")),
-					new ParametersCompiled (null, p), null);
-
-				d.CreateType ();
-				d.DefineType ();
-				d.Define ();
-				d.Emit ();
-
-				site.AddDelegate (d);
-				del_type = new TypeExpression (d.Definition, loc);
-			}
-
-			TypeExpr site_type = new GenericTypeExpr (TypeManager.generic_call_site_type, new TypeArguments (del_type), loc);
-			return site_type;
+			return new MemberAccess (new TypeExpression (binder_type, loc), name, loc);
 		}
 	}
 
 	//
 	// Dynamic member access compound assignment for events
 	//
-	class DynamicEventCompoundAssign : DynamicExpressionStatement, IDynamicBinder
+	class DynamicEventCompoundAssign : ExpressionStatement
 	{
-		string name;
-		Statement condition;
-
-		public DynamicEventCompoundAssign (string name, Arguments args, ExpressionStatement assignment, ExpressionStatement invoke, Location loc)
-			: base (null, args, loc)
+		class IsEvent : DynamicExpressionStatement, IDynamicBinder
 		{
-			this.name = name;
-			base.binder = this;
+			string name;
 
-			// Used by += or -= only
-			type = TypeManager.bool_type;
+			public IsEvent (string name, Arguments args, Location loc)
+				: base (null, args, loc)
+			{
+				this.name = name;
+				binder = this;
+			}
 
-			condition = new If (
-				new Binary (Binary.Operator.Equality, this, new BoolLiteral (true, loc), loc),
-				new StatementExpression (invoke), new StatementExpression (assignment),
-				loc);
+			public Expression CreateCallSiteBinder (ResolveContext ec, Arguments args)
+			{
+				type = ec.BuiltinTypes.Bool;
+
+				Arguments binder_args = new Arguments (3);
+
+				binder_args.Add (new Argument (new BinderFlags (0, this)));
+				binder_args.Add (new Argument (new StringLiteral (ec.BuiltinTypes, name, loc)));
+				binder_args.Add (new Argument (new TypeOf (ec.CurrentType, loc)));
+
+				return new Invocation (GetBinder ("IsEvent", loc), binder_args);
+			}
 		}
 
-		public Expression CreateCallSiteBinder (ResolveContext ec, Arguments args)
+		Expression condition;
+		ExpressionStatement invoke, assign;
+
+		public DynamicEventCompoundAssign (string name, Arguments args, ExpressionStatement assignment, ExpressionStatement invoke, Location loc)
 		{
-			Arguments binder_args = new Arguments (3);
+			condition = new IsEvent (name, args, loc);
+			this.invoke = invoke;
+			this.assign = assignment;
+			this.loc = loc;
+		}
 
-			binder_args.Add (new Argument (new BinderFlags (0, this)));
-			binder_args.Add (new Argument (new StringLiteral (name, loc)));
-			binder_args.Add (new Argument (new TypeOf (new TypeExpression (ec.CurrentType, loc), loc)));
+		public override Expression CreateExpressionTree (ResolveContext ec)
+		{
+			return condition.CreateExpressionTree (ec);
+		}
 
-			return new Invocation (GetBinder ("IsEvent", loc), binder_args);
+		protected override Expression DoResolve (ResolveContext rc)
+		{
+			type = rc.BuiltinTypes.Dynamic;
+			eclass = ExprClass.Value;
+			condition = condition.Resolve (rc);
+			return this;
+		}
+
+		public override void Emit (EmitContext ec)
+		{
+			var rc = new ResolveContext (ec.MemberContext);
+			var expr = new Conditional (new BooleanExpression (condition), invoke, assign, loc).Resolve (rc);
+			expr.Emit (ec);
 		}
 
 		public override void EmitStatement (EmitContext ec)
 		{
-			condition.Emit (ec);
+			var stmt = new If (condition, new StatementExpression (invoke), new StatementExpression (assign), loc);
+			using (ec.With (BuilderContext.Options.OmitDebugInfo, true)) {
+				stmt.Emit (ec);
+			}
+		}
+
+		public override void FlowAnalysis (FlowAnalysisContext fc)
+		{
+			invoke.FlowAnalysis (fc);
 		}
 	}
 
@@ -491,8 +649,8 @@ namespace Mono.CSharp
 			flags |= ec.HasSet (ResolveContext.Options.CheckedScope) ? CSharpBinderFlags.CheckedContext : 0;
 
 			binder_args.Add (new Argument (new BinderFlags (flags, this)));
-			binder_args.Add (new Argument (new TypeOf (new TypeExpression (type, loc), loc)));
-			binder_args.Add (new Argument (new TypeOf (new TypeExpression (ec.CurrentType, loc), loc)));
+			binder_args.Add (new Argument (new TypeOf (type, loc)));
+			binder_args.Add (new Argument (new TypeOf (ec.CurrentType, loc)));
 			return new Invocation (GetBinder ("Convert", loc), binder_args);
 		}
 	}
@@ -511,7 +669,7 @@ namespace Mono.CSharp
 			Arguments binder_args = new Arguments (3);
 
 			binder_args.Add (new Argument (new BinderFlags (0, this)));
-			binder_args.Add (new Argument (new TypeOf (new TypeExpression (ec.CurrentType, loc), loc)));
+			binder_args.Add (new Argument (new TypeOf (ec.CurrentType, loc)));
 			binder_args.Add (new Argument (new ImplicitlyTypedArrayCreation (args.CreateDynamicBinderArguments (ec), loc)));
 
 			return new Invocation (GetBinder ("InvokeConstructor", loc), binder_args);
@@ -520,26 +678,73 @@ namespace Mono.CSharp
 
 	class DynamicIndexBinder : DynamicMemberAssignable
 	{
+		bool can_be_mutator;
+
 		public DynamicIndexBinder (Arguments args, Location loc)
 			: base (args, loc)
 		{
 		}
 
+		public DynamicIndexBinder (CSharpBinderFlags flags, Arguments args, Location loc)
+			: this (args, loc)
+		{
+			base.flags = flags;
+		}
+
+		protected override Expression DoResolve (ResolveContext ec)
+		{
+			can_be_mutator = true;
+			return base.DoResolve (ec);
+		}
+
 		protected override Expression CreateCallSiteBinder (ResolveContext ec, Arguments args, bool isSet)
 		{
 			Arguments binder_args = new Arguments (3);
 
-			binder_args.Add (new Argument (new BinderFlags (0, this)));
-			binder_args.Add (new Argument (new TypeOf (new TypeExpression (ec.CurrentType, loc), loc)));
+			binder_args.Add (new Argument (new BinderFlags (flags, this)));
+			binder_args.Add (new Argument (new TypeOf (ec.CurrentType, loc)));
 			binder_args.Add (new Argument (new ImplicitlyTypedArrayCreation (args.CreateDynamicBinderArguments (ec), loc)));
 
+			isSet |= (flags & CSharpBinderFlags.ValueFromCompoundAssignment) != 0;
 			return new Invocation (GetBinder (isSet ? "SetIndex" : "GetIndex", loc), binder_args);
 		}
+
+		protected override Arguments CreateSetterArguments (ResolveContext rc, Expression rhs)
+		{
+			//
+			// Indexer has arguments which complicates things as the setter and getter
+			// are called in two steps when unary mutator is used. We have to make a
+			// copy of all variable arguments to not duplicate any side effect.
+			//
+			// ++d[++arg, Foo ()]
+			//
+
+			if (!can_be_mutator)
+				return base.CreateSetterArguments (rc, rhs);
+
+			var setter_args = new Arguments (Arguments.Count + 1);
+			for (int i = 0; i < Arguments.Count; ++i) {
+				var expr = Arguments[i].Expr;
+
+				if (expr is Constant || expr is VariableReference || expr is This) {
+					setter_args.Add (Arguments [i]);
+					continue;
+				}
+
+				LocalVariable temp = LocalVariable.CreateCompilerGenerated (expr.Type, rc.CurrentBlock, loc);
+				expr = new SimpleAssign (temp.CreateReferenceExpression (rc, expr.Location), expr).Resolve (rc);
+				Arguments[i].Expr = temp.CreateReferenceExpression (rc, expr.Location).Resolve (rc);
+				setter_args.Add (Arguments [i].Clone (expr));
+			}
+
+			setter_args.Add (new Argument (rhs));
+			return setter_args;
+		}
 	}
 
 	class DynamicInvocation : DynamicExpressionStatement, IDynamicBinder
 	{
-		ATypeNameExpression member;
+		readonly ATypeNameExpression member;
 
 		public DynamicInvocation (ATypeNameExpression member, Arguments args, Location loc)
 			: base (null, args, loc)
@@ -548,15 +753,6 @@ namespace Mono.CSharp
 			this.member = member;
 		}
 
-		//
-		// When a return type is known not to be dynamic
-		//
-		public DynamicInvocation (ATypeNameExpression member, Arguments args, TypeSpec type, Location loc)
-			: this (member, args, loc)
-		{
-			this.type = type;
-		}
-
 		public static DynamicInvocation CreateSpecialNameInvoke (ATypeNameExpression member, Arguments args, Location loc)
 		{
 			return new DynamicInvocation (member, args, loc) {
@@ -580,14 +776,14 @@ namespace Mono.CSharp
 			binder_args.Add (new Argument (new BinderFlags (call_flags, this)));
 
 			if (is_member_access)
-				binder_args.Add (new Argument (new StringLiteral (member.Name, member.Location)));
+				binder_args.Add (new Argument (new StringLiteral (ec.BuiltinTypes, member.Name, member.Location)));
 
 			if (member != null && member.HasTypeArguments) {
 				TypeArguments ta = member.TypeArguments;
 				if (ta.Resolve (ec)) {
 					var targs = new ArrayInitializer (ta.Count, loc);
 					foreach (TypeSpec t in ta.Arguments)
-						targs.Add (new TypeOf (new TypeExpression (t, loc), loc));
+						targs.Add (new TypeOf (t, loc));
 
 					binder_args.Add (new Argument (new ImplicitlyTypedArrayCreation (targs, loc)));
 				}
@@ -595,7 +791,7 @@ namespace Mono.CSharp
 				binder_args.Add (new Argument (new NullLiteral (loc)));
 			}
 
-			binder_args.Add (new Argument (new TypeOf (new TypeExpression (ec.CurrentType, loc), loc)));
+			binder_args.Add (new Argument (new TypeOf (ec.CurrentType, loc)));
 
 			Expression real_args;
 			if (args == null) {
@@ -629,15 +825,22 @@ namespace Mono.CSharp
 			this.name = name;
 		}
 
+		public DynamicMemberBinder (string name, CSharpBinderFlags flags, Arguments args, Location loc)
+			: this (name, args, loc)
+		{
+			base.flags = flags;
+		}
+
 		protected override Expression CreateCallSiteBinder (ResolveContext ec, Arguments args, bool isSet)
 		{
 			Arguments binder_args = new Arguments (4);
 
-			binder_args.Add (new Argument (new BinderFlags (0, this)));
-			binder_args.Add (new Argument (new StringLiteral (name, loc)));
-			binder_args.Add (new Argument (new TypeOf (new TypeExpression (ec.CurrentType, loc), loc)));
+			binder_args.Add (new Argument (new BinderFlags (flags, this)));
+			binder_args.Add (new Argument (new StringLiteral (ec.BuiltinTypes, name, loc)));
+			binder_args.Add (new Argument (new TypeOf (ec.CurrentType, loc)));
 			binder_args.Add (new Argument (new ImplicitlyTypedArrayCreation (args.CreateDynamicBinderArguments (ec), loc)));
 
+			isSet |= (flags & CSharpBinderFlags.ValueFromCompoundAssignment) != 0;
 			return new Invocation (GetBinder (isSet ? "SetMember" : "GetMember", loc), binder_args);
 		}
 	}
@@ -666,17 +869,23 @@ namespace Mono.CSharp
 
 		protected abstract Expression CreateCallSiteBinder (ResolveContext ec, Arguments args, bool isSet);
 
+		protected virtual Arguments CreateSetterArguments (ResolveContext rc, Expression rhs)
+		{
+			var setter_args = new Arguments (Arguments.Count + 1);
+			setter_args.AddRange (Arguments);
+			setter_args.Add (new Argument (rhs));
+			return setter_args;
+		}
+
 		public override Expression DoResolveLValue (ResolveContext rc, Expression right_side)
 		{
-			if (right_side == EmptyExpression.OutAccess.Instance) {
+			if (right_side == EmptyExpression.OutAccess) {
 				right_side.DoResolveLValue (rc, this);
 				return null;
 			}
 
 			if (DoResolveCore (rc)) {
-				setter_args = new Arguments (Arguments.Count + 1);
-				setter_args.AddRange (Arguments);
-				setter_args.Add (new Argument (right_side));
+				setter_args = CreateSetterArguments (rc, right_side);
 				setter = CreateCallSiteBinder (rc, setter_args, true);
 			}
 
@@ -709,7 +918,7 @@ namespace Mono.CSharp
 			throw new NotImplementedException ();
 		}
 
-		public void EmitAssign (EmitContext ec, Expression source, bool leave_copy, bool prepare_for_load)
+		public void EmitAssign (EmitContext ec, Expression source, bool leave_copy, bool isCompound)
 		{
 			EmitCall (ec, setter, setter_args, !leave_copy);
 		}
@@ -719,15 +928,23 @@ namespace Mono.CSharp
 
 	class DynamicUnaryConversion : DynamicExpressionStatement, IDynamicBinder
 	{
-		string name;
+		readonly string name;
 
 		public DynamicUnaryConversion (string name, Arguments args, Location loc)
 			: base (null, args, loc)
 		{
 			this.name = name;
 			base.binder = this;
-			if (name == "IsTrue" || name == "IsFalse")
-				type = TypeManager.bool_type;
+		}
+
+		public static DynamicUnaryConversion CreateIsTrue (ResolveContext rc, Arguments args, Location loc)
+		{
+			return new DynamicUnaryConversion ("IsTrue", args, loc) { type = rc.BuiltinTypes.Bool };
+		}
+
+		public static DynamicUnaryConversion CreateIsFalse (ResolveContext rc, Arguments args, Location loc)
+		{
+			return new DynamicUnaryConversion ("IsFalse", args, loc) { type = rc.BuiltinTypes.Bool };
 		}
 
 		public Expression CreateCallSiteBinder (ResolveContext ec, Arguments args)
@@ -741,56 +958,29 @@ namespace Mono.CSharp
 
 			binder_args.Add (new Argument (new BinderFlags (flags, this)));
 			binder_args.Add (new Argument (new MemberAccess (new MemberAccess (sle, "ExpressionType", loc), name, loc)));
-			binder_args.Add (new Argument (new TypeOf (new TypeExpression (ec.CurrentType, loc), loc)));
+			binder_args.Add (new Argument (new TypeOf (ec.CurrentType, loc)));
 			binder_args.Add (new Argument (new ImplicitlyTypedArrayCreation (args.CreateDynamicBinderArguments (ec), loc)));
 
 			return new Invocation (GetBinder ("UnaryOperation", loc), binder_args);
 		}
 	}
 
-	public class DynamicSiteClass : HoistedStoreyClass
+	sealed class DynamicSiteClass : HoistedStoreyClass
 	{
-		//
-		// Holds the type to access the site. It gets inflated
-		// by MVARs for generic call sites
-		//
-		TypeSpec instance_type;
-
-		public DynamicSiteClass (TypeContainer parent, MemberBase host, TypeParameter[] tparams)
-			: base (parent, MakeMemberName (host, "DynamicSite", 0, tparams, Location.Null), tparams, Modifiers.STATIC)
+		public DynamicSiteClass (TypeDefinition parent, MemberBase host, TypeParameters tparams)
+			: base (parent, MakeMemberName (host, "DynamicSite", parent.DynamicSitesCounter, tparams, Location.Null), tparams, Modifiers.STATIC, MemberKind.Class)
 		{
-			if (tparams != null) {
-				mutator = new TypeParameterMutator (tparams, CurrentTypeParameters);
-			}
+			parent.DynamicSitesCounter++;
 		}
 
 		public FieldSpec CreateCallSiteField (FullNamedExpression type, Location loc)
 		{
-			int index = fields == null ? 0 : fields.Count;
+			int index = AnonymousMethodsCounter++;
 			Field f = new HoistedField (this, type, Modifiers.PUBLIC | Modifiers.STATIC, "Site" + index.ToString ("X"), null, loc);
 			f.Define ();
 
 			AddField (f);
-
-			var fs = f.Spec;
-			if (mutator != null) {
-				//
-				// Inflate the field, no need to keep it in MemberCache as it's accessed only once
-				//
-				var inflator = new TypeParameterInflator (instance_type, spec.MemberDefinition.TypeParameters, instance_type.TypeArguments);
-				fs = (FieldSpec) fs.InflateMember (inflator);
-			}
-
-			return fs;
-		}
-
-		protected override bool DoResolveTypeParameters ()
-		{
-			instance_type = spec;
-			if (mutator != null)
-				instance_type = instance_type.MakeGenericType (mutator.MethodTypeParameters.Select (l => l.Type).ToArray ());
-
-			return true;
+			return f.Spec;
 		}
 	}
 }
