@@ -19,15 +19,18 @@
 #include <mono/metadata/threadpool-internals.h>
 #include <mono/metadata/exception.h>
 #include <mono/metadata/environment.h>
+#include <mono/metadata/mono-config.h>
 #include <mono/metadata/mono-mlist.h>
 #include <mono/metadata/mono-perfcounters.h>
 #include <mono/metadata/socket-io.h>
 #include <mono/metadata/mono-cq.h>
 #include <mono/metadata/mono-wsq.h>
+#include <mono/metadata/mono-ptr-array.h>
 #include <mono/io-layer/io-layer.h>
 #include <mono/utils/mono-time.h>
 #include <mono/utils/mono-proclib.h>
 #include <mono/utils/mono-semaphore.h>
+#include <mono/utils/atomic.h>
 #include <errno.h>
 #ifdef HAVE_SYS_TIME_H
 #include <sys/time.h>
@@ -81,13 +84,13 @@ enum {
 };
 
 typedef struct {
-	CRITICAL_SECTION io_lock; /* access to sock_to_state */
+	mono_mutex_t io_lock; /* access to sock_to_state */
 	int inited; // 0 -> not initialized , 1->initializing, 2->initialized, 3->cleaned up
 	MonoGHashTable *sock_to_state;
 
 	gint event_system;
 	gpointer event_data;
-	void (*modify) (gpointer event_data, int fd, int operation, int events, gboolean is_new);
+	void (*modify) (gpointer p, int fd, int operation, int events, gboolean is_new);
 	void (*wait) (gpointer sock_data);
 	void (*shutdown) (gpointer event_data);
 } SocketIOData;
@@ -150,13 +153,15 @@ static void socket_io_cleanup (SocketIOData *data);
 static MonoObject *get_io_event (MonoMList **list, gint event);
 static int get_events_from_list (MonoMList *list);
 static int get_event_from_state (MonoSocketAsyncResult *state);
+static void check_for_interruption_critical (void);
 
 static MonoClass *async_call_klass;
 static MonoClass *socket_async_call_klass;
 static MonoClass *process_async_call_klass;
 
 static GPtrArray *wsqs;
-CRITICAL_SECTION wsqs_lock;
+mono_mutex_t wsqs_lock;
+static gboolean suspended;
 
 /* Hooks */
 static MonoThreadPoolFunc tp_start_func;
@@ -188,7 +193,7 @@ enum {
 #include <mono/metadata/tpool-poll.c>
 #ifdef HAVE_EPOLL
 #include <mono/metadata/tpool-epoll.c>
-#elif defined(HAVE_KQUEUE)
+#elif defined(USE_KQUEUE_FOR_THREADPOOL)
 #include <mono/metadata/tpool-kqueue.c>
 #endif
 /*
@@ -214,30 +219,11 @@ is_corlib_type (MonoDomain *domain, MonoClass *klass)
 /*
  * Note that we call it is_socket_type() where 'socket' refers to the image
  * that contains the System.Net.Sockets.Socket type.
- * For moonlight there is a System.Net.Sockets.Socket class in both System.dll and System.Net.dll.
 */
 static gboolean
 is_socket_type (MonoDomain *domain, MonoClass *klass)
 {
-	static const char *version = NULL;
-	static gboolean moonlight;
-
-	if (is_system_type (domain, klass))
-		return TRUE;
-
-	/* If moonlight, check if the type is in System.Net.dll too */
-	if (version == NULL) {
-		version = mono_get_runtime_info ()->framework_version;
-		moonlight = !strcmp (version, "2.1");
-	}
-
-	if (!moonlight)
-		return FALSE;
-
-	if (domain->system_net_dll == NULL)
-		domain->system_net_dll = mono_image_loaded ("System.Net");
-	
-	return klass->image == domain->system_net_dll;
+	return is_system_type (domain, klass);
 }
 
 #define check_type_cached(domain, ASSEMBLY, _class, _namespace, _name, loc) do { \
@@ -328,14 +314,14 @@ get_events_from_list (MonoMList *list)
 static void
 socket_io_cleanup (SocketIOData *data)
 {
-	EnterCriticalSection (&data->io_lock);
+	mono_mutex_lock (&data->io_lock);
 	if (data->inited != 2) {
-		LeaveCriticalSection (&data->io_lock);
+		mono_mutex_unlock (&data->io_lock);
 		return;
 	}
 	data->inited = 3;
 	data->shutdown (data->event_data);
-	LeaveCriticalSection (&data->io_lock);
+	mono_mutex_unlock (&data->io_lock);
 }
 
 static int
@@ -397,8 +383,14 @@ threadpool_jobs_inc (MonoObject *obj)
 static gboolean
 threadpool_jobs_dec (MonoObject *obj)
 {
-	MonoDomain *domain = obj->vtable->domain;
-	int remaining_jobs = InterlockedDecrement (&domain->threadpool_jobs);
+	MonoDomain *domain;
+	int remaining_jobs;
+
+	if (obj == NULL)
+		return FALSE;
+
+	domain = obj->vtable->domain;
+	remaining_jobs = InterlockedDecrement (&domain->threadpool_jobs);
 	if (remaining_jobs == 0 && domain->cleanup_semaphore) {
 		ReleaseSemaphore (domain->cleanup_semaphore, 1, NULL);
 		return TRUE;
@@ -451,15 +443,15 @@ mono_thread_pool_remove_socket (int sock)
 	if (socket_io_data.inited == 0)
 		return;
 
-	EnterCriticalSection (&socket_io_data.io_lock);
+	mono_mutex_lock (&socket_io_data.io_lock);
 	if (socket_io_data.sock_to_state == NULL) {
-		LeaveCriticalSection (&socket_io_data.io_lock);
+		mono_mutex_unlock (&socket_io_data.io_lock);
 		return;
 	}
 	list = mono_g_hash_table_lookup (socket_io_data.sock_to_state, GINT_TO_POINTER (sock));
 	if (list)
 		mono_g_hash_table_remove (socket_io_data.sock_to_state, GINT_TO_POINTER (sock));
-	LeaveCriticalSection (&socket_io_data.io_lock);
+	mono_mutex_unlock (&socket_io_data.io_lock);
 	
 	while (list) {
 		state = (MonoSocketAsyncResult *) mono_mlist_get_data (list);
@@ -481,9 +473,15 @@ static void
 init_event_system (SocketIOData *data)
 {
 #ifdef HAVE_EPOLL
-	if (data->event_system == EPOLL_BACKEND)
+	if (data->event_system == EPOLL_BACKEND) {
 		data->event_data = tp_epoll_init (data);
-#elif defined(HAVE_KQUEUE)
+		if (data->event_data == NULL) {
+			if (g_getenv ("MONO_DEBUG"))
+				g_message ("Falling back to poll()");
+			data->event_system = POLL_BACKEND;
+		}
+	}
+#elif defined(USE_KQUEUE_FOR_THREADPOOL)
 	if (data->event_system == KQUEUE_BACKEND)
 		data->event_data = tp_kqueue_init (data);
 #endif
@@ -508,11 +506,11 @@ socket_io_init (SocketIOData *data)
 		}
 	}
 
-	EnterCriticalSection (&data->io_lock);
+	mono_mutex_lock (&data->io_lock);
 	data->sock_to_state = mono_g_hash_table_new_type (g_direct_hash, g_direct_equal, MONO_HASH_VALUE_GC);
 #ifdef HAVE_EPOLL
 	data->event_system = EPOLL_BACKEND;
-#elif defined(HAVE_KQUEUE)
+#elif defined(USE_KQUEUE_FOR_THREADPOOL)
 	data->event_system = KQUEUE_BACKEND;
 #else
 	data->event_system = POLL_BACKEND;
@@ -522,7 +520,7 @@ socket_io_init (SocketIOData *data)
 
 	init_event_system (data);
 	mono_thread_create_internal (mono_get_root_domain (), data->wait, data, TRUE, SMALL_STACK);
-	LeaveCriticalSection (&data->io_lock);
+	mono_mutex_unlock (&data->io_lock);
 	data->inited = 2;
 	threadpool_start_thread (&async_io_tp);
 }
@@ -545,9 +543,9 @@ socket_io_add (MonoAsyncResult *ares, MonoSocketAsyncResult *state)
 	MONO_OBJECT_SETREF (state, ares, ares);
 
 	fd = GPOINTER_TO_INT (state->handle);
-	EnterCriticalSection (&data->io_lock);
+	mono_mutex_lock (&data->io_lock);
 	if (data->sock_to_state == NULL) {
-		LeaveCriticalSection (&data->io_lock);
+		mono_mutex_unlock (&data->io_lock);
 		return;
 	}
 	list = mono_g_hash_table_lookup (data->sock_to_state, GINT_TO_POINTER (fd));
@@ -561,8 +559,8 @@ socket_io_add (MonoAsyncResult *ares, MonoSocketAsyncResult *state)
 
 	mono_g_hash_table_replace (data->sock_to_state, state->handle, list);
 	ievt = get_events_from_list (list);
-	LeaveCriticalSection (&data->io_lock);
-	data->modify (data->event_data, fd, state->operation, ievt, is_new);
+	/* The modify function leaves the io_lock critical section. */
+	data->modify (data, fd, state->operation, ievt, is_new);
 }
 
 #ifndef DISABLE_SOCKETS
@@ -605,6 +603,7 @@ mono_async_invoke (ThreadPool *tp, MonoAsyncResult *ares)
 	MonoObject *res, *exc = NULL;
 	MonoArray *out_args = NULL;
 	HANDLE wait_event = NULL;
+	MonoInternalThread *thread = mono_thread_internal_current ();
 
 	if (ares->execution_context) {
 		/* use captured ExecutionContext (if available) */
@@ -617,7 +616,10 @@ mono_async_invoke (ThreadPool *tp, MonoAsyncResult *ares)
 	if (ac == NULL) {
 		/* Fast path from ThreadPool.*QueueUserWorkItem */
 		void *pa = ares->async_state;
-		mono_runtime_delegate_invoke (ares->async_delegate, &pa, &exc);
+		/* The debugger needs this */
+		thread->async_invoke_method = ((MonoDelegate*)ares->async_delegate)->method;
+		res = mono_runtime_delegate_invoke (ares->async_delegate, &pa, &exc);
+		thread->async_invoke_method = NULL;
 	} else {
 		MonoObject *cb_exc = NULL;
 
@@ -640,8 +642,9 @@ mono_async_invoke (ThreadPool *tp, MonoAsyncResult *ares)
 		if (ac != NULL && ac->cb_method) {
 			void *pa = &ares;
 			cb_exc = NULL;
+			thread->async_invoke_method = ac->cb_method;
 			mono_runtime_invoke (ac->cb_method, ac->cb_target, pa, &cb_exc);
-			MONO_OBJECT_SETREF (ac->msg, exc, cb_exc);
+			thread->async_invoke_method = NULL;
 			exc = cb_exc;
 		} else {
 			exc = NULL;
@@ -671,7 +674,9 @@ threadpool_start_idle_threads (ThreadPool *tp)
 			if (InterlockedCompareExchange (&tp->nthreads, n + 1, n) == n)
 				break;
 		}
+#ifndef DISABLE_PERFCOUNTERS
 		mono_perfcounter_update_value (tp->pc_nthreads, TRUE, 1);
+#endif
 		mono_thread_create_internal (mono_get_root_domain (), tp->async_invoke, tp, TRUE, stack_size);
 		SleepEx (100, TRUE);
 	} while (1);
@@ -688,6 +693,7 @@ threadpool_init (ThreadPool *tp, int min_threads, int max_threads, void (*async_
 	MONO_SEM_INIT (&tp->new_job, 0);
 }
 
+#ifndef DISABLE_PERFCOUNTERS
 static void *
 init_perf_counter (const char *category, const char *counter)
 {
@@ -706,6 +712,7 @@ init_perf_counter (const char *category, const char *counter)
 	machine = mono_string_new (root, ".");
 	return mono_perfcounter_get_impl (category_str, counter_str, NULL, machine, &type, &custom);
 }
+#endif
 
 #ifdef DEBUG
 static void
@@ -724,11 +731,11 @@ print_pool_info (ThreadPool *tp)
 	g_print ("Queued: %d\n", (tp->tail - tp->head));
 	if (tp == &async_tp) {
 		int i;
-		EnterCriticalSection (&wsqs_lock);
+		mono_mutex_lock (&wsqs_lock);
 		for (i = 0; i < wsqs->len; i++) {
 			g_print ("\tWSQ %d: %d\n", i, mono_wsq_count (g_ptr_array_index (wsqs, i)));
 		}
-		LeaveCriticalSection (&wsqs_lock);
+		mono_mutex_unlock (&wsqs_lock);
 	} else {
 		g_print ("\tSockets: %d\n", mono_g_hash_table_size (socket_io_data.sock_to_state));
 	}
@@ -762,9 +769,10 @@ monitor_thread (gpointer unused)
 	pools [0] = &async_tp;
 	pools [1] = &async_io_tp;
 	thread = mono_thread_internal_current ();
-	ves_icall_System_Threading_Thread_SetName_internal (thread, mono_string_new (mono_domain_get (), "Threapool monitor"));
+	ves_icall_System_Threading_Thread_SetName_internal (thread, mono_string_new (mono_domain_get (), "Threadpool monitor"));
 	while (1) {
 		ms = 500;
+		i = 10; //number of spurious awakes we tolerate before doing a round of rebalancing.
 		do {
 			guint32 ts;
 			ts = mono_msec_ticks ();
@@ -775,11 +783,14 @@ monitor_thread (gpointer unused)
 				break;
 			if (THREAD_WANTS_A_BREAK (thread))
 				mono_thread_interruption_checkpoint ();
-		} while (ms > 0);
+		} while (ms > 0 && i--);
 
 		if (mono_runtime_is_shutting_down ())
 			break;
 
+		if (suspended)
+			continue;
+
 		for (i = 0; i < 2; i++) {
 			ThreadPool *tp;
 			tp = pools [i];
@@ -787,7 +798,7 @@ monitor_thread (gpointer unused)
 				continue;
 			need_one = (mono_cq_count (tp->queue) > 0);
 			if (!need_one && !tp->is_io) {
-				EnterCriticalSection (&wsqs_lock);
+				mono_mutex_lock (&wsqs_lock);
 				for (i = 0; wsqs != NULL && i < wsqs->len; i++) {
 					MonoWSQ *wsq;
 					wsq = g_ptr_array_index (wsqs, i);
@@ -796,7 +807,7 @@ monitor_thread (gpointer unused)
 						break;
 					}
 				}
-				LeaveCriticalSection (&wsqs_lock);
+				mono_mutex_unlock (&wsqs_lock);
 			}
 			if (need_one)
 				threadpool_start_thread (tp);
@@ -805,7 +816,13 @@ monitor_thread (gpointer unused)
 }
 
 void
-mono_thread_pool_init ()
+mono_thread_pool_init_tls (void)
+{
+	mono_wsq_init ();
+}
+
+void
+mono_thread_pool_init (void)
 {
 	gint threads_per_cpu = 1;
 	gint thread_count;
@@ -825,7 +842,7 @@ mono_thread_pool_init ()
 	}
 
 	MONO_GC_REGISTER_ROOT_FIXED (socket_io_data.sock_to_state);
-	InitializeCriticalSection (&socket_io_data.io_lock);
+	mono_mutex_init_recursive (&socket_io_data.io_lock);
 	if (g_getenv ("MONO_THREADS_PER_CPU") != NULL) {
 		threads_per_cpu = atoi (g_getenv ("MONO_THREADS_PER_CPU"));
 		if (threads_per_cpu < 1)
@@ -840,10 +857,10 @@ mono_thread_pool_init ()
 	async_call_klass = mono_class_from_name (mono_defaults.corlib, "System", "MonoAsyncCall");
 	g_assert (async_call_klass);
 
-	InitializeCriticalSection (&wsqs_lock);
+	mono_mutex_init_recursive (&wsqs_lock);
 	wsqs = g_ptr_array_sized_new (MAX (100 * cpu_count, thread_count));
-	mono_wsq_init ();
 
+#ifndef DISABLE_PERFCOUNTERS
 	async_tp.pc_nitems = init_perf_counter ("Mono Threadpool", "Work Items Added");
 	g_assert (async_tp.pc_nitems);
 
@@ -855,6 +872,7 @@ mono_thread_pool_init ()
 
 	async_io_tp.pc_nthreads = init_perf_counter ("Mono Threadpool", "# of IO Threads");
 	g_assert (async_io_tp.pc_nthreads);
+#endif
 	tp_inited = 2;
 #ifdef DEBUG
 	signal (SIGALRM, signal_handler);
@@ -988,12 +1006,12 @@ mono_thread_pool_cleanup (void)
 	}
 
 	if (wsqs) {
-		EnterCriticalSection (&wsqs_lock);
+		mono_mutex_lock (&wsqs_lock);
 		mono_wsq_cleanup ();
 		if (wsqs)
 			g_ptr_array_free (wsqs, TRUE);
 		wsqs = NULL;
-		LeaveCriticalSection (&wsqs_lock);
+		mono_mutex_unlock (&wsqs_lock);
 		MONO_SEM_DESTROY (&async_tp.new_job);
 	}
 }
@@ -1007,7 +1025,9 @@ threadpool_start_thread (ThreadPool *tp)
 	stack_size = (!tp->is_io) ? 0 : SMALL_STACK;
 	while (!mono_runtime_is_shutting_down () && (n = tp->nthreads) < tp->max_threads) {
 		if (InterlockedCompareExchange (&tp->nthreads, n + 1, n) == n) {
+#ifndef DISABLE_PERFCOUNTERS
 			mono_perfcounter_update_value (tp->pc_nthreads, TRUE, 1);
+#endif
 			mono_thread_create_internal (mono_get_root_domain (), tp->async_invoke, tp, TRUE, stack_size);
 			return TRUE;
 		}
@@ -1046,12 +1066,16 @@ threadpool_append_jobs (ThreadPool *tp, MonoObject **jobs, gint njobs)
 		return;
 
 	if (tp->pool_status == 0 && InterlockedCompareExchange (&tp->pool_status, 1, 0) == 0) {
-		if (!tp->is_io)
+		if (!tp->is_io) {
 			mono_thread_create_internal (mono_get_root_domain (), monitor_thread, NULL, TRUE, SMALL_STACK);
+			threadpool_start_thread (tp);
+		}
 		/* Create on demand up to min_threads to avoid startup penalty for apps that don't use
 		 * the threadpool that much
-		* mono_thread_create_internal (mono_get_root_domain (), threadpool_start_idle_threads, tp, TRUE, SMALL_STACK);
-		*/
+		 */
+		if (mono_config_is_server_mode ()) {
+			mono_thread_create_internal (mono_get_root_domain (), threadpool_start_idle_threads, tp, TRUE, SMALL_STACK);
+		}
 	}
 
 	for (i = 0; i < njobs; i++) {
@@ -1063,7 +1087,9 @@ threadpool_append_jobs (ThreadPool *tp, MonoObject **jobs, gint njobs)
 			o->add_time = mono_100ns_ticks ();
 		}
 		threadpool_jobs_inc (ar); 
+#ifndef DISABLE_PERFCOUNTERS
 		mono_perfcounter_update_value (tp->pc_nitems, TRUE, 1);
+#endif
 		if (!tp->is_io && mono_wsq_local_push (ar))
 			continue;
 
@@ -1078,10 +1104,13 @@ static void
 threadpool_clear_queue (ThreadPool *tp, MonoDomain *domain)
 {
 	MonoObject *obj;
-	MonoMList *other;
+	MonoMList *other = NULL;
+	MonoCQ *queue = tp->queue;
 
-	other = NULL;
-	while (mono_cq_dequeue (tp->queue, &obj)) {
+	if (!queue)
+		return;
+
+	while (mono_cq_dequeue (queue, &obj)) {
 		if (obj == NULL)
 			continue;
 		if (obj->vtable->domain != domain)
@@ -1089,12 +1118,32 @@ threadpool_clear_queue (ThreadPool *tp, MonoDomain *domain)
 		threadpool_jobs_dec (obj);
 	}
 
+	if (mono_runtime_is_shutting_down ())
+		return;
+
 	while (other) {
 		threadpool_append_job (tp, (MonoObject *) mono_mlist_get_data (other));
 		other = mono_mlist_next (other);
 	}
 }
 
+static gboolean
+remove_sockstate_for_domain (gpointer key, gpointer value, gpointer user_data)
+{
+	MonoMList *list = value;
+	gboolean remove = FALSE;
+	while (list) {
+		MonoObject *data = mono_mlist_get_data (list);
+		if (mono_object_domain (data) == user_data) {
+			remove = TRUE;
+			mono_mlist_set_data (list, NULL);
+		}
+		list = mono_mlist_next (list);
+	}
+	//FIXME is there some sort of additional unregistration we need to perform here?
+	return remove;
+}
+
 /*
  * Clean up the threadpool of all domain jobs.
  * Can only be called as part of the domain unloading process as
@@ -1112,6 +1161,12 @@ mono_thread_pool_remove_domain_jobs (MonoDomain *domain, int timeout)
 	threadpool_clear_queue (&async_tp, domain);
 	threadpool_clear_queue (&async_io_tp, domain);
 
+	mono_mutex_lock (&socket_io_data.io_lock);
+	if (socket_io_data.sock_to_state)
+		mono_g_hash_table_foreach_remove (socket_io_data.sock_to_state, remove_sockstate_for_domain, domain);
+
+	mono_mutex_unlock (&socket_io_data.io_lock);
+	
 	/*
 	 * There might be some threads out that could be about to execute stuff from the given domain.
 	 * We avoid that by setting up a semaphore to be pulsed by the thread that reaches zero.
@@ -1163,21 +1218,21 @@ add_wsq (void)
 	int i;
 	MonoWSQ *wsq;
 
-	EnterCriticalSection (&wsqs_lock);
+	mono_mutex_lock (&wsqs_lock);
 	wsq = mono_wsq_create ();
 	if (wsqs == NULL) {
-		LeaveCriticalSection (&wsqs_lock);
+		mono_mutex_unlock (&wsqs_lock);
 		return NULL;
 	}
 	for (i = 0; i < wsqs->len; i++) {
 		if (g_ptr_array_index (wsqs, i) == NULL) {
 			wsqs->pdata [i] = wsq;
-			LeaveCriticalSection (&wsqs_lock);
+			mono_mutex_unlock (&wsqs_lock);
 			return wsq;
 		}
 	}
 	g_ptr_array_add (wsqs, wsq);
-	LeaveCriticalSection (&wsqs_lock);
+	mono_mutex_unlock (&wsqs_lock);
 	return wsq;
 }
 
@@ -1189,9 +1244,9 @@ remove_wsq (MonoWSQ *wsq)
 	if (wsq == NULL)
 		return;
 
-	EnterCriticalSection (&wsqs_lock);
+	mono_mutex_lock (&wsqs_lock);
 	if (wsqs == NULL) {
-		LeaveCriticalSection (&wsqs_lock);
+		mono_mutex_unlock (&wsqs_lock);
 		return;
 	}
 	g_ptr_array_remove_fast (wsqs, wsq);
@@ -1207,7 +1262,7 @@ remove_wsq (MonoWSQ *wsq)
 		}
 	}
 	mono_wsq_destroy (wsq);
-	LeaveCriticalSection (&wsqs_lock);
+	mono_mutex_unlock (&wsqs_lock);
 }
 
 static void
@@ -1224,7 +1279,7 @@ try_steal (MonoWSQ *local_wsq, gpointer *data, gboolean retry)
 		if (mono_runtime_is_shutting_down ())
 			return;
 
-		EnterCriticalSection (&wsqs_lock);
+		mono_mutex_lock (&wsqs_lock);
 		for (i = 0; wsqs != NULL && i < wsqs->len; i++) {
 			MonoWSQ *wsq;
 
@@ -1233,11 +1288,11 @@ try_steal (MonoWSQ *local_wsq, gpointer *data, gboolean retry)
 				continue;
 			mono_wsq_try_steal (wsqs->pdata [i], data, ms);
 			if (*data != NULL) {
-				LeaveCriticalSection (&wsqs_lock);
+				mono_mutex_unlock (&wsqs_lock);
 				return;
 			}
 		}
-		LeaveCriticalSection (&wsqs_lock);
+		mono_mutex_unlock (&wsqs_lock);
 		ms += 10;
 	} while (retry && ms < 11);
 }
@@ -1245,9 +1300,10 @@ try_steal (MonoWSQ *local_wsq, gpointer *data, gboolean retry)
 static gboolean
 dequeue_or_steal (ThreadPool *tp, gpointer *data, MonoWSQ *local_wsq)
 {
-	if (mono_runtime_is_shutting_down ())
+	MonoCQ *queue = tp->queue;
+	if (mono_runtime_is_shutting_down () || !queue)
 		return FALSE;
-	mono_cq_dequeue (tp->queue, (MonoObject **) data);
+	mono_cq_dequeue (queue, (MonoObject **) data);
 	if (!tp->is_io && !*data)
 		try_steal (local_wsq, data, FALSE);
 	return (*data != NULL);
@@ -1340,26 +1396,63 @@ should_i_die (ThreadPool *tp)
 	return result;
 }
 
+static void
+set_tp_thread_info (ThreadPool *tp)
+{
+	const gchar *name;
+	MonoInternalThread *thread = mono_thread_internal_current ();
+
+	mono_profiler_thread_start (thread->tid);
+	name = (tp->is_io) ? "IO Threadpool worker" : "Threadpool worker";
+	mono_thread_set_name_internal (thread, mono_string_new (mono_domain_get (), name), FALSE);
+}
+
+static void
+clear_thread_state (void)
+{
+	MonoInternalThread *thread = mono_thread_internal_current ();
+	/* If the callee changes the background status, set it back to TRUE */
+	mono_thread_clr_state (thread , ~ThreadState_Background);
+	if (!mono_thread_test_state (thread , ThreadState_Background))
+		ves_icall_System_Threading_Thread_SetState (thread, ThreadState_Background);
+}
+
+static void
+check_for_interruption_critical (void)
+{
+	MonoInternalThread *thread;
+	/*RULE NUMBER ONE OF SKIP_THREAD: NEVER POKE MANAGED STATE.*/
+	mono_gc_set_skip_thread (FALSE);
+
+	thread = mono_thread_internal_current ();
+	if (THREAD_WANTS_A_BREAK (thread))
+		mono_thread_interruption_checkpoint ();
+
+	/*RULE NUMBER TWO OF SKIP_THREAD: READ RULE NUMBER ONE.*/
+	mono_gc_set_skip_thread (TRUE);
+}
+
+static void
+fire_profiler_thread_end (void)
+{
+	MonoInternalThread *thread = mono_thread_internal_current ();
+	mono_profiler_thread_end (thread->tid);
+}
+
 static void
 async_invoke_thread (gpointer data)
 {
 	MonoDomain *domain;
-	MonoInternalThread *thread;
 	MonoWSQ *wsq;
 	ThreadPool *tp;
 	gboolean must_die;
-	const gchar *name;
   
 	tp = data;
 	wsq = NULL;
 	if (!tp->is_io)
 		wsq = add_wsq ();
 
-	thread = mono_thread_internal_current ();
-
-	mono_profiler_thread_start (thread->tid);
-	name = (tp->is_io) ? "IO Threadpool worker" : "Threadpool worker";
-	ves_icall_System_Threading_Thread_SetName_internal (thread, mono_string_new (mono_domain_get (), name));
+	set_tp_thread_info (tp);
 
 	if (tp_start_func)
 		tp_start_func (tp_hooks_user_data);
@@ -1425,20 +1518,8 @@ async_invoke_thread (gpointer data)
 					exc = mono_async_invoke (tp, ar);
 					if (tp_item_end_func)
 						tp_item_end_func (tp_item_user_data);
-					if (exc && mono_runtime_unhandled_exception_policy_get () == MONO_UNHANDLED_POLICY_CURRENT) {
-						gboolean unloaded;
-						MonoClass *klass;
-
-						klass = exc->vtable->klass;
-						unloaded = is_appdomainunloaded_exception (exc->vtable->domain, klass);
-						if (!unloaded && klass != mono_defaults.threadabortexception_class) {
-							mono_unhandled_exception (exc);
-							if (mono_environment_exitcode_get () == 1)
-								exit (255);
-						}
-						if (klass == mono_defaults.threadabortexception_class)
-							mono_thread_internal_reset_abort (thread);
-					}
+					if (exc)
+						mono_internal_thread_unhandled_exception (exc);
 					if (is_socket && tp->is_io) {
 						MonoSocketAsyncResult *state = (MonoSocketAsyncResult *) data;
 
@@ -1453,10 +1534,7 @@ async_invoke_thread (gpointer data)
 				}
 				mono_thread_pop_appdomain_ref ();
 				InterlockedDecrement (&tp->busy_threads);
-				/* If the callee changes the background status, set it back to TRUE */
-				mono_thread_clr_state (thread , ~ThreadState_Background);
-				if (!mono_thread_test_state (thread , ThreadState_Background))
-					ves_icall_System_Threading_Thread_SetState (thread, ThreadState_Background);
+				clear_thread_state ();
 			}
 		}
 
@@ -1471,6 +1549,17 @@ async_invoke_thread (gpointer data)
 			gboolean res;
 
 			InterlockedIncrement (&tp->waiting);
+
+			// Another thread may have added a job into its wsq since the last call to dequeue_or_steal
+			// Check all the queues again before entering the wait loop
+			dequeue_or_steal (tp, &data, wsq);
+			if (data) {
+				InterlockedDecrement (&tp->waiting);
+				break;
+			}
+
+			mono_gc_set_skip_thread (TRUE);
+
 #if defined(__OpenBSD__)
 			while (mono_cq_count (tp->queue) == 0 && (res = mono_sem_wait (&tp->new_job, TRUE)) == -1) {// && errno == EINTR) {
 #else
@@ -1478,10 +1567,12 @@ async_invoke_thread (gpointer data)
 #endif
 				if (mono_runtime_is_shutting_down ())
 					break;
-				if (THREAD_WANTS_A_BREAK (thread))
-					mono_thread_interruption_checkpoint ();
+				check_for_interruption_critical ();
 			}
 			InterlockedDecrement (&tp->waiting);
+
+			mono_gc_set_skip_thread (FALSE);
+
 			if (mono_runtime_is_shutting_down ())
 				break;
 			must_die = should_i_die (tp);
@@ -1506,12 +1597,14 @@ async_invoke_thread (gpointer data)
 				if (!down && nt <= tp->min_threads)
 					break;
 				if (down || InterlockedCompareExchange (&tp->nthreads, nt - 1, nt) == nt) {
+#ifndef DISABLE_PERFCOUNTERS
 					mono_perfcounter_update_value (tp->pc_nthreads, TRUE, -1);
+#endif
 					if (!tp->is_io) {
 						remove_wsq (wsq);
 					}
 
-					mono_profiler_thread_end (thread->tid);
+					fire_profiler_thread_end ();
 
 					if (tp_finish_func)
 						tp_finish_func (tp_hooks_user_data);
@@ -1625,3 +1718,39 @@ mono_install_threadpool_item_hooks (MonoThreadPoolItemFunc begin_func, MonoThrea
 	tp_item_user_data = user_data;
 }
 
+void
+mono_internal_thread_unhandled_exception (MonoObject* exc)
+{
+	if (mono_runtime_unhandled_exception_policy_get () == MONO_UNHANDLED_POLICY_CURRENT) {
+		gboolean unloaded;
+		MonoClass *klass;
+
+		klass = exc->vtable->klass;
+		unloaded = is_appdomainunloaded_exception (exc->vtable->domain, klass);
+		if (!unloaded && klass != mono_defaults.threadabortexception_class) {
+			mono_unhandled_exception (exc);
+			if (mono_environment_exitcode_get () == 1)
+				exit (255);
+		}
+		if (klass == mono_defaults.threadabortexception_class)
+		 mono_thread_internal_reset_abort (mono_thread_internal_current ());
+	}
+}
+
+/*
+ * Suspend creation of new threads.
+ */
+void
+mono_thread_pool_suspend (void)
+{
+	suspended = TRUE;
+}
+
+/*
+ * Resume creation of new threads.
+ */
+void
+mono_thread_pool_resume (void)
+{
+	suspended = FALSE;
+}
