@@ -2,6 +2,7 @@
 #define __MONO_MINI_S390X_H__
 
 #include <mono/arch/s390x/s390x-codegen.h>
+#include <mono/utils/mono-context.h>
 #include <signal.h>
 
 #define MONO_ARCH_CPU_SPEC s390x_cpu_desc
@@ -13,29 +14,6 @@
 /* Parameters used by the register allocator */
 /*-------------------------------------------*/
 
-#define S390_LONG(loc, opy, op, r, ix, br, off)					\
-	if (has_ld) {								\
-		if (s390_is_imm20(off)) {					\
-			s390_##opy (loc, r, ix, br, off);			\
-		} else {							\
-			s390_basr (code, s390_r13, 0);				\
-			s390_j    (code, 6);					\
-			s390_llong(code, off);					\
-			s390_lg   (code, s390_r13, 0, s390_r13, 4);		\
-			s390_##op (code, r, s390_r13, br, 0);			\
-		}								\
-	} else {								\
-		if (s390_is_uimm12(off)) {					\
-			s390_##op (loc, r, ix, br, off);			\
-		} else {							\
-			s390_basr (code, s390_r13, 0);				\
-			s390_j    (code, 6);					\
-			s390_llong(code, off);					\
-			s390_lg   (code, s390_r13, 0, s390_r13, 4);		\
-			s390_##op (code, r, s390_r13, br, 0);			\
-		}								\
-	}
-
 struct MonoLMF {
 	gpointer    previous_lmf;
 	gpointer    lmf_addr;
@@ -47,8 +25,6 @@ struct MonoLMF {
 	gdouble     fregs[16];
 };
 
-typedef struct ucontext MonoContext;
-
 typedef struct MonoCompileArch {
 	gpointer    litpool;
 	glong	    litsize;
@@ -64,18 +40,30 @@ typedef struct
 } MonoS390StackFrame;
 
 // #define MONO_ARCH_SIGSEGV_ON_ALTSTACK		1
-#define MONO_ARCH_EMULATE_LCONV_TO_R8_UN 	1
-#define MONO_ARCH_NO_EMULATE_LONG_MUL_OPTS	1
-#define MONO_ARCH_NO_EMULATE_LONG_SHIFT_OPS	1
-#define MONO_ARCH_HAVE_IS_INT_OVERFLOW  	1
-#define MONO_ARCH_NEED_DIV_CHECK		1
-#define MONO_ARCH_HAVE_ATOMIC_ADD 1
-#define MONO_ARCH_HAVE_ATOMIC_EXCHANGE 1
-#define MONO_ARCH_SIGNAL_STACK_SIZE 		256*1024
-#define MONO_ARCH_HAVE_DECOMPOSE_OPTS 1
-// #define MONO_ARCH_HAVE_THROW_CORLIB_EXCEPTION	1
-
-#define MONO_ARCH_USE_SIGACTION 	1
+#define MONO_ARCH_EMULATE_LCONV_TO_R8_UN 		1
+#define MONO_ARCH_NO_EMULATE_LONG_MUL_OPTS		1
+#define MONO_ARCH_NO_EMULATE_LONG_SHIFT_OPS		1
+#define MONO_ARCH_HAVE_IS_INT_OVERFLOW  		1
+#define MONO_ARCH_NEED_DIV_CHECK			1
+#define MONO_ARCH_SIGNAL_STACK_SIZE 			256*1024
+#define MONO_ARCH_HAVE_DECOMPOSE_OPTS 			1
+#define MONO_ARCH_HAVE_CREATE_DELEGATE_TRAMPOLINE	1
+#define MONO_ARCH_HAVE_TLS_GET				1
+#define MONO_ARCH_ENABLE_MONO_LMF_VAR			1
+#define MONO_ARCH_IMT_REG				s390_r9
+#define MONO_ARCH_VTABLE_REG				MONO_ARCH_IMT_REG
+#define MONO_ARCH_RGCTX_REG				MONO_ARCH_IMT_REG
+#define MONO_ARCH_THIS_AS_FIRST_ARG     		1
+#define MONO_ARCH_HAVE_XP_UNWIND			1
+#define MONO_ARCH_HAVE_SIGCTX_TO_MONOCTX		1
+#define MONO_ARCH_SOFT_DEBUG_SUPPORTED			1
+#define MONO_ARCH_HAVE_CONTEXT_SET_INT_REG		1
+#define MONO_ARCH_USE_SIGACTION 			1
+#define MONO_ARCH_GC_MAPS_SUPPORTED			1
+#define MONO_ARCH_GSHARED_SUPPORTED			1
+#define MONO_ARCH_MONITOR_ENTER_ADJUSTMENT		1
+#define MONO_ARCH_HAVE_HANDLER_BLOCK_GUARD		1
+#define MONO_ARCH_HAVE_INVALIDATE_METHOD		1
 
 #define S390_STACK_ALIGNMENT		 8
 #define S390_FIRST_ARG_REG 		s390_r2
@@ -98,6 +86,7 @@ typedef struct
 /* s390_r8..s390_r11 are used for global regalloc       */
 /* s390_r15 is the stack pointer                        */
 /*------------------------------------------------------*/
+
 #define MONO_ARCH_CALLEE_REGS (0xfc)
 
 #define MONO_ARCH_CALLEE_SAVED_REGS 0xff80
@@ -142,32 +131,16 @@ typedef struct
 #define S390_CC_EQ			8
 #define S390_ALIGN(v, a)	(((a) > 0 ? (((v) + ((a) - 1)) & ~((a) - 1)) : (v)))
 
-#define MONO_CONTEXT_SET_IP(ctx,ip) 					\
-	do {								\
-		(ctx)->uc_mcontext.gregs[14] = (unsigned long)ip;	\
-		(ctx)->uc_mcontext.psw.addr = (unsigned long)ip;	\
-	} while (0); 
-
-#define MONO_CONTEXT_SET_SP(ctx,bp) MONO_CONTEXT_SET_BP((ctx),(bp))
-#define MONO_CONTEXT_SET_BP(ctx,bp) 					\
-	do {		 						\
-		(ctx)->uc_mcontext.gregs[15] = (unsigned long)bp;	\
-		(ctx)->uc_stack.ss_sp	     = (void*)bp;		\
-	} while (0) 
-
-#define MONO_CONTEXT_GET_IP(ctx) (gpointer) (ctx)->uc_mcontext.psw.addr
-#define MONO_CONTEXT_GET_BP(ctx) MONO_CONTEXT_GET_SP((ctx))
-#define MONO_CONTEXT_GET_SP(ctx) ((gpointer)((ctx)->uc_mcontext.gregs[15]))
-
 #define MONO_INIT_CONTEXT_FROM_FUNC(ctx,func) do {			\
 		MonoS390StackFrame *sframe;				\
 		__asm__ volatile("lgr   %0,15" : "=r" (sframe));	\
 		MONO_CONTEXT_SET_BP ((ctx), sframe->prev);		\
 		MONO_CONTEXT_SET_SP ((ctx), sframe->prev);		\
-		sframe = (MonoS390StackFrame*)sframe->prev;		\
-		MONO_CONTEXT_SET_IP ((ctx), sframe->return_address);	\
+		MONO_CONTEXT_SET_IP ((ctx), func);			\
 	} while (0)
 
+#define MONO_ARCH_INIT_TOP_LMF_ENTRY(lmf) do { (lmf)->ebp = -1; } while (0)
+
 /*------------------------------------------------------------------*/
 /*                                                                  */
 /* Name		- s390_patch_rel                                    */
